(function() {
    
    if (!OCA.CloudsuiteGearsIntegration) {
		/**
		 * Namespace for the SignLive! app
		 * @namespace OCA.CloudsuiteGearsIntegration
		 */
         OCA.CloudsuiteGearsIntegration = {};
	}

    class SignatureRequestCheckerRunner {
        constructor(conversation_id, timeout_value) {
          this.conversation_id = conversation_id;
          this.timeout_value = timeout_value;

          this.check_aborted = false;
          this.start_time = Date(0);
          this.total_iterations = 0;
        }

        async checkSignatureStatus() {
            console.debug('Check signature status for conversation_id: ' + this.conversation_id);
            this.start_time = Date.now();
            while (true) {
                await new Promise(r => setTimeout(r, this.timeout_value / 10.0));

                this.total_iterations = this.total_iterations + 1;
                if (this.check_aborted == true || this.total_iterations >= 1000) {
                    var total_runtime = Date.now() - this.start_time;
                    console.debug('Signature request was aborted, leaving checkSignature for conversation_id ' + this.conversation_id + ' after ' + (total_runtime.getMilliseconds() / 1000.0) + ' seconds.');
                    break;
                }
                console.debug('Continuing signature status monitoring for conversation_id: ' + this.conversation_id);
            }
        }

        async abortSignatureRequest() {
            console.warn('Aborting signature status monitoring for conversation_id: ' + this.conversation_id+ ' after ' + (total_runtime.getMilliseconds() / 1000.0) + ' seconds.');
            this.check_aborted = true;
        }
    }

    class SignatureRequestCheckDispatcher {
        constructor() {
            this.request_checker_instances = {};
        }

        createSignatureChecker(conversation_id, timeout_value) {
            console.debug('Creating new SignatureRequestCheckerRunner instance for conversation_id \'' + conversation_id + '\' with timeout value ' + timeout_value);
            this.request_checker_instances = new SignatureRequestCheckerRunner(conversation_id, timeout_value);
        }

        startSignatureChecker(conversation_id) {
            console.debug('Starting SignatureRequestCheckerRunner instance for conversation_id ' + conversation_id);
            if (conversation_id in this.request_checker_instances) {
                this.request_checker_instances[conversation_id].checkSignatureStatus();
            } else {
                console.warn('Tried to start non-existent SignatureRequestCheckerRunner: ' + conversation_id);
            }
        }

        deleteSignatureChecker(conversation_id) {
            console.debug('Deleting SignatureRequestCheckerRunner instance for conversation_id \'' + conversation_id + '\'');
            if (conversation_id in this.request_checker_instances) {
                console.debug('Stopping and removing SignatureRequestCheckerRunner instance ' + conversation_id);
                this.request_checker_instances[conversation_id].abortSignatureRequest();
                delete this.request_checker_instances[conversation_id];
            } else {
                console.warn('Attempted to delete a non-existent SignatureRequestCheckerRunner instance: ' + conversation_id);
            }
        }
    }

    class SignatureJobStatusNotifier {
        start_time = Date(0);
        total_iterations = 0;
        job_expiration_interval = 600;
        timeout_value = 1000.0;
        
        constructor() {
            return this.init();
        }

        init() {
            this.start_time = Date(0);
            this.total_iterations = 0;

            this.job_expiration_interval = 600; // seconds
            this.timeout_value = 1000.0; // milliseconds

            this.fileList = null;
            if (OCA === null || OCA === undefined || OCA.Files === null || OCA.Files === undefined || 
                OCA.Files.App === null || OCA.Files.App === undefined || OCA.Files.App.fileList === null || OCA.Files.App.fileList === undefined) {
                console.warn('OCA.Files.fileList instance is invalid or not yet defined, can not refresh file list after signature request jobs finish!');
            } else {
                console.debug('Retrieving OCA.Files.fileList object.');
                this.fileList = OCA.Files.App.fileList;
            }
        }

        async checkSignatureJobsStatus() {
            console.debug('Start check loop for new signature job status notifications.');
            this.start_time = Date.now();
        
            var job_expiration_interval = this.job_expiration_interval;
            if (job_expiration_interval === null || job_expiration_interval === undefined) {
                job_expiration_interval = 600;
            }

            var timeout_value = this.timeout_value;
            if (timeout_value === null || timeout_value === undefined) {
                timeout_value = 1000.0;
            }

            console.debug('start_time = ' + this.start_time);
            console.debug('job_expiration_interval = ' + job_expiration_interval);

            var status_icon_success_url = window.location.protocol + "//" + window.location.host + '/apps/signlive/img/checkmark.svg';
            var status_icon_error_url = window.location.protocol + "//" + window.location.host + '/apps/signlive/img/error.svg';

            var jobs_to_remove_from_status_queue = [];
            var jobs_to_notify_user_of = {};
            var num_jobs_to_notify_user_of = 0;
            var notification_text = '';
            while (true) {
                await new Promise(r => setTimeout(r, timeout_value));

                jobs_to_remove_from_status_queue = [];
                jobs_to_notify_user_of = {};
                num_jobs_to_notify_user_of = 0;

                var rest_url = OC.generateUrl('/apps/signlive/sign_document/get_job_definition_entries');

                // Initialize with default value in case the config option is not available from the backend for whatever reason
                var notification_timeout_in_seconds = 7;
                var show_notifications_for_successful_signature_requests = 0;
                var show_verbose_notifications_for_failed_signature_requests = 0;

                $.ajax({
                    url: rest_url,
                    async : false,
                    type: 'post',
                    success: (response) => {
                        // console.debug('Response received for job definition entries query after waiting for ' + timeout_value + ' ms.');
                        for (const job_uuid in response) {
                            if (job_uuid === 'cloudsuite_gears_notification_duration') {
                                notification_timeout_in_seconds = response[job_uuid];
                                // console.debug('notification_timeout_in_seconds = ' + notification_timeout_in_seconds);
                            } else if (job_uuid === 'cloudsuite_gears_show_success_notifications') {
                                show_notifications_for_successful_signature_requests = response[job_uuid];
                                // console.debug('show_notifications_for_successful_signature_requests = ' + show_notifications_for_successful_signature_requests);
                            } else if (job_uuid === 'cloudsuite_gears_show_verbose_failure_notifications') {
                                show_verbose_notifications_for_failed_signature_requests = response[job_uuid];
                                // console.debug('show_verbose_notifications_for_failed_signature_requests = ' + show_verbose_notifications_for_failed_signature_requests);
                            }
                            else if (job_uuid === 'cloudsuite_gears_notification_check_interval') {
                                timeout_value = parseInt(response[job_uuid]) * 1000.0;
                                // console.debug('timeout_value = ' + timeout_value);
                            }
                            else {
                                console.debug('************************************************');
                                console.debug('Signature job: ' + job_uuid);
                                console.debug('************************************************');
                                
                                var job_status_entry = response[job_uuid];

                                const current_time_epoch = Math.floor(Date.now() / 1000.0);
                                const job_creation_time_epoch = Math.floor(parseInt(job_status_entry['job_creation_time']));

                                const job_lifetime = current_time_epoch - job_creation_time_epoch;
                                console.debug('Time in seconds since job creation time: ' + job_lifetime);
                                if (job_lifetime > job_expiration_interval) {
                                    console.debug('Job ' + job_uuid + ' has timed out, deleting it from job status queue.');
                                    jobs_to_remove_from_status_queue.push(job_uuid);
                                } else {
                                    jobs_to_notify_user_of[job_uuid] = job_status_entry;
                                    num_jobs_to_notify_user_of = num_jobs_to_notify_user_of + 1;
                                }
                            }
                        }

                        // console.debug('jobs_to_remove_from_status_queue.length: ' + jobs_to_remove_from_status_queue.length);
                        // console.debug('num_jobs_to_notify_user_of: ' + num_jobs_to_notify_user_of);              
                    },
                    error: (response) => {
                        console.error('Failed to query signature job status entries from Nextcloud backend: ' + response.message);
                    }
                });

                if (jobs_to_remove_from_status_queue.length > 0) {
                    console.debug('==============================================================');
                    console.debug('Jobs to remove from status queue: ' + jobs_to_remove_from_status_queue.length);
                    console.debug('==============================================================');
                
                    var num_jobs_removed = 0;
                    for (const job_uuid in jobs_to_remove_from_status_queue) {
                        console.debug('Acknowleding expiry of notification(s) for job_uuid ' + jobs_to_remove_from_status_queue[job_uuid]);
                        var ack_url = OC.generateUrl('/apps/signlive/sign_document/confirm_notification');
                        $.ajax({
                            url: ack_url,
                            async: false,
                            type: 'post',
                            data: {
                                job_uuid: jobs_to_remove_from_status_queue[job_uuid],
                                ack_reason: 'JOB_EXPIRED'
                            },
                            success: (response) => {
                                console.debug('Expired signature job entry for job_uuid ' + jobs_to_remove_from_status_queue[job_uuid] + ' successfully removed from signature job status entry queue: ' + response.message);
                            },
                            error: (response) => {
                                console.debug('Failed to remove expired signature job entry for job_uuid ' + jobs_to_remove_from_status_queue[job_uuid] + ' from signature job status entry queue: ' + response.message);
                            }
                        });
                        num_jobs_removed = num_jobs_removed + 1;
                    }
                } 

                
                if (num_jobs_to_notify_user_of > 0) {
                    console.debug('==============================================================');
                    console.debug('Jobs to notify user of: ' + num_jobs_to_notify_user_of);
                    
                    var notification_text = '';
                    var failure_notification_text = '';
                    var num_successful_jobs = 0;
                    var num_failed_jobs = 0;
                    for (const [job_uuid, job_data] of Object.entries(jobs_to_notify_user_of)) {
                        console.debug('==== Job UUID: ' + job_uuid + ' ====');
                        console.debug('job_data[\'file_name\']: ' + job_data['file_name']);
                        console.debug('job_data[\'nextcloud_user\']: ' + job_data['nextcloud_user']);
                        console.debug('job_data[\'signature_type\']: ' + job_data['signature_type']);
                        console.debug('job_data[\'job_creation_time\']: ' + job_data['job_creation_time']);
                        console.debug('job_data[\'job_stage\']: ' + job_data['job_stage']);
                        console.debug('job_data[\'status_message\']: ' + job_data['status_message']);
                        console.debug('job_data[\'job_failed\']: ' + job_data['job_failed']);

                        // Job success assumes 'sendSuccessNotification' has been called in IntarsysJobSignatureController.php
                        var ack_url = OC.generateUrl('/apps/signlive/sign_document/confirm_notification');
                        if (job_data['job_stage'] === 'SIGNATURE_JOB_COMPLETED_SUCCESSFULLY' ||
                            job_data['job_stage'] === 'RESULT_RETRIEVAL_SUCCEEDED') {
                            notification_text = '<li><img src="' + status_icon_success_url + '"/>' + notification_text + job_data['file_name'] + '</li>';
                            num_successful_jobs = num_successful_jobs + 1;
                            // Notification has been shown: Acknowledge notification to remove signature request job entry from database
                            console.debug('Acknowleding display of notification for job_uuid ' + job_uuid);
                            
                            $.ajax({
                                url: ack_url,
                                async: false,
                                type: 'post',
                                data: {
                                    job_uuid: job_uuid,
                                    ack_reason: 'JOB_NOTIFICATION_DISPLAYED'
                                },
                                success: (response) => {
                                    console.debug('Acknowledged notification for signature job entry with job_uuid ' + job_uuid + ': ' + response.message);
                                },
                                error: (response) => {
                                    console.warn('Failed to acknowledge notification for signature job entry with job_uuid ' + job_uuid + ': ' + response.message);
                                }
                            });
                        } else if (job_data['job_failed'] === '1' ||
                                   job_data['job_failed'] === 1) {
                            console.debug('Job with UUID ' + job_uuid + ' is marked as FAILED.');
                            
                            failure_notification_text = failure_notification_text + '<li><img src="' +  status_icon_error_url + '"/>' + job_data['file_name'] + '</li>';
                            num_failed_jobs = num_failed_jobs + 1;

                            // Acknowledge failure notification if no persistent notifications for failed jobs should be shown
                            // I. e. failure notifications through the Notifications app
                            if (show_verbose_notifications_for_failed_signature_requests === 0 ||
                                show_verbose_notifications_for_failed_signature_requests === false ||
                                show_verbose_notifications_for_failed_signature_requests === "false") {
                                console.debug('Acknownleding display of notification for FAILED job with job_uuid: ' + job_uuid);
                                $.ajax({
                                    url: ack_url,
                                    async: false,
                                    type: 'post',
                                    data: {
                                        job_uuid: job_uuid,
                                        ack_reason: 'JOB_NOTIFICATION_DISPLAYED'
                                    },
                                    success: (response) => {
                                        console.debug('Acknowledged notification for signature job entry with job_uuid ' + job_uuid + ': ' + response.message);
                                    },
                                    error: (response) => {
                                        console.warn('Failed to acknowledge notification for  signature job entry with job_uuid ' + job_uuid + ': ' + response.message);
                                    }
                                });
                            }

                        }
                    }

                    console.debug('*********************************************');
                    console.debug('Notification for failed signature jobs...');
                    console.debug('num_failed_jobs = ' + num_failed_jobs);
                    console.debug('*********************************************');

                    if (num_failed_jobs > 0) {
                        var display_text = '';
                        if (num_failed_jobs === 1) {
                            display_text = OC.L10N.translate(
                                'signlive',
                                'One signature request has failed') + ': ';
                        } else {
                            display_text = OC.L10N.translate(
                                'signlive',
                                '{num_failed_jobs} signature requests have failed',
                                {num_failed_jobs: num_failed_jobs}) + ': ';
                        }

                        var status_text = '<p style="color:red;">' + display_text + '</p><ul style="list-style-type:none">' + failure_notification_text + '</ul>';
                        console.debug('Notifications to show for failed signature requests: ' + status_text);
                        
                        OC.Notification.showTemporary(status_text, {timeout: notification_timeout_in_seconds * 1000.0, isHTML: true});
                    }

                    console.debug('*********************************************');
                    console.debug('Notification for successful signature jobs...');
                    console.debug('num_successful_jobs = ' + num_successful_jobs);
                    console.debug('show_notifications_for_successful_signature_requests = ' + show_notifications_for_successful_signature_requests);
                    console.debug('*********************************************');

                    if (num_successful_jobs > 0) {
                        // Refresh the files view in the web UI after a successful signature request to update file modification dates
                        if (this.fileList !== null) {
                            console.debug('Dispatching fileList reload() call after receiving successful status for one or more signature request jobs.');
                            this.fileList.reload();
                        } else {
                            console.warn('Could not dispatch fileList reload() call after successful completion of one or more signature request jobs! fileList object is not valid.');
                        }

                        if (show_notifications_for_successful_signature_requests === 0 || 
                            show_notifications_for_successful_signature_requests === false || 
                            show_notifications_for_successful_signature_requests === 'false') {
                            console.debug('Not showing notifications for successfully processed signature requests.');
                        } else {
                            var display_text = '';
                            if (num_successful_jobs === 1) {
                                display_text = OC.L10N.translate(
                                    'signlive',
                                    'One signature request has been processed successfully') + ': ';
                            } else {
                                display_text = OC.L10N.translate(
                                    'signlive',
                                    '{num_jobs_to_notify_user_of} signature requests have been processed successfully',
                                    {num_jobs_to_notify_user_of: num_jobs_to_notify_user_of}) + ': ';
                            }

                            if (this.fileList === null) {
                                if (OCA === null || OCA === undefined || OCA.Files === null || OCA.Files === undefined || 
                                    OCA.Files.App === null || OCA.Files.App === undefined || OCA.Files.App.fileList === null || OCA.Files.App.fileList === undefined) {
                                    console.warn('OCA.Files.fileList instance is invalid or not yet defined, can not refresh file list after signature request jobs finish!');
                                } else {
                                    console.debug('Retrieving OCA.Files.fileList object.');
                                    this.fileList = OCA.Files.App.fileList;
                                }
                            }

                            var status_text = '<p>' + display_text + '</p><ul style="list-style-type:none">' + notification_text + '</ul>';
                            console.debug('Notifications to show for successful signature requests: ' + status_text);
                            
                            OC.Notification.showTemporary(status_text, {timeout: notification_timeout_in_seconds * 1000.0, isHTML: true});
                        }
                    }
                    console.debug('==============================================================');
                }
                this.total_iterations = this.total_iterations + 1;
                // console.debug('Continuing signature job status notification loop, iteration: ' + this.total_iterations);
            }
        }
    }
    
    OCA.CloudsuiteGearsIntegration.SignatureRequestCheckerRunner = SignatureRequestCheckerRunner;
    OCA.CloudsuiteGearsIntegration.SignatureJobStatusNotifier = SignatureJobStatusNotifier;
})();