<?php

/**
* This file is licensed under the Affero General Public License version 3 or
* later. See the COPYING file.
*
* @author Fabian Aichele
* @copyright intarsys GmbH 2021
*/

namespace OCA\IntarsysCloudSuiteGearsIntegration\Controller;

use OCP\IConfig;
use OCP\IRequest;
use OCP\Util;
use OCP\ILogger;
use OCP\IUserManager;
use OCP\IUserSession;
use OCP\Accounts\IAccount;
use OCP\Accounts\IAccountManager;
use OCP\IURLGenerator;
use OCP\AppFramework\Controller;
use OCP\Files\IRootFolder;
use OCP\Files\IMimeTypeDetector;
use OCP\Files\NotFoundException;
use OCP\Files\NotPermittedException;
use OCP\Files\File;
use OCP\Files\Folder;
use OCP\BackgroundJob\IJobList;
use OCP\AppFramework\Utility\ITimeFactory;
use OCP\Notification\IManager as INotificationManager;

use OCP\AppFramework\Http\DataResponse;
use OCP\AppFramework\Http\TemplateResponse;
use OCP\AppFramework\Http\RedirectResponse;

use OCA\IntarsysCloudSuiteGearsIntegration\Util\FileIO;
use OCA\IntarsysCloudSuiteGearsIntegration\Util\JSONQueryCreator;
use OCA\IntarsysCloudSuiteGearsIntegration\Util\SignLiveDefaultConfigValues;

use OCA\IntarsysCloudSuiteGearsIntegration\Notification\CloudsuiteGearsNotifier;

use GuzzleHttp\Client;
use GuzzleHttp\Message\Response;
use GuzzleHttp\Psr7;
use GuzzleHttp\Exception\RequestException;
use GuzzleHttp\Exception\ClientException;

class IntarsysJobSignatureController extends Controller {
    /** @var \OCP\IConfig */
    protected $config;

    /** @var \OCP\IURLGenerator */
    protected $urlGenerator;

    /** @var IUserSession */
    protected $userSession;

    /** @var INotificationManager */
	private $notificationManager;

    /** @var IUserManager */
    protected $userManager;

    /** @var IAccountManager */
    protected $accountManager;

    /** @var IRootFolder */
    protected $storage;

    /** @var IMimeTypeDetector */
    protected $mimeTypeDetector;

    /** @var IJobList */
    protected $jobList;

    /** @var ITimeFactory */
    protected $timeFactory;

    /** @var FileIO */
    protected $fileIO;

    /** @var JSONQueryCreator */
    protected $jsonQueryCreator;

    public function __construct(
        $appName,
        IRequest $request,
        IConfig $config,
        IURLGenerator $urlGenerator,
        IRootFolder $storage,
        IMimeTypeDetector $mimeTypeDetector,
        IUserSession $userSession,
        IUserManager $userManager,
        IAccountManager $accountManager,
        INotificationManager $notificationManager,
        IJobList $jobList,
        ITimeFactory $timeFactory) {
            parent::__construct($appName, $request);

            $this->request = $request;
            $this->config = $config;
            $this->urlGenerator = $urlGenerator;
            $this->storage = $storage;
            $this->mimeTypeDetector = $mimeTypeDetector;
            $this->userSession = $userSession;
            $this->userManager = $userManager;
            $this->accountManager = $accountManager;
            $this->jobList = $jobList;
            $this->timeFactory = $timeFactory;
            $this->notificationManager = $notificationManager;

            $current_time = $this->timeFactory->getDateTime('now');
            $this->controller_identifier = 'IntarsysJobSignatureController_'.$current_time->format('d.m.Y_h.i.s_u');

            $this->fileIO = new FileIO($this->config,
                                       $this->mimeTypeDetector,
                                       $this->storage,
                                       $this->userSession,
                                       $this->timeFactory);

            $this->jsonQueryCreator = new JSONQueryCreator($this->config,
                                                           $this->mimeTypeDetector,
                                                           $this->storage,
                                                           $this->userSession,
                                                           $this->userManager,
                                                           $this->accountManager,
                                                           $this->timeFactory);

            $this->cloudsuite_gears_http_protocol_type = $this->config->getAppValue('signlive', 'cloudsuite_gears_http_protocol_type', SignLiveDefaultConfigValues::$GEARS_SERVER_PROTOCOL);
            $this->cloudsuite_gears_host = $this->config->getAppValue('signlive', 'cloudsuite_gears_host', SignLiveDefaultConfigValues::$GEARS_SERVER_HOST);
            $this->cloudsuite_gears_port = $this->config->getAppValue('signlive', 'cloudsuite_gears_port', SignLiveDefaultConfigValues::$GEARS_SERVER_PORT);
            $this->cloudsuite_gears_gears_core_path = $this->config->getAppValue('signlive', 'cloudsuite_gears_gears_core_path', SignLiveDefaultConfigValues::$GEARS_SERVER_CORE_PATH);
            
            $this->gears_backend_url = $this->cloudsuite_gears_http_protocol_type.'://'.$this->cloudsuite_gears_host.':'.$this->cloudsuite_gears_port.$this->cloudsuite_gears_gears_core_path;

            $this->gears_rest_client = new \GuzzleHttp\Client();
    }

    /**
     * Get the server root URL including protocol
     * @return string
     */
    protected function url_origin( $s, $use_forwarded_host = false )
    {
        $ssl      = ( ! empty( $s['HTTPS'] ) && $s['HTTPS'] == 'on' );
        $sp       = strtolower( $s['SERVER_PROTOCOL'] );
        $protocol = substr( $sp, 0, strpos( $sp, '/' ) ) . ( ( $ssl ) ? 's' : '' );
        $port     = $s['SERVER_PORT'];
        $port     = ( ( ! $ssl && $port=='80' ) || ( $ssl && $port=='443' ) ) ? '' : ':'.$port;
        $host     = ( $use_forwarded_host && isset( $s['HTTP_X_FORWARDED_HOST'] ) ) ? $s['HTTP_X_FORWARDED_HOST'] : ( isset( $s['HTTP_HOST'] ) ? $s['HTTP_HOST'] : null );
        $host     = isset( $host ) ? $host : $s['SERVER_NAME'] . $port;
        return $protocol . '://' . $host;
    }
    
    /**
     * @NoAdminRequired
     * Get a list of file names to sign passed in the request parameters
     * @return array
     */
    protected function getFilesToSign() {
        \OCP\Util::writeLog('signlive', 'getFilesToSign()', ILogger::DEBUG);
        $file_list_to_sign = array();
        $request_parameters = $this->request->getParams();
        foreach ($request_parameters as $parameter_name => $parameter_value) {
            \OCP\Util::writeLog('signlive', 'Parameter \''.$parameter_name.'\' => \''.$parameter_value.'\'', ILogger::DEBUG);
            if ($parameter_name === 'pdf_documents_to_sign') {
                $file_list_to_sign = explode(";", $parameter_value);
                \OCP\Util::writeLog('signlive', 'List of files to sign: '.print_r($file_list_to_sign, TRUE), ILogger::DEBUG);                
            }
            if ($parameter_name === 'filename') {
                \OCP\Util::writeLog('signlive', 'Single file to sign: '.print_r($parameter_value, TRUE), ILogger::DEBUG);
                array_push($file_list_to_sign, $parameter_value);
            }
        }

        return $file_list_to_sign;
    }

    /**
     * @NoAdminRequired
     * Send a notification for a successfully processed signature request to the Nextcloud web frontend
     * @return bool
    */
    protected function sendSuccessNotification($signature_job_data) {
        \OCP\Util::writeLog('signlive', '=====================================================================', ILogger::DEBUG);
        \OCP\Util::writeLog('signlive', 'sendSuccessNotification() called in signlive/lib/Controller/IntarsysJobSignatureController.php', ILogger::DEBUG);
        
        foreach($signature_job_data as $data_key => $data_value) {
            \OCP\Util::writeLog('signlive', 'Job data attribute '.$data_key.' => '.$data_value, ILogger::DEBUG);
        }
        
        $job_uuid_found = array_key_exists("signature_job_uuid", $signature_job_data);
        $file_name_found = array_key_exists("file_name", $signature_job_data);
        $job_name_found = array_key_exists("signature_job_name", $signature_job_data);
        $profile_type_found = array_key_exists("signature_profile_type", $signature_job_data);
        $job_type_found =  array_key_exists("signature_type", $signature_job_data);
        $user_found = array_key_exists("nextcloud_user", $signature_job_data);
        $status_message_found = array_key_exists("status_message", $signature_job_data);

        \OCP\Util::writeLog('signlive', 'job_uuid_found = '.$job_uuid_found.'; file_name_found = '.$file_name_found.'; user_found = '.$user_found.'; job_type_found = '.$job_type_found, ILogger::DEBUG);

        $show_success_notifications = $this->config->getAppValue('signlive', 'cloudsuite_gears_show_success_notifications', SignLiveDefaultConfigValues::$SHOW_NOTIFICATIONS_FOR_SUCCESSFUL_SIGNATURE_REQUESTS);
        \OCP\Util::writeLog('signlive', 'show_success_notifications = '.$show_success_notifications, ILogger::DEBUG);

        \OCP\Util::writeLog('signlive', '=====================================================================', ILogger::DEBUG);

        // Return early if no success notifications are to be shown
        if ($show_success_notifications === 0) {
            \OCP\Util::writeLog('signlive', 'sendSuccessNotification() returns early: No success notifications are to be shown.', ILogger::DEBUG);
            return false;
        } 

        if ($job_uuid_found &&
            $file_name_found &&
            $user_found && 
            $job_type_found) {
            \OCP\Util::writeLog('signlive', 'sendSuccessNotification() dispatching a new notification to the web UI.', ILogger::DEBUG);
            \OCP\Util::writeLog('signlive', 'signature_job_uuid = \''.$signature_job_data['signature_job_uuid'].'\'; signature_job_file = '.$signature_job_data['signature_job_file'].'; nextcloud_user = '.$signature_job_data['nextcloud_user'].'; signature_job_type = '.$signature_job_data['signature_job_type'], ILogger::DEBUG);
            
            $signature_job_entry_data = $this->getJobDefinitionEntry($signature_job_data['signature_job_uuid']);

            $send_success_notification = $this->config->getAppValue('signlive', 'send_signature_job_success_notifications', SignLiveDefaultConfigValues::$SHOW_NOTIFICATIONS_FOR_SUCCESSFUL_SIGNATURE_REQUESTS);

            // When configured: Send a full notification via the Notification API on successful signature request completion
            if ($send_success_notification === true) {
                $object_id = $signature_job_data['signature_job_uuid'];
                $notification = $this->notificationManager->createNotification();
                try {
                    $notification->setApp('signlive');
                    $notification->setDateTime(new \DateTime());

                    \OCP\Util::writeLog('signlive', '$object_id strlen() = '.strlen($object_id), ILogger::DEBUG);

                    // The NotificationManager throws an InvalidArgumentException if the 'Object' of a notification is longer than 64 characters
                    if (strlen($object_id) >= 64) {
                        $object_id = substr($object_id, 0, 63);
                        \OCP\Util::writeLog('signlive', 'Truncated $object_id to: '.$object_id, ILogger::DEBUG);
                    }

                    $notification->setObject('signlive_job', $object_id);
                    $notification->setUser($signature_job_data['nextcloud_user']);

                    $message_and_subject_parameters = array(
                        'file_name' => $signature_job_data['file_name']
                    );
                    if ($signature_job_data['signature_job_type'] === 'viewer_based_signature') {
                        $notification->setSubject('viewer_based_signature_success', $message_and_subject_parameters);
                    } else if ($signature_job_data['signature_job_type'] === 'automatic_signature') {
                        $notification->setSubject('automatic_signature_success', $message_and_subject_parameters);
                    }
                    
                    $status_message = $signature_job_data['status_message'];
                    \OCP\Util::writeLog('signlive', '$status_message strlen() = '.strlen($status_message), ILogger::DEBUG);
                    // The NotificationManager throws an InvalidArgumentException if the 'Message' of a notification is longer than 64 characters
                    if (strlen($status_message) >= 64) {
                        $status_message = substr($status_message, 0, 63);
                        \OCP\Util::writeLog('signlive', 'Truncated $status_message to: '.$status_message, ILogger::DEBUG);
                    }
                    $notification->setMessage($status_message, $message_and_subject_parameters);

                    $confirm_action = $notification->createAction();
                    $confirm_action->setLabel('confirm');
                    
                    $confirm_url = $this->urlGenerator->getAbsoluteURL($this->urlGenerator->linkTo('', '/index.php/apps/signlive/sign_document/confirm_notification?job_uuid='.$object_id));
                    \OCP\Util::writeLog('signlive', 'confirm URL: '.$confirm_url, ILogger::DEBUG);
                    $confirm_action->setLink($confirm_url, 'POST');
                    $notification->addAction($confirm_action);

                    if (array_key_exists('signature_download_url', $signature_job_entry_data)) {
                        $download_signed_file_action = $notification->createAction();
                        $download_signed_file_action->setLabel('download_signed_file');
                        $link_to_signed_document = $signature_job_entry_data['signature_download_url']; //$this->urlGenerator->getAbsoluteURL($this->urlGenerator->linkTo('files', $signature_job_data['signature_job_file']));
                        
                        \OCP\Util::writeLog('signlive', 'Link to signed document: '.$link_to_signed_document, ILogger::DEBUG);
                        
                        $download_signed_file_action->setLink($link_to_signed_document, 'GET');
                        $notification->addAction($download_signed_file_action);
                    }

                    $this->notificationManager->notify($notification);
                } catch (\InvalidArgumentException $e) {
                    \OCP\Util::writeLog('signlive', 'sendSuccessNotification() failed with an InvalidArgumentException: '.$e, ILogger::ERROR);
                    return false;
                }
            } else {
                // Just dispatch a short-lived message box notification by means of a DataResponse to the web frontend
                // Persist the message to the database for pick-up, display and automated confirmation of display 
                // by the frontend side of the request job monitoring
                $job_success_message = '';
                if ($signature_job_data['signature_job_type'] === 'viewer_based_signature') {
                    $job_success_message = 'Viewer-based signature request job \''.$job_uuid.'\' has been completed successfully for file \''.$signature_job_data['file_name'].'\'.';
                } else if ($signature_job_data['signature_job_type'] === 'automatic_signature') {
                    $job_success_message = 'Automatic signature request job \''.$job_uuid.'\' has been completed successfully for file \''.$signature_job_data['file_name'].'\'.';
                }

                $this->setNewJobState($job_uuid, 'SIGNATURE_JOB_COMPLETED_SUCCESSFULLY', $job_success_message, false);
                
                return true;
            }
            return true;
        }
        
        return false;
    }

    /**
     * @NoAdminRequired
     * Send a notification for failed signature request via the Nextcloud notification API
     * @return bool
    */
    protected function sendFailureNotification($signature_job_data) {
        \OCP\Util::writeLog('signlive', '=====================================================================', ILogger::DEBUG);
        \OCP\Util::writeLog('signlive', 'sendFailureNotification() called in signlive/lib/Controller/IntarsysJobSignatureController.php', ILogger::DEBUG);
        
        foreach($signature_job_data as $data_key => $data_value) {
            \OCP\Util::writeLog('signlive', 'Job data attribute '.$data_key.' => '.$data_value, ILogger::DEBUG);
        }
        
        $job_uuid_found = array_key_exists("signature_job_uuid", $signature_job_data);
        $file_name_found = array_key_exists("file_name", $signature_job_data);
        $job_name_found = array_key_exists("signature_job_name", $signature_job_data);
        $profile_type_found = array_key_exists("signature_profile_type", $signature_job_data);
        $job_type_found = array_key_exists("signature_type", $signature_job_data);
        $user_found = array_key_exists("nextcloud_user", $signature_job_data);
        $status_message_found = array_key_exists("status_message", $signature_job_data);
        $job_stage_found = array_key_exists("job_stage", $signature_job_data);

        \OCP\Util::writeLog('signlive', 'job_uuid_found = '.$job_uuid_found.'; profile_type_found = '.$profile_type_found.'; file_name_found = '.$file_name_found.'; user_found = '.$user_found.'; job_type_found = '.$job_type_found, ILogger::DEBUG);
        $show_persistent_failure_notifications = $this->config->getAppValue('signlive', 'cloudsuite_gears_show_verbose_failure_notifications', SignLiveDefaultConfigValues::$SHOW_PERSISTENT_NOTIFICATIONS_FOR_FAILED_SIGNATURE_REQUESTS);
        \OCP\Util::writeLog('signlive', 'show_persistent_failure_notifications = '.$show_persistent_failure_notifications, ILogger::DEBUG);

        \OCP\Util::writeLog('signlive', '=====================================================================', ILogger::DEBUG);

        if ($job_uuid_found &&
            $file_name_found &&
            $user_found && 
            $job_type_found) {
            \OCP\Util::writeLog('signlive', 'sendFailureNotification() dispatching a new notification to the web UI.', ILogger::DEBUG);
            \OCP\Util::writeLog('signlive', 'signature_job_uuid = \''.$signature_job_data['signature_job_uuid'].'\'; signature_job_file = '.$signature_job_data['file_name'].'; nextcloud_user = '.$signature_job_data['nextcloud_user'].'; signature_type = '.$signature_job_data['signature_type'], ILogger::DEBUG);
        
            // Only show failure notifications via Notifications app if specifically requested in the SignLive settings
            if ($show_persistent_failure_notifications === 1 || $show_persistent_failure_notifications === "true") {
                \OCP\Util::writeLog('signlive', 'sendFailureNotification() sends a new persistent notification via Notification app.', ILogger::DEBUG);

                $object_id = $signature_job_data['signature_job_uuid'];
                $notification = $this->notificationManager->createNotification();
                try {
                    $notification->setApp('signlive');
                    $notification->setDateTime(new \DateTime());

                    \OCP\Util::writeLog('signlive', '$object_id strlen() = '.strlen($object_id), ILogger::DEBUG);

                    // The NotificationManager throws an InvalidArgumentException if the 'Object' of a notification is longer than 64 characters
                    if (strlen($object_id) >= 64) {
                        $object_id = substr($object_id, 0, 63);
                        \OCP\Util::writeLog('signlive', 'Truncated $object_id to: '.$object_id, ILogger::DEBUG);
                    }

                    $notification->setObject('signlive_job', $object_id);
                    $notification->setUser($signature_job_data['nextcloud_user']);

                    $message_and_subject_parameters = array(
                        'file_name' => $signature_job_data['file_name']
                    );
                    if ($signature_job_data['signature_type'] === 'viewer_based_signature') {
                        $notification->setSubject('viewer_based_signature_failure', $message_and_subject_parameters);
                    } else if ($signature_job_data['signature_type'] === 'automatic_signature') {
                        $notification->setSubject('automatic_signature_failure', $message_and_subject_parameters);
                    }

                    $status_message = 'Signature job '.$object_id.' failed!';
                    if ($status_message_found) {
                        $status_message = $signature_job_data['status_message'];
                        \OCP\Util::writeLog('signlive', '$status_message strlen() = '.strlen($status_message), ILogger::DEBUG);
                        // The NotificationManager throws an InvalidArgumentException if the 'Message' of a notification is longer than 64 characters
                        if (strlen($status_message) >= 64) {
                            $status_message = substr($status_message, 0, 63);
                            \OCP\Util::writeLog('signlive', 'Truncated $status_message to: '.$status_message, ILogger::DEBUG);
                        }
                    }

                    $notification->setMessage($status_message, $message_and_subject_parameters);

                    $confirm_action = $notification->createAction();
                    $confirm_action->setLabel('submit_bug_report');
                    
                    $submit_bug_report = $this->urlGenerator->getAbsoluteURL($this->urlGenerator->linkTo('', '/index.php/apps/signlive/submit_bug_report?job_uuid='.$object_id));
                    \OCP\Util::writeLog('signlive', 'submit_bug_report URL: '.$submit_bug_report, ILogger::DEBUG);
                    $confirm_action->setLink($submit_bug_report, 'POST');
                    $notification->addAction($confirm_action);

                    $this->notificationManager->notify($notification);
                } catch (\InvalidArgumentException $e) {
                    \OCP\Util::writeLog('signlive', 'sendFailureNotification() failed with an InvalidArgumentException: '.$e, ILogger::WARN);
                    return false;
                }
            } else {
                \OCP\Util::writeLog('signlive', 'sendFailureNotification() will NOT send a new persistent notification via Notification app.', ILogger::DEBUG);
            }

            // Send a temporarily visible failure notification in any case
            // Dispatch a short-lived message box notification by means of a DataResponse to the web frontend
            // Persist the message to the database for pick-up, display and automated confirmation of display 
            // by the frontend side of the request job monitoring
            $job_failure_message = '';
            if ($signature_job_data['signature_type'] === 'viewer_based_signature') {
                $job_failure_message = 'Viewer-based signature request job \''.$signature_job_data['signature_job_uuid'].'\' has failed for file \''.$signature_job_data['file_name'].'\'!';
                if ($status_message_found) {
                    $job_failure_message = $job_failure_message.'; Status message: '.$signature_job_data['status_message'];
                }
                if ($job_stage_found) {
                    $job_failure_message = $job_failure_message.'; Job stage: '.$signature_job_data['job_stage'];
                }
            } else if ($signature_job_data['signature_type'] === 'automatic_signature') {
                $job_failure_message = 'Automatic signature request job \''.$signature_job_data['signature_job_uuid'].'\' has failed for file \''.$signature_job_data['file_name'].'\'!';
                if ($status_message_found) {
                    $job_failure_message = $job_failure_message.'; Status message: '.$signature_job_data['status_message'];
                }
                if ($job_stage_found) {
                    $job_failure_message = $job_failure_message.'; Job stage: '.$signature_job_data['job_stage'];
                }
            }

            $last_job_state = $this->getJobState($signature_job_data['signature_job_uuid']);
            // Update status message but keep job state as-is
            $this->setNewJobState($signature_job_data['signature_job_uuid'], $last_job_state, $job_failure_message, true);
            
            return true;
        }
        return false;
    }

    /**
     * @NoAdminRequired
     * Create persistent record for signature request job
     * @return string
     */
    protected function createJobDefinitionEntry($signature_job_data) {
        \OCP\Util::writeLog('signlive', 'createJobDefinitionEntry() called in signlive/lib/Controller/IntarsysJobSignatureController.php.', ILogger::DEBUG);
        \OCP\Util::writeLog('signlive', '$signature_job_data = '.print_r($signature_job_data, TRUE), ILogger::DEBUG);

        $uuid_found = 0;
        $uuid_value = '';
        if (array_key_exists('signature_job_uuid', $signature_job_data)) {
            $uuid_found = 1;
            $uuid_value = $signature_job_data['signature_job_uuid'];
        }

        $user_found = 0;
        $user_value = '';
        if (array_key_exists('nextcloud_user', $signature_job_data)) {
            $user_found = 1;
            $user_value = $signature_job_data['nextcloud_user'];
        }

        $creation_time_found = 0;
        $creation_time_value = '';
        if (array_key_exists('job_creation_time', $signature_job_data)) {
            $creation_time_found = 1;
            $creation_time_value = $signature_job_data['job_creation_time'];
        }

        $file_name_found = 0;
        $file_name_value = '';
        if (array_key_exists('file_name', $signature_job_data) ||
            array_key_exists('signature_job_file', $signature_job_data)) {
            $file_name_found = 1;
            if (array_key_exists('file_name', $signature_job_data)) {
                $file_name_value = $signature_job_data['file_name'];
            }
            if (array_key_exists('signature_job_file', $signature_job_data)) {
                $file_name_value = $signature_job_data['signature_job_file'];
            }
        }

        $signature_type_found = 0;
        $signature_type_value = '';
        if (array_key_exists('signature_type', $signature_job_data) ||
            array_key_exists('signature_job_type', $signature_job_data)) {
            $signature_type_found = 1;
            if (array_key_exists('signature_type', $signature_job_data)) {
                $signature_type_value = $signature_job_data['signature_type'];
            }
            if (array_key_exists('signature_job_type', $signature_job_data)) {
                $signature_type_value = $signature_job_data['signature_job_type'];
            }
        }

        $download_url_found = 0;
        $download_url_value = '';
        if (array_key_exists('download_url', $signature_job_data)) {
            $download_url_found = 1;
            $download_url_value = $signature_job_data['download_url'];
        }

        $job_stage_found = 0;
        $job_stage_value = '';
        if (array_key_exists('job_stage', $signature_job_data)) {
            $job_stage_found = 1;
            $job_stage_value = $signature_job_data['job_stage'];
        }

        $original_folder_uri_found = 0;
        $original_folder_uri_value = '';
        if (array_key_exists('original_folder_uri', $signature_job_data)) {
            $original_folder_uri_found = 1;
            $original_folder_uri_value = $signature_job_data['original_folder_uri'];
        }

        \OCP\Util::writeLog('signlive', '$uuid_found = '.$uuid_found.', $user_found = '.$user_found.', $creation_time_found = '.$creation_time_found.', $file_name_found = '.$file_name_found.', $signature_type_found = '.$signature_type_found, ILogger::DEBUG);

        if ($uuid_found &&
            $user_found &&
            $creation_time_found &&
            $file_name_found) {
            \OCP\Util::writeLog('signlive', 'All required attributes for a signature job entry found, creating persistent record.', ILogger::DEBUG);

            $job_record_key = 'csgi_job_def_'.$uuid_value;
            $job_record_value = 'file_name::'.$file_name_value.';;';
            $job_record_value = $job_record_value.'nextcloud_user::'.$user_value.';;';
            $job_record_value = $job_record_value.'job_creation_time::'.$creation_time_value;
            
            // Put optional attributes last
            if ($signature_type_found) {
                $job_record_value = $job_record_value.';;signature_type::'.$signature_type_value;
            }
            if ($job_stage_found) {
                $job_record_value = $job_record_value.';;job_stage::'.$job_stage_value;
            }
            if ($download_url_found) {
                $job_record_value = $job_record_value.';;download_url::'.$download_url_value;
            }
            if ($original_folder_uri_found) {
                $job_record_value = $job_record_value.';;original_folder_uri::'.$original_folder_uri_value;
            }

            $this->config->setAppValue('signlive', $job_record_key, $job_record_value);

            return $signature_job_data['signature_job_uuid'];
        }
        
        // Warning is only really relevant for developers, so report on DEBUG level
        \OCP\Util::writeLog('signlive', 'Not all required attributes for a signature job entry were provided, not creating persistent record.', ILogger::DEBUG);
        \OCP\Util::writeLog('signlive', 'uuid_found = .'.$uuid_found.'; user_found = '.$user_found.'; file_name_found = '.$file_name_found.'; creation_time_found = '.$creation_time_found, ILogger::DEBUG);
        return '';
    }

    /**
     * @NoAdminRequired
     * Get a specific signature request job record
     * @return array
     */
    protected function getJobDefinitionEntry($signature_job_uuid) {
        \OCP\Util::writeLog('signlive', 'getJobDefinitionEntry() called in signlive/lib/Controller/IntarsysJobSignatureController.php. job_uuid = '.$signature_job_uuid, ILogger::DEBUG);
    
        $job_definition_entry_data = [];

        if (empty($signature_job_uuid) || $signature_job_uuid === '') {
            \OCP\Util::writeLog('signlive', 'getJobDefinitionEntry() called with an empty job_uuid!', ILogger::WARN);
            return $job_definition_entry_data;
        }

        $job_record_key = 'csgi_job_def_'.$signature_job_uuid;
        $job_record_value = $this->config->getAppValue('signlive', $job_record_key, '');

        if (empty($job_record_value) || $job_record_value === '') {
            \OCP\Util::writeLog('signlive', 'getJobDefinitionEntry() found no signature request job entry for job_uuid = '.$signature_job_uuid, ILogger::DEBUG);
            return $job_definition_entry_data;
        }

        $job_record_components = explode(';;', $job_record_value);
        foreach ($job_record_components as $record_component) {
            $component_name_and_value = explode('::', $record_component);
            if (count($component_name_and_value) == 2) {
                $component_name = $component_name_and_value[0];
                $component_value = $component_name_and_value[1];
                $job_definition_entry_data[$component_name] = $component_value;
            }
        }
        
        \OCP\Util::writeLog('signlive', 'getJobDefinitionEntry() return value: '.print_r($job_definition_entry_data, TRUE), ILogger::DEBUG);

        return $job_definition_entry_data;
    }

    /**
     * @NoAdminRequired
     * Remove a specific signature job record
     * @return bool
     */
    protected function removeJobDefinitionEntry($signature_job_uuid) {
        \OCP\Util::writeLog('signlive', 'removeJobDefinitionEntry() called in signlive/lib/Controller/IntarsysJobSignatureController.php. job_uuid = '.$signature_job_uuid, ILogger::DEBUG);
        if (empty($signature_job_uuid) || $signature_job_uuid === '') {
            \OCP\Util::writeLog('signlive', 'removeJobDefinitionEntry() called with an empty job_uuid!', ILogger::WARN);
            return false;
        }

        $job_record_key = 'csgi_job_def_'.$signature_job_uuid;
        $job_record_value = $this->config->getAppValue('signlive', $job_record_key, '');

        if (empty($job_record_value) || $job_record_value === '') {
            \OCP\Util::writeLog('signlive', 'removeJobDefinitionEntry() found no signature request job entry for job_uuid = '.$signature_job_uuid, ILogger::WARN);
            return false;
        } else {
            $this->config->deleteAppValue('signlive', $job_record_key);
            \OCP\Util::writeLog('signlive', 'removeJobDefinitionEntry() deleted signature job entry for job_uuid = '.$signature_job_uuid, ILogger::DEBUG);
            
            return true;
        }

        return false;
    }

    /**
     * @NoAdminRequired
     * Set an existing signature job record to a new state
     * @return bool
     */
    protected function setNewJobState(string $job_uuid, string $job_stage, string $status_message, $mark_as_failed = false) {
        \OCP\Util::writeLog('signlive', 'setNewJobState() called in signlive/lib/Controller/IntarsysJobSignatureController.php. $job_uuid = \''.$job_uuid.'\', $job_state = \''.$job_stage.'\'', ILogger::DEBUG);
    
        if (empty($job_uuid) || $job_uuid === '') {
            \OCP\Util::writeLog('signlive', 'setNewJobState() called with an empty job_uuid value, can not continue!', ILogger::WARN);
            return false;
        }

        $signature_job_data = $this->getJobDefinitionEntry($job_uuid);
        
        $user_found = 0;
        $user_value = '';
        if (array_key_exists('nextcloud_user', $signature_job_data)) {
            $user_found = 1;
            $user_value = $signature_job_data['nextcloud_user'];
        }

        $creation_time_found = 0;
        $creation_time_value = '';
        if (array_key_exists('job_creation_time', $signature_job_data)) {
            $creation_time_found = 1;
            $creation_time_value = $signature_job_data['job_creation_time'];
        }

        $file_name_found = 0;
        $file_name_value = '';
        if (array_key_exists('file_name', $signature_job_data) ||
            array_key_exists('signature_job_file', $signature_job_data)) {
            $file_name_found = 1;
            if (array_key_exists('file_name', $signature_job_data)) {
                $file_name_value = $signature_job_data['file_name'];
            }
            if (array_key_exists('signature_job_file', $signature_job_data)) {
                $file_name_value = $signature_job_data['signature_job_file'];
            }
        }

        $signature_type_found = 0;
        $signature_type_value = '';
        if (array_key_exists('signature_type', $signature_job_data) ||
            array_key_exists('signature_job_type', $signature_job_data)) {
            $signature_type_found = 1;
            if (array_key_exists('signature_type', $signature_job_data)) {
                $signature_type_value = $signature_job_data['signature_type'];
            }
            if (array_key_exists('signature_job_type', $signature_job_data)) {
                $signature_type_value = $signature_job_data['signature_job_type'];
            }
        }

        $download_url_found = 0;
        $download_url_value = '';
        if (array_key_exists('download_url', $signature_job_data)) {
            $download_url_found = 1;
            $download_url_value = $signature_job_data['download_url'];
        }

        $original_folder_uri_found = 0;
        $original_folder_uri_value = '';
        if (array_key_exists('original_folder_uri', $signature_job_data)) {
            $original_folder_uri_found = 1;
            $original_folder_uri_value = $signature_job_data['original_folder_uri'];
        }

        $job_stage_found = 0;
        $job_stage_value = '';
        if (array_key_exists('job_stage', $signature_job_data)) {
            $job_stage_found = 1;
            $job_stage_value = $signature_job_data['job_stage'];
        }

        if ($user_found &&
            $creation_time_found &&
            $file_name_found &&
            $signature_type_found) {
            
            \OCP\Util::writeLog('signlive', 'setNewJobState(): All required attributes are present in signature job entry \''.$job_uuid.'\', can continue.', ILogger::DEBUG);

            // Reassemble the settings string with all attributes but the new job state unchanged.
            $job_record_key = 'csgi_job_def_'.$job_uuid;
            $job_record_value = 'file_name::'.$file_name_value.';;';
            $job_record_value = $job_record_value.'nextcloud_user::'.$user_value.';;';
            $job_record_value = $job_record_value.'signature_type::'.$signature_type_value.';;';
            $job_record_value = $job_record_value.'job_creation_time::'.$creation_time_value.';;';
            $job_record_value = $job_record_value.'job_stage::'.$job_stage.';;';
            $job_record_value = $job_record_value.'status_message::'.$status_message;
            
            \OCP\Util::writeLog('signlive', 'setNewJobState(): mark_as_failed =  \''.$mark_as_failed.'\'', ILogger::DEBUG);
            if ($mark_as_failed === true) {
                $job_record_value = $job_record_value.';;job_failed::1';
            } else {
                $job_record_value = $job_record_value.';;job_failed::0';
            }
            
            // Put optional attributes last
            if ($signature_type_found) {
                $job_record_value = $job_record_value.';;signature_type::'.$signature_type_value;
            }
            if ($download_url_found) {
                $job_record_value = $job_record_value.';;download_url::'.$download_url_value;
            }
            if ($original_folder_uri_found) {
                $job_record_value = $job_record_value.';;original_folder_uri::'.$original_folder_uri_value;
            }

            $this->config->setAppValue('signlive', $job_record_key, $job_record_value);

            \OCP\Util::writeLog('signlive', 'setNewJobState(): Signature job entry \''.$job_uuid.'\' updated to new job state \''.$job_stage.'\'. Record is now: \''.$job_record_value.'\'', ILogger::DEBUG);

            return true;
        } else  {
            \OCP\Util::writeLog('signlive', 'setNewJobState(): Not all required attributes are present in signature job entry \''.$job_uuid.'\', can not continue!', ILogger::DEBUG);
            \OCP\Util::writeLog('signlive', 'setNewJobState(): user_found = \''.$user_found.'\', creation_time_found = \''.$creation_time_found.'\', file_name_found = \''.$file_name_found.'\', signature_type_found = \''.$signature_type_found.'\', job_stage_found = \''.$job_stage_found.'\'', ILogger::DEBUG);
            return false;
        }
    }

    /**
     * @NoAdminRequired
     * Return current state for a signature request job
     * @param string
     */
    protected function getJobState(string $job_uuid) {
        \OCP\Util::writeLog('signlive', 'getJobState() called in signlive/lib/Controller/IntarsysJobSignatureController.php. $job_uuid = \''.$job_uuid.'\'', ILogger::DEBUG);
    
        if (empty($job_uuid) || $job_uuid === '') {
            \OCP\Util::writeLog('signlive', 'getJobState() called with an empty job_uuid value, can not continue!', ILogger::WARN);
            return '';
        }

        $signature_job_data = $this->getJobDefinitionEntry($job_uuid);
        $job_stage_found = 0;
        if (array_key_exists('job_stage', $signature_job_data)) {
            $job_stage_found = 1;
        }

        if ($job_stage_found) {
            return $signature_job_data['job_stage'];
        }

        return '';
    }

    /**
	 * @NoAdminRequired
     * Confirm a signature request notification
     * For failure notifications: Confirmed explicitly by user by button click in the Notification app
     * For success notifications: Confirmed automatically after display in the Nextcloud web UI
	 */
	public function confirmSignatureNotification() {
        \OCP\Util::writeLog('signlive', '************************************************************************************************************************', ILogger::DEBUG);
        \OCP\Util::writeLog('signlive', 'confirmSignatureNotification() called in signlive/lib/Controller/IntarsysJobSignatureController.php.', ILogger::DEBUG);
        \OCP\Util::writeLog('signlive', '************************************************************************************************************************', ILogger::DEBUG);
        
        $request_parameters = $this->request->getParams();
        $job_uuid = '';
        $ack_reason = '';
        foreach ($request_parameters as $parameter_name => $parameter_value) {
            \OCP\Util::writeLog('signlive', 'Parameter \''.$parameter_name.'\' => \''.$parameter_value.'\'', ILogger::DEBUG);
            if ($parameter_name === 'job_uuid') {
                $job_uuid = $parameter_value;
            }
            if ($parameter_name === 'ack_reason') {
                $ack_reason = $parameter_value;
            }
        }

        if ($job_uuid === '') {
            \OCP\Util::writeLog('signlive', 'confirmSignatureNotification() called without parameter \'job_uuid\', can not proceed!', ILogger::WARN);
        
            return(new DataResponse([
                'message' => 'confirmSignatureNotification() called without parameter \'job_uuid\', can not proceed!',
                'success' => false
            ]));
        } else {
            \OCP\Util::writeLog('signlive', 'Notification confirmed for signature request job: $job_uuid = '.$job_uuid, ILogger::DEBUG);

            $job_definition_removed = $this->removeJobDefinitionEntry($job_uuid);

            $status_message = 'Notification confirmed for signature request job: '.$job_uuid;

            if ($job_definition_removed === true) {
                if ($ack_reason === 'JOB_EXPIRED') {
                    $status_message = 'Removed expired signature job status for job_uuid \''.$job_uuid.'\'';
                }
                return(new DataResponse([
                    'message' => $status_message,
                    'success' => true
                ]));
            } else {
                $status_message = 'Notification NOT confirmed for signature request job: '.$job_uuid.'! Failed to remove signature request job entry.';
                return(new DataResponse([
                    'message' => $status_message,
                    'success' => false
                ]));
            }
        }
    }

    /**
     * @NoAdminRequired
     * @NoCSRFRequired
     * Get a list of all persisted signature job definitions
     * Returns an array of signature request job entries indexed by job UUIDs filtered by the current Nextcloud user
     * @return list
     */
    public function getJobDefinitionEntries() {
        $app_config_keys = $this->config->getAppKeys('signlive');

        $current_user = $this->userSession->getUser()->getUID();

        $signature_job_records = [];
        foreach ($app_config_keys as $app_key_id => $app_key) {
			if (str_starts_with($app_key, 'csgi_job_def_')) {
                $job_uuid = str_replace('csgi_job_def_', '', $app_key);

                $signature_job_record = $this->getJobDefinitionEntry($job_uuid);
                if ($signature_job_record['nextcloud_user'] === $current_user) {
                    $signature_job_records[$job_uuid] = $signature_job_record;
                }
            }
        }

        // Add the temporary notification display timeout and which signature requests should be reported to the user
        $signature_job_records['cloudsuite_gears_notification_duration'] = $this->config->getAppValue('signlive', 'cloudsuite_gears_notification_duration', SignLiveDefaultConfigValues::$NOTIFICATION_DISPLAY_DURATION);
        $signature_job_records['cloudsuite_gears_show_success_notifications'] = $this->config->getAppValue('signlive', 'cloudsuite_gears_show_success_notifications', SignLiveDefaultConfigValues::$SHOW_NOTIFICATIONS_FOR_SUCCESSFUL_SIGNATURE_REQUESTS);
        $signature_job_records['cloudsuite_gears_show_verbose_failure_notifications'] = $this->config->getAppValue('signlive', 'cloudsuite_gears_show_verbose_failure_notifications', SignLiveDefaultConfigValues::$SHOW_PERSISTENT_NOTIFICATIONS_FOR_FAILED_SIGNATURE_REQUESTS);
        $signature_job_records['cloudsuite_gears_notification_check_interval'] = $this->config->getAppValue('signlive', 'cloudsuite_gears_notification_check_interval', SignLiveDefaultConfigValues::$NOTIFICATION_CHECK_INTERVAL);

        return $signature_job_records;
    }

    /**
     * @NoAdminRequired
     * Log a signature request to the Nextcloud log file
     * Logged with omitted base64-encoded portions for brevity
     * @param signature_job_uuid UUID of signature request job
     * @param signature_job_data array with signature request job attributes
     * @param signature_json_request JSON sent to Gears backend server
     */
    protected function logSignatureRequest($signature_job_uuid, $signature_job_data, $signature_json_request) {
        $log_message = 'Initial JSON request for signature job \''.$signature_job_uuid.' -- ';
        if (array_key_exists('nextcloud_user', $signature_job_data)) {
            $log_message = $log_message.' Nextcloud user: '.$signature_job_data['nextcloud_user'].' --';
        }
        if (array_key_exists('file_name', $signature_job_data)) {
            $log_message = $log_message.' File name: '.$signature_job_data['file_name'].' --';
        }
        if (array_key_exists('signature_type', $signature_job_data)) {
            $log_message = $log_message.'Signature profile type: '.$signature_job_data['signature_type'].' --';
        }
        $log_message = $log_message.' JSON section with replaced variable values (without base64-encoded content): '.$signature_json_request;

        \OCP\Util::writeLog('signlive', $log_message, ILogger::INFO);
    }

     /**
     * @NoAdminRequired
     * Log a signature job result to the Nextcloud log file
     * Logged with omitted base64-encoded portions for brevity
     * @param signature_job_uuid UUID of signature request job
     * @param signature_job_data array with signature request job attributes
     */
    protected function logSignatureJobResult($signature_job_uuid, $signature_job_data) {
        $log_message = 'Result for signature job \''.$signature_job_uuid.'\': ';
        if (array_key_exists('job_failed', $signature_job_data)) {
            if (intval($signature_job_data['job_failed']) == 1) {
                $log_message = $log_message.'FAILURE!';
            } else {
                $log_message = $log_message.'SUCCESS.';
            }
        }
    }

    /**
	 * @NoAdminRequired
	 * Sign a document using the Gears web viewer
	 * @return DataResponse
	 */
    public function signDocumentViaViewerJob() {
        \OCP\Util::writeLog('signlive', 'signDocumentViaViewerJob() called in signlive/lib/Controller/IntarsysJobSignatureController.php.', ILogger::DEBUG);

        $request_parameters = $this->request->getParams();
        $file_name = '';
        $redirect_url = '';
        $profile_type = '';
        $signature_type = '';
        $profile_index = -1;
        $profile_unique_id = '';
        $profile_identifier = '';
        $job_uuid = '';
        $download_url = '';
        $original_folder_uri = '';
        foreach ($request_parameters as $parameter_name => $parameter_value) {
            \OCP\Util::writeLog('signlive', 'Parameter \''.$parameter_name.'\' => \''.$parameter_value.'\'', ILogger::DEBUG);
            if ($parameter_name == 'filename') {
                $file_name = $parameter_value;
            }
            if ($parameter_name === 'profile_unique_id') {
                $profile_unique_id = $parameter_value;
            }
            if ($parameter_name === 'redirect_uri') {
                $redirect_url = $parameter_value;
            }
            if ($parameter_name === 'profile_type') {
                $profile_type = $parameter_value;
            }
            if ($parameter_name === 'signature_type') {
                $signature_type = $parameter_value;
            }
            if ($parameter_name === 'profile_index') {
                $profile_index = $parameter_value;
            }
            if ($parameter_name == 'profile_identifier') {
                $profile_identifier = $parameter_value;
            }
            if ($parameter_name === 'job_uuid') {
                $job_uuid = $parameter_value;
            }
            if ($parameter_name === 'download_url') {
                $download_url = $parameter_value;
            }
            if ($parameter_name === 'original_folder_uri') {
                $original_folder_uri = $parameter_value;
            }
        }
        
        // Verify all required signature request parameters have actually been provided
        $parameters_valid = true;
        $parameter_error_msg = '';
        if (empty($file_name) || $file_name === '') {
            $parameter_error_msg += 'No file name was provided.';
            $parameters_valid = false;
        }
        if (empty($profile_unique_id) || $profile_unique_id === '') {
            $parameter_error_msg = $parameter_error_msg.' -- No valid signature profile unique ID was provided.';
            $parameters_valid = false;
        }
        if (empty($redirect_url) || $redirect_url === '') {
            $parameter_error_msg = $parameter_error_msg.' -- No valid redirect URL was provided.';
            $parameters_valid = false;
        }
        if (empty($profile_type) || $profile_type === '') {
            $parameter_error_msg = $parameter_error_msg.' -- No valid signature profile type was provided.';
            $parameters_valid = false;
        }
        if (empty($job_uuid) || $job_uuid === '') {
            $parameter_error_msg = $parameter_error_msg.' -- No job UUID was provided.';
            $parameters_valid = false;
        }
        if (empty($signature_type) || $signature_type === '') {
            $parameter_error_msg += ' -- No signature type was provided.';
            $signature_type = false;
        }

        // If not all required parameters are present: Return early
        if ($parameters_valid === false) {
            \OCP\Util::writeLog('signlive', 'Received invalid viewer-based signature request query, not all required parameters were provided: '.$parameter_error_msg, ILogger::WARN);
            
            // No signature request job state exists at this point: Just send a failure notification if there is a valid job_uuid
            // Provide what is possible in job record data and let sendFailureNotification take care of assembling a notification
            if (!(empty($job_uuid) || $job_uuid === '')) {
                $job_record_data = [];
                $job_record_data['signature_job_uuid'] = $job_uuid;
                if (!(empty($file_name) || $file_name === '')) {
                    $job_record_data['file_name'] = $file_name;
                }
                if (!(empty($signature_type) || $signature_type === '')) {
                    $job_record_data['signature_type'] = $signature_type;
                }
                $job_record_data['nextcloud_user'] = $this->userSession->getUser()->getUID();
                $job_record_data['status_message'] = 'Received invalid viewer-based signature request query, not all required parameters were provided: '.$parameter_error_msg;
                
                $this->sendFailureNotification($job_record_data);
            }
            return new DataResponse([
                'message' => 'Not all required parameters for a viewer-based signature request were provided, can not continue: '.$parameter_error_msg,
                'viewer_url' => '',
                'signature_type' => 'viewer_based_signature',
                'success' => false
            ]);
        }

        // TODO: Enable reading files from shared folders, i. e. not only from the storage of the currently logged on Nextcloud user
        $file_list_to_sign = $this->getFilesToSign();

        if (count($file_list_to_sign) == 1) {
            \OCP\Util::writeLog('signlive', 'Using signature profile index '.$profile_index.' of type \''.$profile_type.'\'', ILogger::DEBUG);
            \OCP\Util::writeLog('signlive', 'Profile identifier: '.$profile_identifier, ILogger::DEBUG);
            \OCP\Util::writeLog('signlive', 'Signature type: '.$signature_type, ILogger::DEBUG);
            $user_profile_json_section = '';
            $profile_shortened_unique_id = str_replace('cloudsuite_gears_profile_id_', '', $profile_unique_id);
            $profile_json_section_key = 'cloudsuite_gears_profile_json_'.$profile_shortened_unique_id;
            \OCP\Util::writeLog('signlive', 'Query JSON section for profile '.$profile_index.' using key: '.$profile_json_section_key, ILogger::DEBUG);
            
            $profile_json_section = $this->config->getAppValue('signlive', $profile_json_section_key, '');

            \OCP\Util::writeLog('signlive', 'Using JSON section for viewer-based signature: '.$user_profile_json_section, ILogger::DEBUG);
                
            if ($profile_json_section != '') {
                $base64_encoded_document = $this->fileIO->getBase64EncodedFile($file_list_to_sign[0]);

                if ($base64_encoded_document != '') {
                    $profile_data = [
                        'file_name' => $file_name,
                        'redirect_url' => $redirect_url,
                        'json_section' => $profile_json_section,
                        'signature_type' => $signature_type, 
                        'base64_encoded_document' => $base64_encoded_document
                    ];

                    \OCP\Util::writeLog('signlive', '---------------------------------------------------------------------------------------------------', ILogger::DEBUG);
                    \OCP\Util::writeLog('signlive', 'Redirect URL for viewer-based signature request job: '.$redirect_url, ILogger::DEBUG);
                    \OCP\Util::writeLog('signlive', '---------------------------------------------------------------------------------------------------', ILogger::DEBUG);

                    // Get the JSON request for the document to sign
                    $json_for_signature_request = $this->jsonQueryCreator->getInitialViewerJSON($file_list_to_sign[0], $profile_data, false);
                    if ($json_for_signature_request != '') {
                        $json_for_signature_request_for_logging = $this->jsonQueryCreator->getInitialViewerJSON($file_list_to_sign[0], $profile_data, true);
                    
                        // Send the JSON request to the Gears backend, catching any connection-related errors that may occur
                        try {
                            // At this point a valid JSON request for the Gears backend is present: Create a signature request job record here.
                            $signature_job_data = [];
                            $signature_job_data['signature_job_uuid'] = $job_uuid;
                            $signature_job_data['signature_job_name'] = 'Signature job \''.$file_name.'\''; 
                            $signature_job_data['signature_job_type'] = 'viewer_based_signature';
                            $signature_job_data['signature_job_file'] = $file_name;
                            $signature_job_data['signature_profile_type'] = $profile_type;
                            $signature_job_data['nextcloud_user'] = $this->userSession->getUser()->getUid();
                            $signature_job_data['job_creation_time'] = time();
                            $signature_job_data['original_folder_uri'] = $original_folder_uri;

                            // Set new signature request job to initial state at creation time
                            $signature_job_data['job_stage'] = 'STARTED';

                            $this->logSignatureRequest($job_uuid, $signature_job_data, $json_for_signature_request_for_logging);
         
                            // Persist this signature request for job management and notifications
                            $job_def_created = $this->createJobDefinitionEntry($signature_job_data);

                            if ($job_def_created === false) {
                                return new DataResponse([
                                    'message' => 'Failed create signature request job entry! Signature request could not be processed successfully.',
                                    'viewer_url' => '',  
                                    'signature_type' => 'viewer_based_signature',
                                    'success' => false
                                ]);
                            }

                            \OCP\Util::writeLog('signlive', 'Using Gears backend URL: '.$this->gears_backend_url, ILogger::DEBUG);
                            $gears_response = $this->gears_rest_client->post($this->gears_backend_url.'/api/v1/flow/viewer/create', [
                                'headers' => ['Content-Type' => 'application/json', 'Accept' => 'application/json'],
                                'body'     => $json_for_signature_request
                            ]);

                            // Request to create viewer flow has been sent, update job state
                            $this->setNewJobState($job_uuid, 'FLOW_CREATION_REQUEST_SENT', 'Initial request for viewer flow creation sent for job '.$job_uuid.'.', false);
                            
                            $gears_response_code = $gears_response->getStatusCode();
                            $gears_response_reason = $gears_response->getReasonPhrase();

                            \OCP\Util::writeLog('signlive', 'HTTP response from Gears backend: '.intval($gears_response_code).' ('.$gears_response_reason.')', ILogger::DEBUG);

                            // Response for viewer flow creation has been received, update job state
                            $this->setNewJobState($job_uuid, 'FLOW_CREATION_RESPONSE_RECEIVED', 'Viewer flow has been created for job \''.$job_uuid.'\'.', false);

                            $json_response_array = json_decode($gears_response->getBody(), true);
                            
                            \OCP\Util::writeLog('signlive', 'Decoded JSON response (shortened) from Gears backend: '.print_r($json_response_array, TRUE), ILogger::DEBUG);

                            // We expect a 2xx HTTP response code to a successful signature flow creation
                            if (intval($gears_response_code) >= 200 && intval($gears_response_code) <= 299) {
                                // Now parse the reponse and see if there's a redirect URL we need to call, or if there is an error.
                                $viewer_redirect_url = '';
                                $scheme_element = '';
                                $gears_response_ok = false;
                                if (array_key_exists('snapshot', $json_response_array)) {
                                    \OCP\Util::writeLog('signlive', 'Found \'snapshot\' element in JSON response.', ILogger::DEBUG);
                                    if (array_key_exists('stage', $json_response_array['snapshot'])) {
                                        \OCP\Util::writeLog('signlive', 'Found \'stage\' element in JSON response.', ILogger::DEBUG);
                                        if (array_key_exists('scheme', $json_response_array['snapshot']['stage'])) {
                                            $scheme_element = $json_response_array['snapshot']['stage']['scheme'];
                                        }
                                        \OCP\Util::writeLog('signlive', '\'scheme\' element in JSON response: '.$scheme_element, ILogger::DEBUG);
                                        if ($scheme_element != '') {
                                            $scheme_element_parts = explode(':', $scheme_element);
                                            $scheme_stage = '';
                                            if (count($scheme_element_parts) > 0) {
                                                $scheme_stage = $scheme_element_parts[count($scheme_element_parts)-1];
                                            }
                                            \OCP\Util::writeLog('signlive', '\'Stage\' in JSON response: '.$scheme_stage, ILogger::DEBUG);
                                            if ($scheme_stage == 'HttpRedirect') {
                                                \OCP\Util::writeLog('signlive', 'Got expected HttpRedirect stage in response from Gears backend.', ILogger::DEBUG);
                                                if (array_key_exists('url', $json_response_array['snapshot']['stage'])) {                                            
                                                    \OCP\Util::writeLog('signlive', 'Found \'url\' element in JSON response: '.$json_response_array['snapshot']['stage']['url'], ILogger::DEBUG);
                                                    $viewer_redirect_url = $json_response_array['snapshot']['stage']['url'];
                                                    $gears_response_ok = true;
                                                }
                                            }
                                        }
                                    }
                                }

                                if ($gears_response_ok) {
                                    \OCP\Util::writeLog('signlive', 'Received valid response for viewer-based signature flow for document: '.$file_list_to_sign[0], ILogger::INFO);
                                    
                                    // Valid redirect URL for viewer flow has been provided, update job state
                                    $this->setNewJobState($job_uuid, 'FLOW_CREATION_RESPONSE_VALID', 'Valid redirect URL for viewer flow has been provided for job \''.$job_uuid.'\'.', false);

                                    return new DataResponse([
                                        'message' => 'Received valid response for viewer-based signature flow for document: '.$file_list_to_sign[0],
                                        'success' => true,
                                        'viewer_url' => $viewer_redirect_url,
                                        'signature_type' => 'viewer_based_signature'
                                    ]);
                                } else {
                                    $status_message = 'Invalid response has been received for viewer flow creation for signature request job '.$job_uuid.', document: '.$file_list_to_sign[0].'!\n';
                                    $status_message = $status_message.'HTTP status code: '.$gears_response_code.'; Response from Gears backend: '.$gears_response->getBody();
                                    \OCP\Util::writeLog('signlive', $status_message, ILogger::ERROR);
                                    
                                    // Invalid response received for viewer flow creation, update job state
                                    $this->setNewJobState($job_uuid, 'FLOW_CREATION_RESPONSE_INVALID', $status_message, true);
                                    
                                    $signature_job_data['status_message'] = $status_message;
                                    $this->sendFailureNotification($signature_job_data);

                                    return new DataResponse([
                                        'message' => 'Received INVALID response for viewer-based signature flow for document: '.$file_list_to_sign[0],
                                        'viewer_url' => '',
                                        'signature_type' => 'viewer_based_signature',
                                        'success' => false
                                    ]);
                                }
                            } else {
                                \OCP\Util::writeLog('signlive', 'Received an error response from Gears backend: '.$gears_response->getStatusCode().' ('.$gears_response->getReasonPhrase().')', ILogger::ERROR);
                                
                                $job_record_data['status_message'] = 'Invalid response has been received for viewer flow creation for job .'.$job_uuid.'! ';
                                // Client side error
                                if (intval($gears_response_code) >= 400 && intval($gears_response_code) <= 499) {
                                    \OCP\Util::writeLog('signlive', 'Error caused by client side. Error code: '.intval($gears_response_code), ILogger::ERROR);
                                    $signature_job_data['status_message'] = $signature_job_data['status_message'].'Error caused by client side. Error code: '.intval($gears_response_code); 
                                }
                                // Server side error
                                if (intval($gears_response_code) >= 500 && intval($gears_response_code) <= 599) {
                                    \OCP\Util::writeLog('signlive', 'Error caused by server side. Error code: '.intval($gears_response_code), ILogger::ERROR);
                                    $signature_job_data['status_message'] = $signature_job_data['status_message'].'Error caused by server side. Error code: '.intval($gears_response_code);
                                }
                                
                                if (array_key_exists('_error', $json_response_array)) {
                                    if (array_key_exists('code', $json_response_array['_error'])) {
                                        \OCP\Util::writeLog('signlive', 'Gears backend error code: '.$json_response_array['_error']['code'], ILogger::ERROR);
                                        $signature_job_data['status_message'] = $signature_job_data['status_message'].' Gears backend error code: '.$json_response_array['_error']['code'];
                                    }
                                    if (array_key_exists('message', $json_response_array['_error'])) {
                                        \OCP\Util::writeLog('signlive', 'Gears backend error message: '.$json_response_array['_error']['message'], ILogger::ERROR);
                                        $signature_job_data['status_message'] = $signature_job_data['status_message'].' Gears backend error message: '.$json_response_array['_error']['message'];                                        
                                    }
                                }

                                // Invalid response received for viewer flow creation, update job state
                                $this->setNewJobState($job_uuid, 'FLOW_CREATION_RESPONSE_INVALID', 'Invalid response has been received for viewer flow creation for job .'.$job_uuid.'!', true);
                                $this->sendFailureNotification($signature_job_data);

                                return new DataResponse([
                                    'message' => 'Received an error response from Gears backend: '.$gears_response->getStatusCode().' ('.$gears_response->getReasonPhrase().')',
                                    'viewer_url' => '',                    
                                    'signature_type' => 'viewer_based_signature',
                                    'success' => false
                                ]);
                            }
                        } catch (RequestException $e) {
                            \OCP\Util::writeLog('signlive', 'Error occurred while sending signature request to Gears backend: '.Psr7\Message::toString($e->getRequest()), ILogger::ERROR);
                            $signature_job_data['status_message'] = 'Error occurred while sending signature request to Gears backend: '.Psr7\Message::toString($e->getRequest());

                            if ($e->hasResponse()) {
                                \OCP\Util::writeLog('signlive', 'Response data returned by Gears backend: '.Psr7\Message::toString($e->getResponse()), ILogger::ERROR);
                                $signature_job_data['status_message'] = $signature_job_data['status_message'].' Response data returned by Gears backend: '.Psr7\Message::toString($e->getResponse());
                            }

                            // Something went wrong with the request to the Gears backend, update job state
                            $this->setNewJobState($job_uuid, 'FLOW_CREATION_REQUEST_FAILED', 'Exception occurred dispatching request to the Gears backend for viewer flow creation for job .'.$job_uuid.'!', true);
                            $this->sendFailureNotification($signature_job_data);

                            return new DataResponse([
                                'message' => 'Error occurred while communicating with the Gears backend! Signature request failed.',
                                'viewer_url' => '',
                                'signature_type' => 'viewer_based_signature',
                                'success' => false
                            ]);
                        } catch (ClientException $e) {
                            \OCP\Util::writeLog('signlive', 'Error occurred while processing the signature request on the Gears backend! Request data: '.Psr7\Message::toString($e->getRequest()), ILogger::ERROR);
                            \OCP\Util::writeLog('signlive', 'Error occurred while processing the signature request on the Gears backend! Response data: '.Psr7\Message::toString($e->getResponse()), ILogger::ERROR);

                            // Something went wrong on the client side while communicating with the Gears backend, update job state
                            $this->setNewJobState($job_uuid, 'FLOW_CREATION_REQUEST_FAILED', 'Exception occurred dispatching request to the Gears backend for viewer flow creation for job .'.$job_uuid.'!', true);
                            
                            $signature_job_data['status_message'] = 'Error occurred while processing the signature request on the Gears backend! Request data: '.Psr7\Message::toString($e->getRequest());
                            $signature_job_data['status_message'] = $signature_job_data['status_message'].'; Response data: '.Psr7\Message::toString($e->getResponse());
                            $this->sendFailureNotification($signature_job_data);

                            return new DataResponse([
                                'message' => 'Error occurred while processing the response received from the Gears backend! Signature request failed.',
                                'viewer_url' => '',
                                'signature_type' => 'viewer_based_signature',
                                'success' => false
                            ]);
                        }
                    } else {
                        \OCP\Util::writeLog('signlive', 'Failed to sign file using Gears viewer profile: '.$file_to_sign[0].' Creating a valid JSON request for submission of the file to sign failed!', ILogger::ERROR);
 
                        // JSON request for flow creation cloud not be create, update job state
                        $this->setNewJobState($job_uuid, 'FLOW_CREATION_JSON_PROCESSING_FAILED', 'Failed to sign file using Gears viewer profile: '.$file_to_sign[0].' Creating a valid JSON request for submission of the file to sign failed for job '.$job_uuid.'!', true);
                            
                        $signature_job_data['status_message'] = 'Failed to sign file using Gears viewer profile: '.$file_to_sign[0].' Creating a valid JSON request for submission of the file to sign failed for job '.$job_uuid.'!';
                        $this->sendFailureNotification($signature_job_data);

                        return new DataResponse([
                            'message' => 'Failed to sign file using Gears viewer profile: '.$file_to_sign[0].' Creating a valid JSON request for submission of the file to sign failed!',
                            'viewer_url' => '',
                            'signature_type' => 'viewer_based_signature',
                            'success' => false
                        ]);
                    }
                } else {
                    \OCP\Util::writeLog('signlive', 'Failed to sign document \''.$file_to_sign[0].'\': File not found or not readable!', ILogger::ERROR);
 
                    // File to sign could not be found or is not readable, update job state
                    $this->setNewJobState($job_uuid, 'FLOW_CREATION_FILE_NOT_FOUND', 'Failed to sign document \''.$file_to_sign[0].'\': File not found or not readable for job '.$job_uuid.'!', true);
                            
                    $signature_job_data['status_message'] = 'Failed to sign document \''.$file_to_sign[0].'\': File not found or not readable for job '.$job_uuid.'!';
                    $this->sendFailureNotification($signature_job_data);

                    return new DataResponse([
                        'message' => 'Failed to sign document \''.$file_to_sign[0].'\': File not found or not readable!',
                        'viewer_url' => '',
                        'signature_type' => 'viewer_based_signature',
                        'success' => false
                    ]);
                }
            } else {
                // No valid JSON section could be found for the given profile, update job state
                $this->setNewJobState($job_uuid, 'FLOW_CREATION_JSON_PROCESSING_FAILED', 'Found no valid JSON section for profile \''.$profile_shortened_unique_id.'\' of type \''.$profile_type.'\'. Could not sign document '.$file_to_sign[0].' in job '.$job_uuid.'!', true);
                            
                $signature_job_data['status_message'] = 'Found no valid JSON section for profile \''.$profile_shortened_unique_id.'\' of type \''.$profile_type.'\'. Could not sign document '.$file_to_sign[0].' in job '.$job_uuid.'!';
                $this->sendFailureNotification($signature_job_data);

                return new DataResponse([
                    'message' => 'Found no valid JSON section for profile \''.$profile_shortened_unique_id.'\' of type \''.$profile_type.'\'. Could not sign document '.$file_to_sign[0].'!',
                    'viewer_url' => '',
                    'signature_type' => 'viewer_based_signature',
                    'success' => false
                ]);
            }
        } else {
            \OCP\Util::writeLog('signlive', 'signDocumentViaViewerJob() will only accept a single document to sign! Doing nothing.', ILogger::ERROR);           
            
            // Only a single document is supported per signature request job, update job state
            $this->setNewJobState($job_uuid, 'FLOW_CREATION_GENERAL_FAILURE', 'signDocumentViaViewerJob() will only accept a single document to sign in job '.$job_uuid.'!', true);
                        
            $signature_job_data['status_message'] = 'signDocumentViaViewerJob() will only accept a single document to sign in job '.$job_uuid.'!';
            $this->sendFailureNotification($signature_job_data);
            
            return new DataResponse([
                'message' => 'signDocumentViaViewerJob() will only accept a single document to sign! Doing nothing.',
                'viewer_url' => '',
                'signature_type' => 'viewer_based_signature',
                'success' => false
            ]);   
        }
    }

    /**
	 * @NoAdminRequired
     * @UseSession
     * @NoCSRFRequired
     * TODO: CSRF support?
     * Process the Gears response for a viewer-based signature request 
     * 
	 * @return RedirectResponse
	 */
    public function processViewerJobResponse() {
        \OCP\Util::writeLog('signlive', '====================================================================================================================', ILogger::DEBUG);
        \OCP\Util::writeLog('signlive', 'processViewerJobResponse() called in signlive/lib/Controller/IntarsysJobSignatureController.php.', ILogger::DEBUG);
        \OCP\Util::writeLog('signlive', '====================================================================================================================', ILogger::DEBUG);
    
        \OCP\Util::writeLog('signlive', 'Response with final result received from viewer-based signature request.', ILogger::DEBUG);
        $request_parameters = $this->request->getParams();

        $cs_conversation = '';
        $cs_stage = '';
        $original_folder_uri = '';
        $document_name = '';
        $job_uuid = '';
        $redirect_url = '';

        foreach ($request_parameters as $parameter_name => $parameter_value) {
            \OCP\Util::writeLog('signlive', 'Parameter \''.$parameter_name.'\' => \''.$parameter_value.'\'', ILogger::DEBUG);
            // Parameters expected to be present, sent URL-encoded from the Nextcloud frontend
            if ($parameter_name === "job_uuid") {
                $job_uuid = $parameter_value;
            }
            if ($parameter_name === "cs_conversation") {
                $cs_conversation = $parameter_value;
            }
            if ($parameter_name === "cs_stage") {
                $cs_stage = $parameter_value;
            }
        }

        // Verify all required signature request parameters have actually been provided
        $parameters_valid = true;
        $parameter_error_msg = '';
        if (empty($cs_conversation) || $cs_conversation === '') {
            $parameter_error_msg = $parameter_error_msg.' -- No valid Gears conversation ID was provided.';
            $parameters_valid = false;
        }
        if (empty($job_uuid) || $job_uuid === '') {
            $parameter_error_msg = $parameter_error_msg.' -- No job UUID was provided.';
            $parameters_valid = false;
        }

        // If not all required parameters are present: Return early
        if ($parameters_valid === false) {
            \OCP\Util::writeLog('signlive', 'Received invalid answer to viewer-based signature request query, not all required parameters were provided by the caller: '.$parameter_error_msg, ILogger::WARN);
            
            // No signature request job state exists at this point: Just send a failure notification if there is a valid job_uuid
            // Provide what is possible in job record data and let sendFailureNotification take care of assembling a notification
            if (!(empty($job_uuid) || $job_uuid === '')) {
                $job_record_data = [];
                $job_record_data['signature_job_uuid'] = $job_uuid;
                if (!(empty($file_name) || $file_name === '')) {
                    $job_record_data['file_name'] = $file_name;
                }
                if (!(empty($signature_type) || $signature_type === '')) {
                    $job_record_data['signature_type'] = $signature_type;
                }
                $job_record_data['nextcloud_user'] = $this->userSession->getUser()->getUID();
                $job_record_data['status_message'] = 'Received invalid viewer-based signature request query, not all required parameters were provided: '.$parameter_error_msg;
                
                $this->sendFailureNotification($job_record_data);
            }

            return new RedirectResponse($redirect_url);
        }

        $signature_job_data = $this->getJobDefinitionEntry($job_uuid);

        $document_name = $signature_job_data['file_name'];
        $original_folder_uri = $signature_job_data['original_folder_uri'];

        \OCP\Util::writeLog('signlive', '===================================================================================================================================', ILogger::DEBUG);
        \OCP\Util::writeLog('signlive', 'cs_conversation = '.$cs_conversation.', document_name = '.$document_name, ILogger::DEBUG);

        if ($original_folder_uri !== '') {
            $redirect_url = $original_folder_uri;
            \OCP\Util::writeLog('signlive', 'redirect_url = '.$redirect_url, ILogger::DEBUG);
            \OCP\Util::writeLog('signlive', '===================================================================================================================================', ILogger::DEBUG);
        // Redirect to the Nextcloud files app index page if there is no original folder in the redirect URI for some reason 
        } else {
            $redirect_url = $this->urlGenerator->linkToRouteAbsolute(
            'files.viewcontroller.index');

            \OCP\Util::writeLog('signlive', 'No valid redirect URL was provided to return to Nextcloud for signature request job '.$job_uuid.'! Redirecting to files app main page now.', ILogger::WARN);

            $signature_job_data['status_message'] = 'No valid redirect URL was provided to return to Nextcloud for signature request job '.$job_uuid.'! Redirecting to files app main page now.';
            
            // No redirect back to Nextcloud is possible, update job state accordingly
            $this->setNewJobState($job_uuid, 'RESULT_RETRIEVAL_GENERAL_FAILURE', 'No valid redirect URL back to Nextcloud has been provided for signature request job .'.$job_uuid.'!', true);
            $this->sendFailureNotification($signature_job_data);
            \OCP\Util::writeLog('signlive', '===================================================================================================================================', ILogger::DEBUG);

            return new RedirectResponse($redirect_url);
        }

        // At this point all required information for retrieving the signature result from the Gears backend is present
        if ($cs_conversation != '' && $document_name != '') {
            $cloudsuite_gears_http_protocol_type = $this->config->getAppValue('signlive', 'cloudsuite_gears_http_protocol_type', SignLiveDefaultConfigValues::$GEARS_SERVER_PROTOCOL);
            $cloudsuite_gears_host = $this->config->getAppValue('signlive', 'cloudsuite_gears_host', SignLiveDefaultConfigValues::$GEARS_SERVER_HOST);
            $cloudsuite_gears_port = $this->config->getAppValue('signlive', 'cloudsuite_gears_port', SignLiveDefaultConfigValues::$GEARS_SERVER_PORT);
            $cloudsuite_gears_gears_core_path = $this->config->getAppValue('signlive', 'cloudsuite_gears_gears_core_path', SignLiveDefaultConfigValues::$GEARS_SERVER_CORE_PATH);
            
            $json_query = $this->jsonQueryCreator->getViewerResultJSON($cs_conversation);

            \OCP\Util::writeLog('signlive', 'JSON query string (shortened) for viewer-based result retrieval request: '.$json_query, ILogger::DEBUG);

            $gears_backend_url = $cloudsuite_gears_http_protocol_type.'://'.$cloudsuite_gears_host.':'.$cloudsuite_gears_port.$cloudsuite_gears_gears_core_path.'/api/v1/flow/conversation/acknowledge';

            // Result retrieval is pending, update job state accordingly
            $this->setNewJobState($job_uuid, 'RESULT_RETRIEVAL_PENDING', 'Result retrieval pending for signature request job .'.$job_uuid.'.', true);

            $gears_rest_client = new \GuzzleHttp\Client();

            try {
                $gears_response = $gears_rest_client->post($gears_backend_url, [
                    'headers' => ['Content-Type' => 'application/json', 'Accept' => 'application/json'],
                    'body'     => $json_query
                    ]);
            } catch (RequestException $e) {
                \OCP\Util::writeLog('signlive', 'Error occurred while sending signature request to Gears backend: '.Psr7\Message::toString($e->getRequest()), ILogger::ERROR);

                $status_message = 'Result retrieval failed due to connection issues for signature request job .'.$job_uuid.'!';
                if ($e->hasResponse()) {
                    \OCP\Util::writeLog('signlive', 'Response data returned by Gears backend: '.Psr7\Message::toString($e->getResponse()), ILogger::ERROR);
                    $status_message = $status_message.' Response data returned by Gears backend: '.Psr7\Message::toString($e->getResponse());
                }
                // Communication failure with Gears backend, update job state accordingly
                $this->setNewJobState($job_uuid, 'RESULT_RETRIEVAL_COMMUNICATION_FAILURE', $status_message, true);
                
                $signature_job_data['status_message'] = $status_message;
                $this->sendFailureNotification($signature_job_data);

                return new RedirectResponse($redirect_url);
            } catch (ClientException $e) {
                \OCP\Util::writeLog('signlive', 'Error occurred while processing the signature request on the Gears backend! Request data: '.Psr7\Message::toString($e->getRequest()), ILogger::ERROR);
                \OCP\Util::writeLog('signlive', 'Error occurred while processing the signature request on the Gears backend! Response data: '.Psr7\Message::toString($e->getResponse()), ILogger::ERROR);
                
                $status_message = 'Result retrieval failed due to client-side issues for signature request job .'.$job_uuid.'!';
                $status_message = $status_message.' -- Request data: '.Psr7\Message::toString($e->getRequest());
                $status_message = $status_message.' -- Response data: '.Psr7\Message::toString($e->getResponse());

                // Client-side failure during result retrieval from backend, update job state accordingly
                $this->setNewJobState($job_uuid, 'RESULT_RETRIEVAL_CLIENT_SIDE_FAILURE', $status_message, true);

                $signature_job_data['status_message'] = $status_message;
                $this->sendFailureNotification($signature_job_data);

                return new RedirectResponse($redirect_url);
            }

            $decoded_gears_response_body = json_decode($gears_response->getBody(), true);
            $json_response_array = (array) $decoded_gears_response_body;

            // Omit logging for potentially large string-type data: base64-encoded signature result!
            // \OCP\Util::writeLog('signlive', 'JSON response from Gears backend (shortened): \"'.print_r($json_response_array, TRUE).'\"', ILogger::DEBUG);

            $document_names = array();
            $signatures = array();
            $gears_result_valid = false;
            if (array_key_exists('snapshot', $json_response_array)) {
                \OCP\Util::writeLog('signlive', 'Found \'snapshot\' element in JSON response.', ILogger::DEBUG);
                if (array_key_exists('stage', $json_response_array['snapshot'])) {
                    \OCP\Util::writeLog('signlive', 'Found \'stage\' element in JSON response.', ILogger::DEBUG);
                    if (array_key_exists('result', $json_response_array['snapshot']['stage'])) {
                        \OCP\Util::writeLog('signlive', 'Found \'result\' element in JSON response.', ILogger::DEBUG);
                        if (array_key_exists('value', $json_response_array['snapshot']['stage']['result'])) {
                            \OCP\Util::writeLog('signlive', 'Found outer \'value\' element in JSON response.', ILogger::DEBUG);
                            if (array_key_exists('artifacts', $json_response_array['snapshot']['stage']['result']['value'])) {
                                \OCP\Util::writeLog('signlive', 'Found \'artifacts\' element in JSON response: artifacts count = '.count($json_response_array['snapshot']['stage']['result']['value']['artifacts']), ILogger::DEBUG);
                                
                                // Using the signature viewer only allows for one document to be digitally signed at a time
                                // Consequently everything that returns multiple signed documents is ignored here
                                if (count($json_response_array['snapshot']['stage']['result']['value']['artifacts']) > 0) {
                                    $artifacts = $json_response_array['snapshot']['stage']['result']['value']['artifacts'];
                                    
                                    // Look for the kind of result object that is present in the 'artifacts' section of the response
                                    // Expected values are one of two different @class-es in a viewer signature response
                                    // * intarsys.cloudsuite.gears.core.service.common.api.DtoResultDocument
                                    // * intarsys.cloudsuite.gears.core.service.signer.api.ResultSigner
                                    
                                    $result_signer_present = 0;
                                    $result_signer_section = null;
                                    $dto_result_document_present = 0;
                                    $dto_result_document_section = null;
                                    
                                    $artifact_index = 0;
                                    foreach ($artifacts as $artifact) {
                                        \OCP\Util::writeLog('signlive', 'Looking at \'artifact\' element: '.$artifact_index, ILogger::DEBUG);
                                        \OCP\Util::writeLog('signlive', '\'artifact\' element: '.print_r($artifact, TRUE), ILogger::DEBUG);
                                        if (array_key_exists('value', $artifact)) {
                                            \OCP\Util::writeLog('signlive', 'Found outer \'value\' element in artifact.', ILogger::DEBUG);
                                            if (array_key_exists('@class', $artifact['value'])) {
                                                $class_value = $artifact['value']['@class'];
                                                \OCP\Util::writeLog('signlive', 'Found \'@class\' element in artifact: '.$class_value, ILogger::DEBUG);
                                                if ($class_value === 'de.intarsys.cloudsuite.gears.core.service.signer.api.ResultSigner') {
                                                    \OCP\Util::writeLog('signlive', 'Found \'ResultSigner\' element in artifacts.', ILogger::DEBUG);

                                                    $result_signer_present = 1;
                                                    $result_signer_section = $artifact;
                                                }
                                                if ($class_value === 'de.intarsys.cloudsuite.gears.core.service.common.api.DtoResultDocument') {
                                                    \OCP\Util::writeLog('signlive', 'Found \'DtoResultDocument\' element in artifacts.', ILogger::DEBUG);
                                                    
                                                    $dto_result_document_present = 1;
                                                    $dto_result_document_section= $artifact;
                                                }
                                            }
                                        }
                                        $artifact_index += 1;
                                    }

                                    \OCP\Util::writeLog('signlive', '$result_signer_present = '.$result_signer_present.'; $dto_result_document_present = '.$dto_result_document_present, ILogger::DEBUG);

                                    $signature_result_section = null;
                                    \OCP\Util::writeLog('signlive', '===================================================================================================================================', ILogger::DEBUG);
                                    if ($result_signer_present === 1 && $dto_result_document_present === 0) {
                                        \OCP\Util::writeLog('signlive', 'ONLY a \'ResultSigner\' element is present, using it.', ILogger::DEBUG);
                                        $signature_result_section = $result_signer_section;
                                    } elseif ($result_signer_present === 0 && $dto_result_document_present === 1) {
                                        \OCP\Util::writeLog('signlive', 'ONLY a \'DtoResultDocument\' element is present, using it.', ILogger::DEBUG);
                                        $signature_result_section = $dto_result_document_section;
                                    } elseif ($result_signer_present === 1 && $dto_result_document_present === 1) {
                                        \OCP\Util::writeLog('signlive', 'BOTH a \'ResultSigner\' AND a \'DtoResultDocument\' element are present, using the former.', ILogger::DEBUG);
                                        $signature_result_section = $result_signer_section;
                                    } elseif ($result_signer_present === 0 && $dto_result_document_present === 0) {
                                        // This very likely indicates an empty result returned by the Gears backend
                                        // I. e. the user did not sign or modify the the document otherwise
                                        \OCP\Util::writeLog('signlive', 'NEITHER a \'ResultSigner\' NOR a \'DtoResultDocument\' element is present, Gears returned an empty result.', ILogger::DEBUG);
                                        // "Empty success" is handled at the end of this function
                                        $gears_response_valid = true;
                                    }
                                    \OCP\Util::writeLog('signlive', '===================================================================================================================================', ILogger::DEBUG);

                                    if ($result_signer_present === 1 && $signature_result_section !== null) {
                                        if (array_key_exists('value', $signature_result_section)) {
                                            \OCP\Util::writeLog('signlive', 'Found outer \'value\' element in ResultSigner element.', ILogger::DEBUG);
                                            if (array_key_exists('value', $signature_result_section['value'])) {
                                                \OCP\Util::writeLog('signlive', 'Found inner \'value\' element in ResultSigner element.', ILogger::DEBUG);
                                                if (array_key_exists('documentNames', $signature_result_section['value']['value'])) {
                                                    $document_names = $signature_result_section['value']['value']['documentNames'];
                                                    \OCP\Util::writeLog('signlive', 'documentNames in response: '.count($document_names), ILogger::DEBUG);
                                                
                                                    // There is one element per signature operation in the 'signatures' array
                                                    // Only the last one is relevant, as each stage only includes the signatures created up to its array index
                                                    // Consequently only the last array element is relevant here since it is the only one that actually includes all signatures
                                                    if (array_key_exists('signatures', $signature_result_section['value']['value'])) {
                                                        $signatures = $signature_result_section['value']['value']['signatures'];
                                                        \OCP\Util::writeLog('signlive', 'signatures in response: '.count($signatures), ILogger::DEBUG);
                                                        if (count($signatures) > 0) {
                                                            // Retrieve the base64-encoded signed PDF document from the last element in the 'signatures' array
                                                            $relevant_signature_element = $signatures[count($signatures) - 1];
                                                            if (array_key_exists('content', $relevant_signature_element)) {
                                                                \OCP\Util::writeLog('signlive', 'Found base64-encoded signed PDF document.', ILogger::DEBUG);
                                                                $base64_decoded_signed_content = base64_decode($relevant_signature_element['content']);
                                                                
                                                                $current_user_folder = $this->storage->getUserFolder($this->userSession->getUser()->getUid());

                                                                \OCP\Util::writeLog('signlive', 'Overwriting original unsigned PDF document, saving signed document to file: '.$relevant_signature_element['name'], ILogger::DEBUG);
                                                                    
                                                                $file_name_to_save = $document_name;
                                                                $current_user_folder->newFile($file_name_to_save, $base64_decoded_signed_content);

                                                                \OCP\Util::writeLog('signlive', '===================================================================================================================================', ILogger::INFO);
                                                                \OCP\Util::writeLog('signlive', 'PDF document signed successfully using viewer-based signature request: '.$file_name_to_save.', unsigned original has been replaced.', ILogger::INFO);
                                                                \OCP\Util::writeLog('signlive', '===================================================================================================================================', ILogger::INFO);

                                                                $gears_result_valid = true;
                                                                // At this point a valid signature result has been retrieved, update job state accordingly
                                                                $this->setNewJobState($job_uuid, 'RESULT_RETRIEVAL_SUCCEEDED', 'Document \''.$document_name.'\' signed successfully with viewer-based signature request job '.$job_uuid.'.', false);
                                                                
                                                                $signature_job_data['status_message'] = 'Document \''.$document_name.'\' signed successfully with viewer-based signature request job '.$job_uuid.'.';
                                                                $this->sendSuccessNotification($signature_job_data);
                                                            
                                                                return new RedirectResponse($redirect_url);
                                                            }
                                                        } else {
                                                            \OCP\Util::writeLog('signlive', '===================================================================================================================================', ILogger::INFO);
                                                            \OCP\Util::writeLog('signlive', 'No \'signature\' element found in ResultSigner element in signature request job '.$job_uuid.'.', ILogger::WARN);
                                                            \OCP\Util::writeLog('signlive', '===================================================================================================================================', ILogger::INFO);
                                                            
                                                            // No valid signed document has been returned by Gears, update job state accordingly
                                                            $this->setNewJobState($job_uuid, 'RESULT_RETRIEVAL_EMPTY_RESPONSE', 'No \'signature\' element found in ResultSigner element in signature request job '.$job_uuid.'.', false);
                                                            
                                                            $signature_job_data['status_message'] = 'No \'signature\' element found in ResultSigner element in signature request job '.$job_uuid.'.';
                                                            $this->sendSuccessNotification($signature_job_data);
                                                            
                                                            return new RedirectResponse($redirect_url);
                                                        }
                                                    }
                                                } else {
                                                    // This can also happen if the user did not sign the document on purpose, so report DEBUG only
                                                    \OCP\Util::writeLog('signlive', '===================================================================================================================================', ILogger::INFO);
                                                    \OCP\Util::writeLog('signlive', 'No \'documentNames\' element found in ResultSigner element in signature request job '.$job_uuid.'.', ILogger::DEBUG);
                                                    \OCP\Util::writeLog('signlive', '===================================================================================================================================', ILogger::INFO);

                                                    // No valid signed document has been returned by Gears, update job state accordingly
                                                    $this->setNewJobState($job_uuid, 'RESULT_RETRIEVAL_EMPTY_RESPONSE', 'No \'documentNames\' element found in ResultSigner element in signature request job '.$job_uuid.'.', false);
                                                    
                                                    $signature_job_data['status_message'] = 'No \'documentNames\' element found in ResultSigner element in signature request job '.$job_uuid.'.';
                                                    $this->sendSuccessNotification($signature_job_data);
                                                    
                                                    return new RedirectResponse($redirect_url);
                                                }
                                            }
                                        }                                        
                                    } elseif ($dto_result_document_present === 1 && $dto_result_document_section !== null) {
                                        if (array_key_exists('value', $signature_result_section)) {
                                            \OCP\Util::writeLog('signlive', 'Found outer \'value\' element in DtoResultDocument element.', ILogger::DEBUG);
                                            if (array_key_exists('value', $signature_result_section['value'])) {
                                                \OCP\Util::writeLog('signlive', 'Found inner \'value\' element in DtoResultDocument element.', ILogger::DEBUG);
                                                if (array_key_exists('document', $signature_result_section['value']['value'])) {
                                                    \OCP\Util::writeLog('signlive', 'Found \'document\' element in DtoResultDocument element.', ILogger::DEBUG);
                                                    if (array_key_exists('content', $signature_result_section['value']['value']['document'])) {
                                                        \OCP\Util::writeLog('signlive', 'Found \'content\' element in DtoResultDocument element.', ILogger::DEBUG);
                                                        $base64_decoded_signed_content = base64_decode($signature_result_section['value']['value']['document']['content']);
                                                        
                                                        $current_user_folder = $this->storage->getUserFolder($this->userSession->getUser()->getUid());

                                                        \OCP\Util::writeLog('signlive', 'Overwriting original unsigned PDF document, saving signed document to file: '.$relevant_signature_element['name'], ILogger::DEBUG);
                                                            
                                                        $file_name_to_save = $document_name;
                                                        $current_user_folder->newFile($file_name_to_save, $base64_decoded_signed_content);

                                                        \OCP\Util::writeLog('signlive', 'PDF document signed successfully using viewer-based signature request: '.$file_name_to_save.', unsigned original has been replaced.', ILogger::INFO);

                                                        $gears_result_valid = true;
                                                        // At this point a valid signature result has been retrieved, update job state accordingly
                                                        $this->setNewJobState($job_uuid, 'RESULT_RETRIEVAL_SUCCEEDED', 'Document \''.$document_name.'\' signed successfully with viewer-based signature request job '.$job_uuid.'.', false);
                                                        
                                                        $signature_job_data['status_message'] = 'Document \''.$document_name.'\' signed successfully with viewer-based signature request job '.$job_uuid.'.';
                                                        $this->sendSuccessNotification($signature_job_data);
                                                    
                                                        return new RedirectResponse($redirect_url);
                                                    }
                                                }
                                            }
                                        }
                                    }
                                } else {
                                    \OCP\Util::writeLog('signlive', 'No \'artifacts\' element found in result for viewer-based signature request job '.$job_uuid.'. This indicates there was no signature added.', ILogger::DEBUG);
                                    
                                    // No 'artifacts' element was found in the Gears result, i. e. no signature was added. Update job state accordingly, treating this as success.
                                    $this->setNewJobState($job_uuid, 'RESULT_RETRIEVAL_NO_SIGNATURE_ADDED', 'No signature was added in viewer-based signature request job '.$job_uuid.'.', false);
                                    $signature_job_data['status_message'] = 'No signature was added in viewer-based signature request job '.$job_uuid.'.';                                                  
                                    $this->sendSuccessNotification($signature_job_data);
                                                                        
                                    return new RedirectResponse($redirect_url);
                                }
                            }
                        }
                    }
                }
            }

            // Should something have gone wrong while parsing the response data received from the Gears backend that has not been taken care of specifically: 
            // Log the relevant response data ('snapshot' and content) and update the job state accordingly 
            if ($gears_result_valid === false) {
                $status_message = 'Parsing the response of the Gears backend failed viewer-based signature request job '.$job_uuid.'!';
                $status_message = $status_message.' Response received from Gears backend: '.$gears_response->getBody();
                $this->setNewJobState($job_uuid, 'RESULT_RETRIEVAL_FAILURE_INVALID', $status_message, true);
                
                $signature_job_data['status_message'] = $status_message;                                                  
                $this->sendFailureNotification($signature_job_data);
            } else {
                // Control flow should never get here, successful signature job results are taken of earlier -- unless Gears returned an empty response, which we still consider a successful job completion.
                \OCP\Util::writeLog('signlive', 'Received an empty PDF document from Gears backend server. Redirecting back to: '.$redirect_url, ILogger::DEBUG);

                \OCP\Util::writeLog('signlive', '=================================================================================================', ILogger::DEBUG);
                \OCP\Util::writeLog('signlive', 'Reached the end of processViewerJobResponse(). No result was obtained, about to trigger redirect.', ILogger::DEBUG);
                \OCP\Util::writeLog('signlive', '=================================================================================================', ILogger::DEBUG);    
            }
            return new RedirectResponse($redirect_url);
        } else {
            \OCP\Util::writeLog('signlive', 'Final result for viewer-based signature request failed: No valid conversation_id and/or no valid document name were returned in the reply from the Gears backend!', ILogger::ERROR);

            $status_message = 'Final result for viewer-based signature request failed: No valid conversation_id and/or no valid document name were returned in the reply from the Gears backend for viewer-based signature request job '.$job_uuid.'!';
            $this->setNewJobState($job_uuid, 'RESULT_RETRIEVAL_FAILURE_INVALID', $status_message, true);
            
            $signature_job_data['status_message'] = $status_message;                                                  
            $this->sendFailureNotification($signature_job_data);

            return new RedirectResponse($redirect_url);
        }
    }

    /**
	 * @NoAdminRequired
	 *
	 * Sign a document using a parameterized REST call
	 *
	 * @return DataResponse
	 */
    public function signDocumentViaProfileJob() {
        \OCP\Util::writeLog('signlive', '=====================================================================================================================', ILogger::DEBUG);
        \OCP\Util::writeLog('signlive', 'signDocumentViaProfileJob() called in signlive/lib/Controller/IntarsysJobSignatureController.php.', ILogger::DEBUG);
        
        $request_parameters = $this->request->getParams();
        $file_name = '';
        $profile_unique_id = '';
        $redirect_url = '';
        $profile_type = '';
        $signature_type = '';
        $profile_index = -1;
        $profile_identifier = '';
        $job_uuid = '';
        foreach ($request_parameters as $parameter_name => $parameter_value) {
            \OCP\Util::writeLog('signlive', 'Parameter \''.$parameter_name.'\' => \''.$parameter_value.'\'', ILogger::DEBUG);
            if ($parameter_name == 'filename') {
                $file_name = $parameter_value;
            }
            if ($parameter_name === 'profile_unique_id') {
                $profile_unique_id = $parameter_value;
            }
            if ($parameter_name === 'redirect_uri') {
                $redirect_url = $parameter_value;
            }
            if ($parameter_name === 'profile_type') {
                $profile_type = $parameter_value;
            }
            if ($parameter_name === 'signature_type') {
                $signature_type = $parameter_value;
            }
            if ($parameter_name === 'profile_index') {
                $profile_index = $parameter_value;
            }
            if ($parameter_name === 'profile_identifier') {
                $profile_identifier = $parameter_value;
            }
            if ($parameter_name === 'profile_unique_id') {
                $signature_profile_id = $parameter_value;
            }
            if ($parameter_name === 'profile_type') {
                $signature_profile_type = $parameter_value;
            }
            if ($parameter_name === 'job_uuid') {
                $job_uuid = $parameter_value;
            }
        }

        // Verify all required signature request parameters have actually been provided
        $parameters_valid = true;
        $parameter_error_msg = '';
        if (empty($file_name) || $file_name === '') {
            \OCP\Util::writeLog('signlive', 'signDocumentViaProfileJob(): No file name was provided!', ILogger::WARN);
            
            $parameter_error_msg = 'No file name was provided.';
            $parameters_valid = false;
        }
        if (empty($signature_profile_id) || $signature_profile_id === '') {
            \OCP\Util::writeLog('signlive', 'signDocumentViaProfileJob(): No valid signature unique ID was provided!', ILogger::WARN);
            $parameter_error_msg = $parameter_error_msg.' -- No valid signature profile unique ID was provided.';
            $parameters_valid = false;
        }
        if (empty($signature_profile_type) || $signature_profile_type === '') {
            \OCP\Util::writeLog('signlive', 'signDocumentViaProfileJob(): No valid signature profile type was provided!', ILogger::WARN);
            $parameter_error_msg = $parameter_error_msg.' -- No valid signature profile type was provided.';
            $parameters_valid = false;
        }
        if (empty($job_uuid) || $job_uuid === '') {
            \OCP\Util::writeLog('signlive', 'signDocumentViaProfileJob(): No valid job UUID was provided!', ILogger::WARN);
            $parameter_error_msg = $parameter_error_msg.' -- No job UUID was provided.';
            $parameters_valid = false;
        }

        // If not all required parameters are present: Return early
        if ($parameters_valid === false) {
            \OCP\Util::writeLog('signlive', 'Received invalid automatic signature request query, not all required parameters were provided by the caller: '.$parameter_error_msg, ILogger::WARN);
            
            // No signature request job state exists at this point: Just send a failure notification if there is a valid job_uuid
            // Provide what is possible in job record data and let sendFailureNotification take care of assembling a notification
            if (!(empty($job_uuid) || $job_uuid === '')) {
                $job_record_data = [];
                $job_record_data['signature_job_uuid'] = $job_uuid;
                if (!(empty($file_name) || $file_name === '')) {
                    $job_record_data['file_name'] = $file_name;
                }
                if (!(empty($signature_type) || $signature_type === '')) {
                    $job_record_data['signature_type'] = $signature_type;
                }
                $job_record_data['nextcloud_user'] = $this->userSession->getUser()->getUID();
                
                $job_record_data['status_message'] = 'Received invalid automatic signature request query, not all required parameters were provided: '.$parameter_error_msg;
                $this->sendFailureNotification($job_record_data);
            }
            \OCP\Util::writeLog('signlive', '=====================================================================================================================', ILogger::DEBUG);
            return new DataResponse([
                'message' => 'Not all required parameters for an automatic signature request were provided, can not continue: '.$parameter_error_msg,
                'signature_type' => 'automatic_signature',
                'success' => false
            ]);
        }

        \OCP\Util::writeLog('signlive', '=====================================================================================================================', ILogger::DEBUG);

        $signature_job_data['signature_job_uuid'] = $job_uuid;
        $signature_job_data['signature_job_name'] = 'Signature job \''.$file_name.'\''; 
        $signature_job_data['signature_job_type'] = 'automatic_signature';
        $signature_job_data['signature_job_file'] = $file_name;
        $signature_job_data['signature_profile_type'] = $signature_profile_type;
        $signature_job_data['nextcloud_user'] = $this->userSession->getUser()->getUid();
        $signature_job_data['job_creation_time'] = time();
        
        $file_list_to_sign = $this->getFilesToSign();

        if (count($file_list_to_sign) == 1) {
            \OCP\Util::writeLog('signlive', 'Using signature profile unique ID \''.$profile_unique_id.'\' of type \''.$profile_type.'\'', ILogger::DEBUG);
            
            $user_profile_json_section = '';
            $profile_shortened_unique_id = str_replace('cloudsuite_gears_profile_id_', '', $profile_unique_id);
            $profile_json_section_key = 'cloudsuite_gears_profile_json_'.$profile_shortened_unique_id;
            \OCP\Util::writeLog('signlive', 'Query JSON section for profile '.$profile_index.' using key: '.$profile_json_section_key, ILogger::DEBUG);
            
            $profile_json_section = $this->config->getAppValue('signlive', $profile_json_section_key, '');

            \OCP\Util::writeLog('signlive', 'Using JSON section for automatic signature: '.$user_profile_json_section, ILogger::DEBUG);
    
            if ($profile_json_section != '') {
                $base64_encoded_document = $this->fileIO->getBase64EncodedFile($file_list_to_sign[0]);

                if ($base64_encoded_document != '') {
                    $profile_data = [
                        'file_name' => $file_name,
                        'redirect_url' => $redirect_url,
                        'json_section' => $profile_json_section,
                        'signature_type' => $signature_type, 
                        'base64_encoded_document' => $base64_encoded_document
                    ];

                    // Get the JSON request for the document to sign
                    $json_for_signature_request = $this->jsonQueryCreator->getJSONQuery($file_list_to_sign[0], $profile_data, false);
                    
                    \OCP\Util::writeLog('signlive', 'signature_type = '.$signature_type, ILogger::DEBUG);
                    \OCP\Util::writeLog('signlive', 'Using JSON section for profile-based signature: '.$user_profile_json_section, ILogger::DEBUG);
                
                    if ($json_for_signature_request != '') {
                        $json_for_signature_request_for_logging = $this->jsonQueryCreator->getJSONQuery($file_list_to_sign[0], $profile_data, true);
                        
                        // \OCP\Util::writeLog('signlive', 'JSON section (shortened) with replaced variable values: '.$json_for_signature_request_for_logging, ILogger::DEBUG);

                        // Send the JSON request to the Gears backend, catching any connection-related errors that may occur
                        try {
                            // At this point a valid JSON request for the Gears backend is present: Create a signature request job record here.
                            $signature_job_data['signature_job_uuid'] = $job_uuid;
                            $signature_job_data['signature_job_name'] = 'Signature job \''.$file_name.'\''; 
                            $signature_job_data['signature_job_type'] = 'automatic_signature';
                            $signature_job_data['signature_job_file'] = $file_name;
                            $signature_job_data['signature_profile_type'] = $profile_type;
                            $signature_job_data['nextcloud_user'] = $this->userSession->getUser()->getUid();
                            $signature_job_data['job_creation_time'] = time();
                            
                            // Set initial state of new signature request job at creation time
                            $job_record_data['job_stage'] = 'STARTED';

                            $this->logSignatureRequest($job_uuid, $signature_job_data, $json_for_signature_request_for_logging);
         
                            // Persist this signature request for job management and notifications
                            $job_def_created = $this->createJobDefinitionEntry($signature_job_data);

                            if ($job_def_created === false) {
                                \OCP\Util::writeLog('signlive', 'Creation of job record for signature request job '.$job_uuid.' failed! Can not continue.', ILogger::WARN);
                                return new DataResponse([
                                    'message' => 'Creation of job record for signature request job '.$job_uuid.' failed! Can not continue.',
                                    'signature_type' => 'automatic_signature',
                                    'success' => false
                                ]);
                            }

                            \OCP\Util::writeLog('signlive', 'Using Gears backend URL: '.$this->gears_backend_url, ILogger::DEBUG);
                            $gears_response = $this->gears_rest_client->post($this->gears_backend_url.'/api/v1/flow/signer/create', [
                                'headers' => ['Content-Type' => 'application/json', 'Accept' => 'application/json'],
                                'body'     => $json_for_signature_request
                            ]);

                            // Request to create signer flow has been sent, update job state
                            $this->setNewJobState($job_uuid, 'FLOW_CREATION_REQUEST_SENT', 'Initial request for signer flow creation sent for job '.$job_uuid.'.', false);

                            $gears_response_code = $gears_response->getStatusCode();
                            $gears_response_reason = $gears_response->getReasonPhrase();

                            \OCP\Util::writeLog('signlive', 'HTTP response from Gears backend: '.intval($gears_response_code).' ('.$gears_response_reason.')', ILogger::DEBUG);

                             // Response for viewer flow creation has been received, update job state
                             $this->setNewJobState($job_uuid, 'FLOW_CREATION_RESPONSE_RECEIVED', 'Signer flow has been created for job .'.$job_uuid.'.', false);

                            $json_response_array = json_decode($gears_response->getBody());
                            
                            // This results in very long log message texts dumped to the Nextcloud log
                            // In contrast to requests base64-encoded documents can not easily replaced in JSON responses
                            // So leave logging the response commented out
                            // \OCP\Util::writeLog('signlive', 'Decoded JSON response from Gears backend: '.print_r($json_response_array, TRUE), ILogger::DEBUG);
                        
                            // We expect a 2xx HTTP response code to a successful signature flow creation
                            if (intval($gears_response_code) >= 200 && intval($gears_response_code) <= 299) {
                                if (property_exists($json_response_array, 'snapshot')) {
                                    \OCP\Util::writeLog('signlive', 'Found \'snapshot\' element in JSON response.', ILogger::DEBUG);
                                    if (property_exists($json_response_array->snapshot, 'stage')) {
                                        \OCP\Util::writeLog('signlive', 'Found \'stage\' element in JSON response.', ILogger::DEBUG);
                                        if (property_exists($json_response_array->snapshot->stage, 'scheme')) {
                                            \OCP\Util::writeLog('signlive', 'Found \'scheme\' element in JSON response.', ILogger::DEBUG);
                                            $scheme_value = $json_response_array->snapshot->stage->scheme;
                                            
                                            // TODO: Adapt to polling (out of band) signature requests
                                            if (!empty($scheme_value)) {
                                                \OCP\Util::writeLog('signlive', '\'scheme\' value in JSON response: '.$scheme_value, ILogger::DEBUG);
                                                if ($scheme_value === 'urn:intarsys:names:conversation:1.0:schemes:Error') {
                                                    if (property_exists($json_response_array->snapshot->stage, 'error')) {
                                                        $error_value = print_r(json_encode($json_response_array->snapshot->stage), true);

                                                        \OCP\Util::writeLog('signlive', 'An error occurred while processing a signer flow creation request: \''.$error_value.'\'', ILogger::ERROR);

                                                        $status_message = 'An error occurred while processing a signer flow creation request: \''.$error_value.'\' in signature request job '.$job_uuid.'!';
                                                        // Invalid/error response received for signer flow creation, update job state
                                                        $this->setNewJobState($job_uuid, 'FLOW_CREATION_RESPONSE_INVALID', $status_message, true);

                                                        $job_record_data['status_message'] = $status_message;
                                                        $this->sendFailureNotification($job_record_data);

                                                        return new DataResponse([
                                                            'message' => $status_message,
                                                            'signature_type' => 'automatic_signature',
                                                            'success' => false
                                                        ]);
                                                    } else {
                                                        \OCP\Util::writeLog('An unspecified error occurred in the Gears backend while processing an automatic_signature-type signature request!', ILogger::ERROR);
                                                        
                                                        // As-of-yet unhandled error condition occurred in signer flow creation, update job state
                                                        $this->setNewJobState($job_uuid, 'FLOW_CREATION_RESPONSE_INVALID', 'As-of-yet unhandled response received for signer flow creation in signature request job '.$job_uuid.'!', true);

                                                        $job_record_data['status_message'] = 'As-of-yet unhandled error condition occurred for signer flow creation in signature request job '.$job_uuid.'!';
                                                        $this->sendFailureNotification($job_record_data);

                                                        return new DataResponse([
                                                            'message' => 'As-of-yet unhandled error condition occurred for signer flow creation in signature request job '.$job_uuid.'!',
                                                            'signature_type' => 'automatic_signature',
                                                            'success' => false
                                                        ]);
                                                    }
                                                } else {
                                                    // 'urn:intarsys:names:conversation:1.0:schemes:Result' is handled further below
                                                    if ($scheme_value !== 'urn:intarsys:names:conversation:1.0:schemes:Result') {
                                                        // TODO: Handle other 'scheme' values.
                                                        \OCP\Util::writeLog('signlive', 'Unhandled \'scheme\' in signer flow creation request: \''.$scheme_value.'\' in signature request job '.$job_uuid.'!', ILogger::DEBUG);

                                                        // As-of-yet unhandled response received for signer flow creation, update job state
                                                        $this->setNewJobState($job_uuid, 'FLOW_CREATION_RESPONSE_UNHANDLED_SCHEME', 'Unhandled \'scheme\' in signer flow creation request: \''.$scheme_value.'\' in signature request job '.$job_uuid.'!', true);
                                                    
                                                        $job_record_data['status_message'] = 'Unhandled \'scheme\' in signer flow creation request: \''.$scheme_value.'\' in signature request job '.$job_uuid.'!';
                                                        $this->sendFailureNotification($job_record_data);

                                                        return new DataResponse([
                                                            'message' => 'Unhandled \'scheme\' in signer flow creation request: \''.$scheme_value.'\' in signature request job '.$job_uuid.'!',
                                                            'signature_type' => 'automatic_signature',
                                                            'success' => false
                                                        ]);
                                                    }
                                                }
                                            } else {
                                                \OCP\Util::writeLog('signlive', 'Empty \'scheme\' value in JSON response for signature request job '.$job_uuid.'!', ILogger::ERROR);
                                                
                                                $this->setNewJobState($job_uuid, 'RESULT_RETRIEVAL_FAILURE_EMPTY', 'Empty \'scheme\' value in JSON response for signature request job '.$job_uuid.'!', true);
                                                
                                                $job_record_data['status_message'] = 'Empty \'scheme\' value in JSON response for signature request job '.$job_uuid.'!';
                                                $this->sendFailureNotification($job_record_data);

                                                return new DataResponse([
                                                    'message' => 'Empty \'scheme\' value in JSON response for signature request job '.$job_uuid.'!',
                                                    'signature_type' => 'automatic_signature',
                                                    'success' => false
                                                ]);
                                            }
                                        }

                                        // No error for in-band signature request found, update job state
                                        // TODO: Adapt to polling (out of band) signature requests
                                        $this->setNewJobState($job_uuid, 'FLOW_CREATION_RESPONSE_VALID', 'Valid redirect URL for viewer flow has been provided for job \''.$job_uuid.'\'.', false);

                                        if (property_exists($json_response_array->snapshot->stage, 'result')) {
                                            \OCP\Util::writeLog('signlive', 'Found \'result\' element in JSON response.', ILogger::DEBUG);
                                            if (property_exists($json_response_array->snapshot->stage->result, 'value')) {
                                                \OCP\Util::writeLog('signlive', 'Found \'value\' element in JSON response.', ILogger::DEBUG);
                                                if (property_exists($json_response_array->snapshot->stage->result->value, 'signatures')) {
                                                    \OCP\Util::writeLog('signlive', 'Found \'signatures\' element in JSON response.', ILogger::DEBUG);
                                                    $signatures_array = json_decode(json_encode($json_response_array->snapshot->stage->result->value->signatures), true);
                                                    \OCP\Util::writeLog('signlive', 'Signatures element count: '.count($signatures_array), ILogger::DEBUG);
                                                    if (count($signatures_array) > 0) {
                                                        $current_user_folder = $this->storage->getUserFolder($this->userSession->getUser()->getUid());
                                                        
                                                        foreach ($signatures_array as $signature) {
                                                            \OCP\Util::writeLog('signlive', 'Signed document: '.$signature['name'], ILogger::DEBUG);

                                                            $base64_decoded_signed_content = base64_decode($signature['content']);
                                                        
                                                            $file_name_to_save = str_replace(":::", "/", $signature['name']);
                                                            \OCP\Util::writeLog('signlive', 'Document \''.$file_name.'\' signed successfully with automatic signature request job '.$job_uuid.'.', ILogger::INFO);

                                                            $current_user_folder->newFile($file_name_to_save, $base64_decoded_signed_content);

                                                            // At this point a valid signature result has been retrieved, update job state accordingly
                                                            $this->setNewJobState($job_uuid, 'RESULT_RETRIEVAL_SUCCEEDED', 'Document \''.$file_name.'\' signed successfully with automatic signature request job '.$job_uuid.'.', false);
                                                            
                                                            $signature_job_data['status_message'] = 'Document \''.$file_name.'\' signed successfully with automatic signature request job '.$job_uuid.'.';
                                                            $this->sendSuccessNotification($signature_job_data);

                                                            return new DataResponse([
                                                                'message' => 'Document \''.$file_name.'\' signed successfully with automatic signature request job '.$job_uuid.'.',
                                                                'success' => true,
                                                                'signature_type' => 'automatic_signature'
                                                            ]);
                                                        }
                                                    }
                                                }
                                            }
                                        }    
                                    }
                                }
                            }
                        } catch (RequestException $e) {
                            $status_message = 'Error occurred while sending signature request to Gears backend in signature request job '.$job_uuid.': '.Psr7\Message::toString($e->getRequest());
                            \OCP\Util::writeLog('signlive', $status_message, ILogger::ERROR);
                            if ($e->hasResponse()) {
                                \OCP\Util::writeLog('signlive', 'Response data returned by Gears backend: '.Psr7\Message::toString($e->getResponse()), ILogger::ERROR);
                                $status_message = $status_message.'Response data returned by Gears backend: '.Psr7\Message::toString($e->getResponse());
                            }

                            // At this point a valid signature result has been retrieved, update job state accordingly
                            $this->setNewJobState($job_uuid, 'RESULT_RETRIEVAL_COMMUNICATION_FAILURE', $status_message, true);

                            $job_record_data['status_message'] = $status_message;
                            $this->sendFailureNotification($job_record_data);

                            return new DataResponse([
                                'message' => $status_message,
                                'success' => false,
                                'signature_type' => 'automatic_signature'
                            ]);
                        } catch (ClientException $e) {
                            $status_message = 'Client-side rror occurred while processing the signature request in signature request job '.$job_uuid.'! Request data: '.Psr7\Message::toString($e->getRequest());
                            $status_message = $status_message.'Response data: '.Psr7\Message::toString($e->getResponse());
                            \OCP\Util::writeLog('signlive', $status_message, ILogger::ERROR);
                            
                            $this->setNewJobState($job_uuid, 'RESULT_RETRIEVAL_CLIENT_SIDE_FAILURE', $status_message, true);
                            $job_record_data['status_message'] = $status_message;
                            $this->sendFailureNotification($job_record_data);

                            return new DataResponse([
                                'message' => $status_message,
                                'success' => false,
                                'signature_type' => 'automatic_signature'
                            ]);
                        }
                    }
                } else {
                    \OCP\Util::writeLog('signlive', 'Failed to base64-encode file to sign: '.$file_to_sign[0].' in signature request job '.$job_uuid.'!', ILogger::ERROR);
                
                    // Could not read and base64-encode file to sign, update job state accordingly
                    $this->setNewJobState($job_uuid, 'FLOW_CREATION_FILE_NOT_FOUND', $status_message, true);

                    $job_record_data['status_message'] = 'Failed to base64-encode file to sign: '.$file_to_sign[0].' in signature request job '.$job_uuid.'!';
                    $this->sendFailureNotification($job_record_data);
                    
                    return new DataResponse([
                        'message' => 'Failed to base64-encode file to sign: '.$file_to_sign[0].' in signature request job '.$job_uuid.'!',
                        'success' => false,
                        'signature_type' => 'automatic_signature'
                    ]);
                }
            } else {
                \OCP\Util::writeLog('signlive', 'Found no valid JSON section for profile unique ID \''.$profile_shortened_unique_id.'\' of type \''.$profile_type.'\'. Could not sign document '.$file_to_sign[0].'!', ILogger::ERROR);
                
                // No valid JSON section could be found for the given profile, update job state
                $this->setNewJobState($job_uuid, 'FLOW_CREATION_JSON_PROCESSING_FAILED', 'Found no valid JSON section for profile \''.$profile_shortened_unique_id.'\' of type \''.$profile_type.'\'. Could not sign document '.$file_to_sign[0].' in job '.$job_uuid.'!', true);
                            
                $job_record_data['status_message'] = 'Found no valid JSON section for profile \''.$profile_shortened_unique_id.'\' of type \''.$profile_type.'\'. Could not sign document '.$file_to_sign[0].' in job '.$job_uuid.'!';
                $this->sendFailureNotification($job_record_data);

                return new DataResponse([
                    'message' => 'Found no valid JSON section for profile unique ID \''.$profile_shortened_unique_id.'\' of type \''.$profile_type.'\'. Could not sign document '.$file_to_sign[0].'!',
                    'success' => false,
                    'signature_type' => 'automatic_signature'
                ]);
            }
        } else {
            \OCP\Util::writeLog('signlive', 'signDocumentViaProfileJob() will only accept a single document to sign! Doing nothing.', ILogger::ERROR);
            
            // Only a single document is supported per signature request job, update job state
            $this->setNewJobState($job_uuid, 'FLOW_CREATION_GENERAL_FAILURE', 'signDocumentViaProfileJob() will only accept a single document to sign in job '.$job_uuid.'!', true);
                        
            $job_record_data['status_message'] = 'signDocumentViaProfileJob() will only accept a single document to sign in job '.$job_uuid.'!';
            $this->sendFailureNotification($job_record_data);
            
            return new DataResponse([
                'message' => 'signDocumentViaProfileJob() will only accept a single document to sign! Doing nothing.',
                'success' => false,
                'signature_type' => 'automatic_signature'
            ]);
        }
    }
}