<?php

/**
* This file is licensed under the Affero General Public License version 3 or
* later. See the COPYING file.
*
* @author Fabian Aichele
* @copyright intarsys GmbH 2021
*/

namespace OCA\IntarsysCloudSuiteGearsIntegration\Controller;

use OCP\AppFramework\Controller;
use OCP\AppFramework\Http\DataResponse;
use OCP\IConfig;
use OCP\IL10N;
use OCP\IRequest;
use OCP\IURLGenerator;
use OCP\IGroupManager;
use OCP\IUserManager;
use OCP\IUserSession;
use OCP\Security\ISecureRandom;
use OCP\Util;
use OCP\ILogger;

use OCA\IntarsysCloudSuiteGearsIntegration\Util\VariableSubstitution;
use OCA\IntarsysCloudSuiteGearsIntegration\Util\SignLiveDefaultConfigValues;
use OCA\IntarsysCloudSuiteGearsIntegration\Util\SignatureProfilesSetup;

class IntarsysSettingsController extends Controller {

	/** @var \OCP\IConfig */
	protected $config;

	/** @var \OCP\Security\ISecureRandom */
	protected $random;

	/** @var \OCP\IURLGenerator */
	protected $urlGenerator;

	/** @var \OCP\IL10N */
	protected $l10n;

	/** @var string */
	protected $user;

	/** @var IGroupManager */
	protected $groupManager;

	/** @var IUserManager */
	protected $userManager;

	/** @var IUserSession */
	protected $userSession;

	/** @var array */
	protected $required_profile_properties;

	/** @var VariableSubstitution */
	protected $variableSubstitution;

	/**
	 * @param string $appName
	 * @param IRequest $request
	 * @param IL10N $l10n
	 * @param IConfig $config
	 * @param IUserSession $userSession
	 */
	public function __construct(
								$appName,
								IRequest $request,
								IConfig $config,
								ISecureRandom $random,
								IURLGenerator $urlGenerator,
								IGroupManager $groupManager,
								IUserManager $userManager,
								IUserSession $userSession,
								IL10N $l10n,
								VariableSubstitution $variableSubstitution) {
		parent::__construct($appName, $request);
		$this->config = $config;
		$this->random = $random;
		$this->urlGenerator = $urlGenerator;
		$this->groupManager = $groupManager;
		$this->userManager = $userManager;
		$this->userSession = $userSession;
		$this->l10n = $l10n;
		$this->variableSubstitution = $variableSubstitution;

		$this->uid = $this->userSession->getUser()->getUID();
        $this->user = $this->userManager->get($this->uid);

		$this->required_profile_properties = ['cloudsuite_gears_profile_id',
			'cloudsuite_gears_profile_name',
			'cloudsuite_gears_profile_type',
			'cloudsuite_gears_profile_json_1',
			'cloudsuite_gears_default_signature_profile',
			'cloudsuite_gears_nextcloud_groups',
			'cloudsuite_gears_signature_profile_menu_priority'
			];
	}

	/**
	 * @NoAdminRequired
	 * 
	 * Returns the number of defined signature profiles in the Nextcloud DB
	 * 
	 * @return int
	 */
	private function getNumSignatureProfiles() {
		$num_signature_profiles = 0;
		$app_config_keys = $this->config->getAppKeys('signlive');

		foreach ($app_config_keys as $app_key_id => $app_key) {
			if (str_starts_with($app_key, "cloudsuite_gears_signature_profile_id_")) {
				$num_signature_profiles += 1;
			}
		}
		\OCP\Util::writeLog('signlive', 'num_signature_profiles = '.$num_signature_profiles, ILogger::DEBUG);

		return $num_signature_profiles;
	}

	/**
	 * @NoAdminRequired
	 * 
	 * Returns the number of defined signature profiles in the Nextcloud DB as REST-compatible DataResponse
	 * @return DataResponse
	 */
	public function getSignatureProfileCount() {
		$num_signature_profiles = $this->getNumSignatureProfiles();
		\OCP\Util::writeLog('signlive', 'getSignatureProfileCount(): '.$num_signature_profiles, ILogger::DEBUG);
		return new DataResponse([
			'num_signature_profiles' => $num_signature_profiles
		]);
	}

	/**
	 * Sets the Intarsys signature server settings
	 *
	 * @param string $cloudsuite_gears_http_protocol_type
	 * @param string $cloudsuite_gears_host
	 * @param int $cloudsuite_gears_port
	 * @param string $cloudsuite_gears_unsigned_pdf_documents_folder
	 * @param string $cloudsuite_gears_signed_pdf_documents_folder
	 * @return DataResponse
	 */
	public function setSignatureServerSettings($cloudsuite_gears_http_protocol_type = 'https',
											   $cloudsuite_gears_host = 'cloudsuite.intarsys.de',
											   $cloudsuite_gears_port = 443,
											   $cloudsuite_gears_unsigned_pdf_documents_folder = '.intarsys/cloudsuite_gears/unsigned_documents',
											   $cloudsuite_gears_signed_pdf_documents_folder = '.intarsys/cloudsuite_gears/unsigned_documents') {							
		\OCP\Util::writeLog('signlive', 'setSignatureServerSettings() called in signlive/lib/Controller/IntarsysSettingsController.php.', ILogger::DEBUG);
		\OCP\Util::writeLog('signlive', 'setSignatureServerSettings() request parameters: '.print_r($this->request->getParams(), TRUE), ILogger::DEBUG);

		$num_profiles = $this->getNumSignatureProfiles();
		$request_parameters = $this->request->getParams();
		\OCP\Util::writeLog('signlive', 'setSignatureServerSettings() called in signlive/lib/Controller/IntarsysSettingsController.php.', ILogger::DEBUG);
		foreach ($request_parameters as $parameter_name => $parameter_value) {
			$value_to_store = $parameter_value;
			$store_value = false;
			if ($parameter_name === 'cloudsuite_gears_http_protocol_type') {
				$store_value = true;
				if (empty($parameter_value)) {
					$value_to_store = 'https';
				}
			}
			if ($parameter_name === 'cloudsuite_gears_host') {
				$store_value = true;
				if (empty($parameter_value)) {
					$value_to_store = 'cloudsuite.intarsys.de';
				}
			}
			if ($parameter_name === 'cloudsuite_gears_port') { 
				$store_value = true;
				if (empty($parameter_value)) {
					$value_to_store = '443';
				}
			}
			if ($parameter_name === 'cloudsuite_gears_gears_core_path') {
				$store_value = true;
				if (empty($parameter_value)) {
					$value_to_store = '/cloudsuite-gears/core';
				}
			}

			if ($parameter_name === 'cloudsuite_gears_show_default_signature_entries_in_context_menu') {
				$store_value = true;
				if (empty($parameter_value)) {
					$value_to_store = '1';
				}
			}

			if ($parameter_name === 'cloudsuite_gears_signature_profile_default_menu_priority') {
				$store_value = true;
				if (empty($parameter_value)) {
					$value_to_store = 75;
				}
			}

			if ($parameter_name === 'cloudsuite_gears_global_attribute_company_name') {
				$store_value = true;
				if (empty($parameter_value)) {
					$value_to_store = '';
				}
			}

			if ($parameter_name === 'cloudsuite_gears_notification_duration') {
				$store_value = true;
				if (empty($parameter_value)) {
					$value_to_store = 7;
				}
			}

			if ($parameter_name === 'cloudsuite_gears_notification_check_interval') {
				$store_value = true;
				if (empty($parameter_value)) {
					$value_to_store = 1;
				}
			}

			if ($parameter_name === 'cloudsuite_gears_show_success_notifications') {
				$store_value = true;
				if (empty($parameter_value)) {
					$value_to_store = 0;
				}
			}

			if ($parameter_name === 'cloudsuite_gears_show_verbose_failure_notifications') {
				$store_value = true;
				if (empty($parameter_value)) {
					$value_to_store = 0;
				}
			}

			if ($parameter_name === 'cloudsuite_gears_display_menu_priorities') {
				$store_value = true;
				if (empty($parameter_value)) {
					$value_to_store = 0;
				}
			}

			if ($store_value == true) {
				\OCP\Util::writeLog('signlive', 'Storing parameter: \''.$parameter_name.'\' => \''.$value_to_store.'\'', ILogger::DEBUG);

				$this->config->setAppValue(
					'signlive',
					$parameter_name,
					$value_to_store);
			}
		}

		return new DataResponse([
			'data' => [
				'message' => $this->l10n->t('Intarsys CloudSuite Gears integration app settings have been updated.'),
			],
		]);
	}

	/**
	 * Sets the SignLive app settings for the currently logged-in Nextcloud user
	 *
	 * @NoAdminRequired
	 * @return DataResponse
	 */

	public function setPersonalSettings() {
		\OCP\Util::writeLog('signlive', 'setPersonalSettings() called in signlive/lib/Controller/IntarsysSettingsController.php.', ILogger::DEBUG);
		\OCP\Util::writeLog('signlive', 'setPersonalSettings() request parameters count: '.count($this->request->getParams()), ILogger::DEBUG);
		
		$request_parameters = $this->request->getParams();

		$user_attribute_title = '';
		if (key_exists('cloudsuite_gears_user_attribute_title', $request_parameters)) {
			$user_attribute_title = $request_parameters['cloudsuite_gears_user_attribute_title'];
		}

		$user_attribute_given_names = '';
		if (key_exists('cloudsuite_gears_user_attribute_given_names', $request_parameters)) {
			$user_attribute_given_names = $request_parameters['cloudsuite_gears_user_attribute_given_names'];
		}

		$user_attribute_surname = '';
		if (key_exists('cloudsuite_gears_user_attribute_surname', $request_parameters)) {
			$user_attribute_surname = $request_parameters['cloudsuite_gears_user_attribute_surname'];
		}

		$user_attribute_msisdn = '';
		if (key_exists('cloudsuite_gears_user_attribute_msisdn', $request_parameters)) {
			$user_attribute_msisdn = $request_parameters['cloudsuite_gears_user_attribute_msisdn'];
		}

		$user_attribute_country = '';
		if (key_exists('cloudsuite_gears_user_attribute_country', $request_parameters)) {
			$user_attribute_country = $request_parameters['cloudsuite_gears_user_attribute_country'];
		}

		$user_attribute_signme_user_name = '';
		if (key_exists('cloudsuite_gears_user_attribute_signme_user_name', $request_parameters)) {
			$user_attribute_signme_user_name = $request_parameters['cloudsuite_gears_user_attribute_signme_user_name'];
		}

		$user_attribute_encoded_signature_image = '';
		if (key_exists('cloudsuite_gears_user_attribute_encoded_signature_image', $request_parameters)) {
			$user_attribute_encoded_signature_image = $request_parameters['cloudsuite_gears_user_attribute_encoded_signature_image'];
		}

		$user_attribute_company_division = '';
		if (key_exists('cloudsuite_gears_user_attribute_company_division', $request_parameters)) {
			$user_attribute_company_division = $request_parameters['cloudsuite_gears_user_attribute_company_division'];
		}

		$this->config->setUserValue($this->uid,
						'signlive',
						'cloudsuite_gears_user_attribute_title',
						$user_attribute_title);

		$this->config->setUserValue($this->uid,
						'signlive',
						'cloudsuite_gears_user_attribute_given_names',
						$user_attribute_given_names);
		
		$this->config->setUserValue($this->uid,
						'signlive',
						'cloudsuite_gears_user_attribute_surname',
						$user_attribute_surname);

		$this->config->setUserValue($this->uid,
						'signlive',
						'cloudsuite_gears_user_attribute_msisdn',
						$user_attribute_msisdn);

		$this->config->setUserValue($this->uid,
						'signlive',
						'cloudsuite_gears_user_attribute_country',
						$user_attribute_country);
		
		$this->config->setUserValue($this->uid,
						'signlive',
						'cloudsuite_gears_user_attribute_signme_user_name',
						$user_attribute_signme_user_name);

		$this->config->setUserValue($this->uid,
						'signlive',
						'cloudsuite_gears_user_attribute_encoded_signature_image',
						$user_attribute_encoded_signature_image);

		$this->config->setUserValue($this->uid,
						'signlive',
						'cloudsuite_gears_user_attribute_company_division',
						$user_attribute_company_division);	

		return new DataResponse([
			'data' => [
				'message' => $this->l10n->t('Intarsys CloudSuite Gears integration user settings have been updated.'),
			],
		]);
	}

	/**
	 * Retrieve list of signature profile definitions with their properties 
	 *
	 * @param bool $filter_public_profiles
	 * 
	 * @NoAdminRequired
	 * @return DataResponse
	 */
	private function getProfileList($filter_public_profiles) {
		if ($filter_public_profiles !== true && $filter_public_profiles !== false) {
			return false;
		}

		$public_signature_profiles = [];
		$restricted_signature_profiles = [];
		$num_profiles = 0;
		$num_public_profiles = 0;
		$num_restricted_profiles = 0;

		$menu_entries_priority_value = $this->config->getAppValue('signlive', 'cloudsuite_gears_signature_profile_menu_priority', SignLiveDefaultConfigValues::$DEFAULT_MENU_PRIORITY_VALUE);

		$app_config_keys = $this->config->getAppKeys('signlive');
		foreach ($app_config_keys as $config_key_id => $config_key) {
			if (str_starts_with($config_key, 'cloudsuite_gears_profile_id_')) {
				$unique_profile_key = str_replace('cloudsuite_gears_profile_id_', '', $config_key);
				\OCP\Util::writeLog('signlive', 'Signature profile found: \''.$unique_profile_key.'\'', ILogger::DEBUG);
				$profile_name_value = $this->config->getAppValue('signlive',
															  	 'cloudsuite_gears_profile_name_'.$unique_profile_key, '');
				$profile_type_value = $this->config->getAppValue('signlive',
															  	 'cloudsuite_gears_profile_type_'.$unique_profile_key, '');
				$profile_json_section_value = $this->config->getAppValue('signlive',
															  	 'cloudsuite_gears_profile_json_'.$unique_profile_key, '');
				$default_profile_value = $this->config->getAppValue('signlive',
															  	    'cloudsuite_gears_default_profile_'.$unique_profile_key, '');
				$nextcloud_groups_value = $this->config->getAppValue('signlive',
															  	    'cloudsuite_gears_profile_nextcloud_groups_'.$unique_profile_key, '');
				$nextcloud_menu_priority_value = $this->config->getAppValue('signlive',
															  	    'cloudsuite_gears_profile_menu_priority_'.$unique_profile_key, SignLiveDefaultConfigValues::$DEFAULT_MENU_PRIORITY_VALUE);
																	
				\OCP\Util::writeLog('signlive', 'Signature profile \''.$unique_profile_key.'\' default_profile_value = \''.$default_profile_value.'\'', ILogger::DEBUG);																	  
				
				if ($default_profile_value === true || 
					$default_profile_value === "true" || 
					$default_profile_value === "1" ||
					intval($default_profile_value) === 1) {
					\OCP\Util::writeLog('signlive', 'Adding default (public) signature profile: '.$unique_profile_key, ILogger::DEBUG);
					$public_signature_profiles['cloudsuite_gears_profile_id_'.$unique_profile_key] = [
						'profile_id' => $config_key,
						'profile_name' => $profile_name_value,
						'profile_type' => $profile_type_value,
						'profile_json_section' => $profile_json_section_value,
						'profile_default_profile' => $default_profile_value,
						'profile_nextcloud_groups' => $nextcloud_groups_value,
						'profile_menu_priority' => $nextcloud_menu_priority_value
					];
					$num_profiles += 1;
					$num_public_profiles += 1;
				} else if ($default_profile_value === false || 
						   $default_profile_value === "false" || 
						   $default_profile_value === "0" ||
						   intval($default_profile_value) === 0) {
					\OCP\Util::writeLog('signlive', 'Adding restricted (visible for certain groups only) signature profile: '.$unique_profile_key, ILogger::DEBUG);
					$restricted_signature_profiles['cloudsuite_gears_profile_id_'.$unique_profile_key] = [
						'profile_id' => $config_key,
						'profile_name' => $profile_name_value,
						'profile_type' => $profile_type_value,
						'profile_json_section' => $profile_json_section_value,
						'profile_default_profile' => $default_profile_value,
						'profile_nextcloud_groups' => $nextcloud_groups_value,
						'profile_menu_priority' => $nextcloud_menu_priority_value
					];
					$num_profiles += 1;
					$num_restricted_profiles += 1;
				} else {
					\OCP\Util::writeLog('signlive', 'Signature profile \''.$unique_profile_key.'\' default_profile_value is EMPTY?', ILogger::DEBUG);
				}
			}
		}
		$data_response = [];

		if ($filter_public_profiles === true) {
			$show_public_profiles_in_context_menu = $this->config->getAppValue('signlive', 'cloudsuite_gears_show_default_signature_entries_in_context_menu', false);
			$show_public_profiles_in_context_menu_value = 0;
			if ($show_public_profiles_in_context_menu === true ||
				$show_public_profiles_in_context_menu === "true") {
				$show_public_profiles_in_context_menu_value = 1;
			}
			return ([
					'data' => [
						'show_public_profiles_in_context_menu' => $show_public_profiles_in_context_menu_value,
						'num_public_profiles' => $num_public_profiles,
						'num_profiles' => $num_profiles,
						'menu_entry_priority_value' => $menu_entries_priority_value,
						'public_signature_profiles' => $public_signature_profiles
					]
				]);
		} else {
			return ([
				'data' => [
					'num_restricted_profiles' => $num_restricted_profiles,
					'num_profiles' => $num_profiles,
					'menu_entry_priority_value' => $menu_entries_priority_value,
					'restricted_signature_profiles' => $restricted_signature_profiles
				]
			]);
		}

		return $data_response; 
	}

	/**
	 * @NoAdminRequired
	 *
	 * Get the list of public signature profiles 
	 * (usable for all Nextcloud users)
	 *
	 * @return DataResponse
	 */
	public function getPublicSignatureProfiles() {
		$data_response = $this->getProfileList(true);
		
		\OCP\Util::writeLog('signlive', 'getPublicSignatureProfiles() called.', ILogger::DEBUG);
	
		return new DataResponse($data_response);
	}

	/**
	 * @NoAdminRequired
	 *
	 * Get the list of restricted signature profiles 
	 * (usable only for Nextcloud users who are member of associated user groups)
	 *
	 * @return DataResponse
	 */
	public function getRestrictedSignatureProfiles() {
		$data_response = $this->getProfileList(false);

		\OCP\Util::writeLog('signlive', 'getRestrictedSignatureProfiles() called.', ILogger::DEBUG);
		
		return new DataResponse($data_response);
	}

	/**
	 * @NoAdminRequired
	 *
	 * Get all signature Nextcloud user group names
	 *
	 * @return DataResponse
	 */
	public function getAllNextcloudGroupNames() {
		$nextcloud_groups = $this->groupManager->search('');
		$nextcloud_group_names = [];
		$num_nextcloud_groups = 0;
		foreach ($nextcloud_groups as $nextcloud_group) {
			$nextcloud_group_names[$nextcloud_group->getGID()] =  $nextcloud_group->getDisplayName();
			$num_nextcloud_groups += 1;
		}

		return new DataResponse([
			'num_nextcloud_groups' => $num_nextcloud_groups,
			'nextcloud_group_names' => $nextcloud_group_names
		]);
	}

	/**
	 * @NoAdminRequired
	 *
	 * Get all defined signature profiles
	 *
	 * @return DataResponse
	 */
	public function getAllSignatureProfiles() {
		$data_response_public_profiles = $this->getProfileList(true);
		$data_response_restricted_profiles = $this->getProfileList(false);

		\OCP\Util::writeLog('signlive', 'getAllSignatureProfiles() called.', ILogger::DEBUG);
		
		$all_profiles = array_merge($data_response_restricted_profiles['data']['restricted_signature_profiles'], $data_response_public_profiles['data']['public_signature_profiles']);
		
		// Add further profile-related configuration values for admin settings JS script
		$all_profiles['menu_entry_priority_value'] = $this->config->getAppValue('signlive', 'cloudsuite_gears_signature_profile_menu_priority', SignLiveDefaultConfigValues::$DEFAULT_MENU_PRIORITY_VALUE);

		$current_source_file = __FILE__;
		$document_root = $_SERVER['DOCUMENT_ROOT'];
		\OCP\Util::writeLog('signlive', 'Request path for settings controller: \''.$current_source_file.'\', document root: \''.$document_root.'\'', ILogger::DEBUG);
		$request_uri = str_replace($document_root, '', $current_source_file);
		$relative_app_path = explode('signlive', $request_uri);
		\OCP\Util::writeLog('signlive', 'Relative app path: \''.print_r($relative_app_path, TRUE), ILogger::DEBUG);
		$all_profiles['relative_app_path'] = $relative_app_path[0];

		return new DataResponse($all_profiles);
	}

	/**
	 * @NoAdminRequired
	 *
	 * Get all signature profiles the current Nextcloud user may use
	 *
	 * @return DataResponse
	 */
	public function getSignatureProfilesForCurrentUser() {
		\OCP\Util::writeLog('signlive', '======================================================================================================', ILogger::DEBUG);
		\OCP\Util::writeLog('signlive', 'getSignatureProfilesForCurrentUser() called in signlive/lib/Controller/IntarsysSettingsController.php.', ILogger::DEBUG);
		
		// Duplicate the layout of the DataResponse as in getAllSignatureProfiles()
		$signature_profile_list = [];
		$signature_profile_list['data'] = [];
		$signature_profile_list['data']['restricted_signature_profiles'] = [];
		$signature_profile_list['data']['public_signature_profiles'] = [];

		$user_groups = $this->groupManager->getUserGroups($this->user);
		\OCP\Util::writeLog('signlive', 'Current user \''.$this->user->getUid().'\' Nextcloud groups count: '.count($user_groups), ILogger::DEBUG);

		$public_profiles = $this->getProfileList(true);
		$restricted_profiles = $this->getProfileList(false);

		$menu_entries_priority_value = $this->config->getAppValue('signlive', 'cloudsuite_gears_signature_profile_menu_priority', SignLiveDefaultConfigValues::$DEFAULT_MENU_PRIORITY_VALUE);
		$display_menu_priority_values = $this->config->getAppValue('signlive', 'cloudsuite_gears_display_menu_priorities', SignLiveDefaultConfigValues::$SHOW_PRIORITY_VALUES_IN_MENU);
		$cloudsuite_gears_notification_check_interval = $this->config->getAppValue('signlive', 'cloudsuite_gears_notification_check_interval', SignLiveDefaultConfigValues::$NOTIFICATION_CHECK_INTERVAL);
		
		\OCP\Util::writeLog('signlive', 'Public signature profiles count    : '.count($public_profiles['data']['public_signature_profiles']), ILogger::DEBUG);
		\OCP\Util::writeLog('signlive', 'Restricted signature profiles count: '.count($restricted_profiles['data']['restricted_signature_profiles']), ILogger::DEBUG);

		// First check all signature profiles that are not visible for all Nextcloud users if they are allowed for the current user
		$restricted_profile_data = $restricted_profiles['data']['restricted_signature_profiles'];
		$num_allowed_restricted_profiles = 0;
		foreach ($user_groups as $group_id => $group) {
			foreach ($restricted_profile_data as $profile_id => $profile_data) {
				// \OCP\Util::writeLog('signlive', 'Profile data for profile ID \''.$profile_id.'\': '.print_r($profile_data, TRUE), ILogger::DEBUG);
				$profile_groups = $profile_data['profile_nextcloud_groups'];
				$profile_name = $profile_data['profile_name'];
				\OCP\Util::writeLog('signlive', 'Groups allowed to use profile \''.$profile_id.'\': '.$profile_groups, ILogger::DEBUG);
				$valid_groups = explode(';;', $profile_groups);
				
				\OCP\Util::writeLog('signlive', 'Checking if user account \''.$this->user->getUid().'\' is allowed to use signature profile \''.$profile_id.'\'', ILogger::DEBUG);
				$profile_allowed_for_user = false;

				// No groups explicitly specified for profile: Deny access
				if (count($valid_groups) == 0) {
					\OCP\Util::writeLog('signlive', 'No groups for signature profile \''.$profile_id.'\' have been specified explicitly. Assuming access denied.', ILogger::DEBUG);
				} else {
					// Otherwise check if current user is member of at least one of the explicitly specified groups
					foreach ($valid_groups as $valid_group) {
						if (empty($valid_group))
							continue;

						\OCP\Util::writeLog('signlive', 'Compare group names: \''.$valid_group.'\' == \''.$group_id.'\'', ILogger::DEBUG);
						if ($valid_group == $group_id) {
							\OCP\Util::writeLog('signlive', 'Match found.', ILogger::DEBUG);
							$profile_allowed_for_user = true;
							break;
						}
					}
				}

				if ($profile_allowed_for_user) {
					$num_allowed_restricted_profiles += 1;
					\OCP\Util::writeLog('signlive', 'Adding restricted profile: \''.$profile_name.'\' (\''.$profile_id.'\') to list of allowed signature profiles for user \''.$this->user->getUid().'\'', ILogger::DEBUG);
					$signature_profile_list['data']['restricted_signature_profiles'][$profile_id] = $restricted_profile_data[$profile_id];
				}
			}
		}

		// Then append all public signature profiles to the list if there are any
		if (count($public_profiles['data']['public_signature_profiles']) > 0) {
			\OCP\Util::writeLog('signlive', 'Adding public signature profiles to DataResponse: '.count($public_profiles['data']['public_signature_profiles']), ILogger::DEBUG);
			$signature_profile_list['data']['public_signature_profiles'] = $public_profiles['data']['public_signature_profiles'];
		}

		// Finally copy remaining signature profile attributes to DataResponse
		$signature_profile_list['data']['show_public_profiles_in_context_menu'] = $public_profiles['data']['show_public_profiles_in_context_menu'];
		$signature_profile_list['data']['num_public_profiles'] = $public_profiles['data']['num_public_profiles'];
		$signature_profile_list['data']['num_profiles'] = $public_profiles['data']['num_profiles'];
		$signature_profile_list['data']['num_restricted_profiles'] = $num_allowed_restricted_profiles;
		$signature_profile_list['data']['menu_entries_priority_value'] = $menu_entries_priority_value;
		$signature_profile_list['data']['cloudsuite_gears_display_menu_priorities'] = $display_menu_priority_values;
		$signature_profile_list['data']['cloudsuite_gears_notification_check_interval'] = $cloudsuite_gears_notification_check_interval;

		$current_source_file = __FILE__;
		$document_root = $_SERVER['DOCUMENT_ROOT'];
		\OCP\Util::writeLog('signlive', 'Request path for settings controller: \''.$current_source_file.'\', document root: \''.$document_root.'\'', ILogger::DEBUG);
		$request_uri = str_replace($document_root, '', $current_source_file);
		$relative_app_path = explode('signlive', $request_uri);
		\OCP\Util::writeLog('signlive', 'Relative app path: \''.print_r($relative_app_path, TRUE), ILogger::DEBUG);
		$signature_profile_list['data']['relative_app_path'] = $relative_app_path[0];
		
		\OCP\Util::writeLog('signlive', '======================================================================================================', ILogger::DEBUG);

		return new DataResponse($signature_profile_list);
	}

	/**
	 * Checks if given profile display name is already used by another signature profile definition
	 *
	 * @return bool
	 */
	protected function checkForDuplicateDisplayName($profile_name) {
		$app_config_keys = $this->config->getAppKeys('signlive');
		$existing_profile_names = [];
		foreach ($app_config_keys as $config_key_id => $config_key) {
			if (str_starts_with($config_key, 'cloudsuite_gears_profile_id_')) {
				$unique_profile_key = str_replace('cloudsuite_gears_profile_id_', '', $config_key);
				\OCP\Util::writeLog('signlive', 'Signature profile found: '.$unique_profile_key, ILogger::DEBUG);
				$profile_name_value = $this->config->getAppValue('signlive',
															  	 'cloudsuite_gears_profile_name_'.$unique_profile_key, '');
				$existing_profile_names[] = $profile_name_value;
			}
		}

		if (in_array($profile_name, $existing_profile_names)) {
			\OCP\Util::writeLog('signlive', 'Signature profile named \''.$profile_name.'\' already exists.', ILogger::DEBUG);
			return true;
		}

		\OCP\Util::writeLog('signlive', 'No signature profile named \''.$profile_name.'\' exists yet.', ILogger::DEBUG);
		return false;
	}

	/**
	 * Add a new signature profile definition
	 *
	 * @return DataResponse
	 */
	public function addSignatureProfile() {
		\OCP\Util::writeLog('signlive', 'addSignatureProfile() called in signlive/lib/Controller/IntarsysSettingsController.php.', ILogger::DEBUG);
		$num_profiles = $this->getNumSignatureProfiles();
		$request_parameters = $this->request->getParams();

		foreach ($request_parameters as $parameter_name => $parameter_value) {
			\OCP\Util::writeLog('signlive', 'Request parameter: '.$parameter_name.' => '.$parameter_value, ILogger::DEBUG);
		}

		if (array_key_exists('cloudsuite_gears_profile_id', $request_parameters) &&
			array_key_exists('cloudsuite_gears_profile_name', $request_parameters) &&
			array_key_exists('cloudsuite_gears_profile_type', $request_parameters) &&
			array_key_exists('cloudsuite_gears_profile_json', $request_parameters) &&
			array_key_exists('cloudsuite_gears_default_signature_profile', $request_parameters) && 
			array_key_exists('cloudsuite_gears_nextcloud_groups', $request_parameters) &&
			array_key_exists('cloudsuite_gears_signature_profile_menu_priority', $request_parameters)) {

			\OCP\Util::writeLog('signlive', 'All required request parameters present. Saving new profile to settings.', ILogger::DEBUG);

			$profile_id = $request_parameters['cloudsuite_gears_profile_id'];
			$profile_name = $request_parameters['cloudsuite_gears_profile_name'];
			$profile_type = $request_parameters['cloudsuite_gears_profile_type'];
			$profile_json_section = $request_parameters['cloudsuite_gears_profile_json'];
			$profile_default_flag = $request_parameters['cloudsuite_gears_default_signature_profile'];
			$profile_nextcloud_groups = $request_parameters['cloudsuite_gears_nextcloud_groups'];
			$profile_menu_priority = $request_parameters['cloudsuite_gears_signature_profile_menu_priority'];

			// Profile display name must not be empty
			if (empty($profile_name) || $profile_name === '') {
				return new DataResponse([
					'success' => false,
					'message' => 'The display name for a signature profile must not be empty!',
				]);
			}

			// Check if there is already a signature profile with the same display name chosen for the new one
			$profile_name_exists = $this->checkForDuplicateDisplayName($profile_name);
			if ($profile_name_exists === true) {
				return new DataResponse([
					'success' => false,
					'message' => 'A signature profile named \''.$profile_name.'\' already exists, display names must be unique! Please choose a different display name.',
				]);	
			}

			// Check if the given profile_id fits into the 64 character length limit imposed by the oc_appconfig table target column
			// If not: Truncate the profile ID to 64 bit (<prefix> + <counter>;;<display name>)
			// Find out which is the longest non-truncated settings key, then truncate the initial profile ID to accomodate the shortest required cut
			$profile_id_with_prefix = 'cloudsuite_gears_profile_id_'.$profile_id;
			$profile_name_with_prefix = 'cloudsuite_gears_profile_name_'.$profile_id;
			$profile_type_with_prefix = 'cloudsuite_gears_profile_type_'.$profile_id;
			$profile_json_with_prefix = 'cloudsuite_gears_profile_json_'.$profile_id;
			$profile_default_with_prefix ='cloudsuite_gears_default_profile_'.$profile_id;
			$profile_groups_with_prefix = 'cloudsuite_gears_profile_nextcloud_groups_'.$profile_id;
			$profile_menu_priority_with_prefix = 'cloudsuite_gears_profile_menu_priority_'.$profile_id;

			$profile_key_lengths_to_check = [
				'cloudsuite_gears_profile_id_' => $profile_id_with_prefix,
				'cloudsuite_gears_profile_name_' => $profile_name_with_prefix,
				'cloudsuite_gears_profile_type_' => $profile_type_with_prefix,
				'cloudsuite_gears_profile_json_' => $profile_json_with_prefix,
				'cloudsuite_gears_default_profile_' => $profile_default_with_prefix,
				'cloudsuite_gears_profile_nextcloud_groups_' => $profile_groups_with_prefix,
				'cloudsuite_gears_profile_menu_priority_' => $profile_menu_priority_with_prefix
			];

			$max_truncation_length_index = -1;
			$max_truncation_length = 64;
			$index_counter = 0;
			$max_truncation_prefix = '';
			foreach ($profile_key_lengths_to_check as $key_prefix => $profile_key_to_check) {
				$prefix_to_check = $key_prefix;
				$prefix_to_check_length = strlen($prefix_to_check);
				$truncation_length = 64 - $prefix_to_check_length;
				\OCP\Util::writeLog('signlive', 'prefix_to_check =  '.$prefix_to_check.', prefix_to_check_length = '.$prefix_to_check_length, ILogger::DEBUG);
				if ($truncation_length < $max_truncation_length) {
					$max_truncation_length = $truncation_length;
					$max_truncation_length_index = $index_counter;
					$max_truncation_prefix = $key_prefix; 
				}
				$index_counter = $index_counter + 1;
			}

			\OCP\Util::writeLog('signlive', 'max_truncation_length for given profile ID \''.$profile_id_with_prefix.'\': '.$max_truncation_length.' for settings key '.$max_truncation_prefix, ILogger::DEBUG);
			
			if ($max_truncation_length_index >= 0) {
				\OCP\Util::writeLog('signlive', 'Given profile ID needs to be truncated to '.$max_truncation_length.' characters.', ILogger::DEBUG);
				$profile_id = substr($profile_id, 0, $max_truncation_length);
				\OCP\Util::writeLog('signlive', 'Truncated unique profile ID: '.$profile_id, ILogger::DEBUG);
			}

			// Final keys with truncated profile_id portion
			$profile_id_with_prefix = 'cloudsuite_gears_profile_id_'.$profile_id;
			$profile_name_with_prefix = 'cloudsuite_gears_profile_name_'.$profile_id;
			$profile_type_with_prefix = 'cloudsuite_gears_profile_type_'.$profile_id;
			$profile_json_with_prefix = 'cloudsuite_gears_profile_json_'.$profile_id;
			$profile_default_with_prefix ='cloudsuite_gears_default_profile_'.$profile_id;
			$profile_groups_with_prefix = 'cloudsuite_gears_profile_nextcloud_groups_'.$profile_id;
			$profile_menu_priority_with_prefix = 'cloudsuite_gears_profile_menu_priority_'.$profile_id;

			$existing_profile_id = $this->config->getAppValue('signlive', $profile_id, '');
			
			\OCP\Util::writeLog('signlive', 'Check for existing profile_id \''.$profile_id.'\' resulted in: \''.$existing_profile_id.'\'', ILogger::DEBUG);

			if (empty($existing_profile_id) || $existing_profile_id === '') {
				$unique_profile_key = $profile_id;

				\OCP\Util::writeLog('signlive', '=================================================================', ILogger::DEBUG);
				\OCP\Util::writeLog('signlive', 'Storing new signature profile with unique ID: '.$unique_profile_key, ILogger::DEBUG);
				
				try {
					\OCP\Util::writeLog('signlive', 'Storing new signature profile unique ID: '.$unique_profile_key, ILogger::DEBUG);
					$this->config->setAppValue( 
						'signlive',
						$profile_id_with_prefix, $unique_profile_key);
					
					
					\OCP\Util::writeLog('signlive', 'Storing new signature profile name \''.$profile_name.'\' with key \''.$profile_name_with_prefix.'\'', ILogger::DEBUG);
					$this->config->setAppValue( 
						'signlive',
						$profile_name_with_prefix, $profile_name);
					
					\OCP\Util::writeLog('signlive', 'Storing new signature profile type \''.$profile_type.'\' with key \''.$profile_type_with_prefix.'\'', ILogger::DEBUG);
					$this->config->setAppValue( 
						'signlive',
						$profile_type_with_prefix, $profile_type);
				
					\OCP\Util::writeLog('signlive', 'Storing new signature profile JSON: '.$profile_json_section, ILogger::DEBUG);
					$this->config->setAppValue( 
						'signlive',
						$profile_json_with_prefix, $profile_json_section);
					
					\OCP\Util::writeLog('signlive', 'Storing new signature profile default profile flag: '.$profile_default_flag, ILogger::DEBUG);
					$this->config->setAppValue( 
						'signlive',
						$profile_default_with_prefix, $profile_default_flag);
					
					\OCP\Util::writeLog('signlive', 'Storing new signature profile Nextcloud groups: '.$profile_nextcloud_groups, ILogger::DEBUG);
					$this->config->setAppValue( 
						'signlive',
						$profile_groups_with_prefix, $profile_nextcloud_groups);

					\OCP\Util::writeLog('signlive', 'Storing new signature profile menu priority: '.$profile_menu_priority, ILogger::DEBUG);
					$this->config->setAppValue( 
						'signlive',
						$profile_menu_priority_with_prefix, $profile_menu_priority);
	
					return new DataResponse([
						'success' => true,
						'message' => 'Successfully created new signature profile \''.$profile_name.'\'',
						'details' => 'New signature profile saved successfully.'
					]);
				} catch (Exception $e) {
					\OCP\Util::writeLog('signlive', 'Exception occured storing new signature profile with unique ID: '.$e->getMessage(), ILogger::DEBUG);
					return new DataResponse([
						'success' => false,
						'message' => 'Exception occurred while storing \''.$profile_name.'\' to Nextcloud settings!',
						'details' => $e->getMessage()
					]);	
				}
				\OCP\Util::writeLog('signlive', '=================================================================', ILogger::DEBUG);
			} else {
				\OCP\Util::writeLog('signlive', 'Can not create new signature profile \''.$profile_name.'\'! Profile with same ID already exists, please call editSignatureProfile() instead or provide a different profile ID.', ILogger::ERROR);
				return new DataResponse([
					'success' => false,
					'message' => 'Can not create new signature profile \''.$profile_name.'\'! Profile with same ID already exists, please use editSignatureProfile() instead or provide a different profile ID.',
					'details' => 'Only one signature profile may exist with unique ID \''.$existing_profile_id.'!\''
				]);
			}

		} else {
			$provided_properties = '';
			$missing_properties = '';
			
			foreach ($this->required_profile_properties as $required_profile_property) {
				if (array_key_exists('cloudsuite_gears_profile_id', $request_parameters)) {
					if ($provided_properties === '') {
						$provided_properties = $required_profile_property;
					} else {
						$provided_properties  = $provided_properties.', '.$required_profile_property;
					} 
				} else {
					if ($missing_properties === '') {
						$missing_properties = $required_profile_property;
					} else {
						$missing_properties = $missing_properties.', '.$required_profile_property;
					}
				}
			}

			\OCP\Util::writeLog('signlive', 'Can not create new profile! One or more required signature profile properties were not provided.', ILogger::ERROR);
			\OCP\Util::writeLog('signlive', 'Provided properties: '.$provided_properties.'; missing properties: '.$missing_properties, ILogger::ERROR);

			return new DataResponse([
				'success' => false,
				'message' => 'Can not create new profile! One or more required signature profile properties were not provided.',
				'details' => 'Provided properties: '.$provided_properties.'; missing properties: '.$missing_properties
			]);
		}
	}

	/**
	 * Change the properties of an existing signature profile definition
	 *
	 * @return DataResponse
	 */
	public function editSignatureProfile() {
		\OCP\Util::writeLog('signlive', '============================================================================================================', ILogger::DEBUG);
		\OCP\Util::writeLog('signlive', 'editSignatureProfile() called in signlive/lib/Controller/IntarsysSettingsController.php.', ILogger::DEBUG);
		
		$num_profiles = $this->getNumSignatureProfiles();
		$request_parameters = $this->request->getParams();

		$profile_id_request_parameter = '';
		foreach ($request_parameters as $parameter_name => $parameter_value) {
			\OCP\Util::writeLog('signlive', 'Request parameter: '.$parameter_name.' => '.$parameter_value, ILogger::DEBUG);
			if ($parameter_name === 'cloudsuite_gears_profile_id') {
				$profile_id_request_parameter = $parameter_value;
			}
		}

		\OCP\Util::writeLog('signlive', 'Profile ID in request_parameters: '.$profile_id_request_parameter, ILogger::DEBUG);
		\OCP\Util::writeLog('signlive', '============================================================================================================', ILogger::DEBUG);

		if ($profile_id_request_parameter !== '') {
			$profile_id = $profile_id_request_parameter;
			$app_config_keys = $this->config->getAppKeys('signlive');
			
			\OCP\Util::writeLog('signlive', 'cloudsuite_gears_profile_id parameter: '.$profile_id, ILogger::DEBUG);
			
			$provided_unique_profile_key = '';
			$provided_profile_id = '';

			$provided_profile_id_parts = explode(';;', $profile_id);
			if (count($provided_profile_id_parts) === 2) {
				$provided_unique_profile_key = $provided_profile_id_parts[1];
				$provided_profile_id_string = str_replace('cloudsuite_gears_profile_id_', '', $provided_profile_id_parts[0]);
				$provided_profile_id = intval($provided_profile_id_string);

				\OCP\Util::writeLog('signlive', 'provided_profile_id         (numerical ID) = '.$provided_profile_id, ILogger::DEBUG);
				\OCP\Util::writeLog('signlive', 'provided_unique_profile_key (string ID)    = '.$provided_unique_profile_key, ILogger::DEBUG);
			} else {
				\OCP\Util::writeLog('signlive', 'Can not modify an existing signature profile without a valid profile ID! Provided profile ID '.$profile_id.' is not existing in the Nextcloud settings.', ILogger::ERROR);
				return new DataResponse([
					'success' => false,
					'message' => 'Can not modify an existing signature profile without a valid profile ID!',
				]);
			}

			$existing_profile_id = -1;
			$existing_unique_profile_key = '';

			foreach ($app_config_keys as $app_key => $app_value) {
				if (str_starts_with($app_value, 'cloudsuite_gears_profile_id_')) {
					\OCP\Util::writeLog('signlive', 'Found profile_id candidate app value: '.$app_value, ILogger::DEBUG);

					$profile_id_parts = explode(';;', $app_value);
					\OCP\Util::writeLog('signlive', 'profile_id_parts count = '.count($profile_id_parts), ILogger::DEBUG);
					if(count($profile_id_parts) === 2) {
						\OCP\Util::writeLog('signlive', 'profile_id_parts[0] (numerical ID) = '.$profile_id_parts[0], ILogger::DEBUG);
						\OCP\Util::writeLog('signlive', 'profile_id_parts[1] (string ID)    = '.$profile_id_parts[1], ILogger::DEBUG);
						
						$current_profile_unique_key = $profile_id_parts[1];
						$current_profile_id_string = str_replace('cloudsuite_gears_profile_id_', '', $profile_id_parts[0]);
						$current_profile_id = intval($current_profile_id_string);

						if ($current_profile_unique_key === $provided_unique_profile_key &&
							$current_profile_id === $provided_profile_id) {
							$existing_unique_profile_key = $profile_id_parts[1];
							$existing_profile_id_string = str_replace('cloudsuite_gears_profile_id_', '', $profile_id_parts[0]);
							$existing_profile_id = intval($existing_profile_id_string);
							\OCP\Util::writeLog('signlive', 'Found matching existing_profile_id: '.$existing_profile_id, ILogger::DEBUG);
							break;
						}
					}
				}
			}

			\OCP\Util::writeLog('signlive', 'Result of existing_profile_id check: $existing_profile_id = \''.$existing_profile_id.'\', $existing_unique_profile_key = \''.$existing_unique_profile_key.'\'', ILogger::DEBUG);
			
			if ($existing_profile_id < 0) {
				\OCP\Util::writeLog('signlive', 'Can not modify signature profile ID \''.$profile_id.'\': No existing signature profile with the specified ID found!', ILogger::DEBUG);
				return new DataResponse([
					'success' => false,
					'message' => 'Can not modify signature profile ID \''.$profile_id.'\': No existing signature profile with the specified ID found!'
				]);
			} else {
				\OCP\Util::writeLog('signlive', '======================== Saving modified profile data ========================', ILogger::DEBUG);
				
				$profile_key_for_storing = $existing_profile_id.';;'.$existing_unique_profile_key;
				if (array_key_exists('cloudsuite_gears_profile_name', $request_parameters)) {
					$profile_name = $request_parameters['cloudsuite_gears_profile_name'];
					\OCP\Util::writeLog('signlive', 'Saving modified profile name: \''.$profile_name.'\'', ILogger::DEBUG);
					
					// This check suppresses saving modified properties even if the display text has NOT been changed.
					// So it is left out for the time being.
					/*$profile_name_exists = $this->checkForDuplicateDisplayName($profile_name);
					if ($profile_name_exists === true) {
						return new DataResponse([
							'success' => false,
							'message' => 'A signature profile named \''.$profile_name.'\' already exists! Can not store modified signature profile properties, please choose a different display name.',
						]);	
					}*/

					if (empty($profile_name) || $profile_name === '') {
						return new DataResponse([
							'success' => false,
							'message' => 'The display name for a signature profile must not be empty!',
						]);
					}
					
					$this->config->setAppValue( 
						'signlive',
						'cloudsuite_gears_profile_name_'.$profile_key_for_storing, $profile_name);
				} else {
					return new DataResponse([
						'success' => false,
						'message' => 'No display name was provided in request to change the properties of an existing signature profile, can not continue!',
					]);
				}

				if (array_key_exists('cloudsuite_gears_profile_json', $request_parameters)) {
					$profile_json_section = $request_parameters['cloudsuite_gears_profile_json'];

					$this->config->setAppValue( 
						'signlive',
						'cloudsuite_gears_profile_json_'.$profile_key_for_storing, $profile_json_section);

					// Check if the variable substitutions in the JSON section look OK - moved to frontend code
					/*$json_validation_result = $this->variableSubstitution->validateJSONSubstitutionVariables($profile_json_section);
					\OCP\Util::writeLog('signlive', 'JSON section validation result: \''.print_r($json_validation_result, TRUE).'\'', ILogger::DEBUG);

					if ($json_validation_result['success'] == false) {
						\OCP\Util::writeLog('signlive', 'Failed to save modified profile JSON section: \''.$json_validation_result['status_message'].'\'', ILogger::WARN);
						return new DataResponse([
							'success' => false,
							'message' => 'Modified JSON section for profile \''.$request_parameters['cloudsuite_gears_profile_name'].'\' contains invalid syntax: \''.$json_validation_result['status_message'].'\'!',
						]);
					} else {
						\OCP\Util::writeLog('signlive', 'Saving modified profile JSON section: \''.$profile_json_section.'\'', ILogger::DEBUG);
						$this->config->setAppValue( 
							'signlive',
							'cloudsuite_gears_profile_json_'.$profile_key_for_storing, $profile_json_section);
					}*/
				} else {
					return new DataResponse([
						'success' => false,
						'message' => 'No JSON section was provided in request to change the properties of an existing signature profile, can not continue!',
					]);
				}

				if (array_key_exists('cloudsuite_gears_profile_type', $request_parameters)) {
					$profile_type = $request_parameters['cloudsuite_gears_profile_type'];
					\OCP\Util::writeLog('signlive', 'Saving modified profile type: \''.$profile_type.'\'', ILogger::DEBUG);
					$this->config->setAppValue( 
						'signlive',
						'cloudsuite_gears_profile_type_'.$profile_key_for_storing, $profile_type);
				}
				
				if (array_key_exists('cloudsuite_gears_default_signature_profile', $request_parameters))  {
					$profile_default_flag = $request_parameters['cloudsuite_gears_default_signature_profile'];
					\OCP\Util::writeLog('signlive', 'Saving modified profile default flag: \''.$profile_default_flag.'\'', ILogger::DEBUG);
					$this->config->setAppValue( 
						'signlive',
						'cloudsuite_gears_default_profile_'.$profile_key_for_storing, $profile_default_flag);
					
				}
				if (array_key_exists('cloudsuite_gears_nextcloud_groups', $request_parameters))  {
					$profile_nextcloud_groups = $request_parameters['cloudsuite_gears_nextcloud_groups'];
					\OCP\Util::writeLog('signlive', 'Saving modified Nextcloud groups list: \''.$profile_nextcloud_groups.'\'', ILogger::DEBUG);
					$this->config->setAppValue( 
						'signlive',
						'cloudsuite_gears_profile_nextcloud_groups_'.$profile_key_for_storing, $profile_nextcloud_groups);
				}
				if (array_key_exists('cloudsuite_gears_signature_profile_menu_priority', $request_parameters))  {
					$profile_menu_priority = $request_parameters['cloudsuite_gears_signature_profile_menu_priority'];
					\OCP\Util::writeLog('signlive', 'Saving modified profile menu priority: \''.$profile_menu_priority.'\'', ILogger::DEBUG);
					$this->config->setAppValue( 
						'signlive',
						'cloudsuite_gears_profile_menu_priority_'.$profile_key_for_storing, $profile_menu_priority);
				}
				\OCP\Util::writeLog('signlive', '======================== Saving modified profile data ========================', ILogger::DEBUG);

				return new DataResponse([
					'success' => true,
					'message' => 'Successfully modified signature profile ID \''.$profile_id.'\'.'
				]);
			}
		} else {
			return new DataResponse([
				'success' => false,
				'message' => 'Can not modify a signature profile without a valid profile ID!',
			]);
		}
	}

	/**
	 * Delete an existing signature profile definition
	 *
	 * @return DataResponse
	 */
	public function deleteSignatureProfile() {
		\OCP\Util::writeLog('signlive', 'deleteSignatureProfile() called in signlive/lib/Controller/IntarsysSettingsController.php.', ILogger::DEBUG);		
		$request_parameters = $this->request->getParams();

		$profile_id_request_parameter = '';
		foreach ($request_parameters as $parameter_name => $parameter_value) {
			\OCP\Util::writeLog('signlive', 'Request parameter: '.$parameter_name.' => '.$parameter_value, ILogger::DEBUG);
			if ($parameter_name === 'cloudsuite_gears_profile_id') {
				$profile_id_request_parameter = $parameter_value;
			}
		}

		if (array_key_exists('cloudsuite_gears_profile_id', $request_parameters)) {
			$profile_id = $profile_id_request_parameter;
			$app_config_keys = $this->config->getAppKeys('signlive');
			
			$provided_unique_profile_key = '';
			$provided_profile_id = '';

			$provided_profile_id_parts = explode(';;', $profile_id);
			if (count($provided_profile_id_parts) === 2) {
				$provided_unique_profile_key = $provided_profile_id_parts[1];
				$provided_profile_id_string = str_replace('cloudsuite_gears_profile_id_', '', $provided_profile_id_parts[0]);
				$provided_profile_id = intval($provided_profile_id_string);

				\OCP\Util::writeLog('signlive', 'provided_profile_id         (numerical ID) = '.$provided_profile_id, ILogger::DEBUG);
				\OCP\Util::writeLog('signlive', 'provided_unique_profile_key (string ID)    = '.$provided_unique_profile_key, ILogger::DEBUG);
			} else {
				\OCP\Util::writeLog('signlive', 'Can not delete an existing signature profile without a valid profile ID! Provided profile ID '.$profile_id.' is not existing in the Nextcloud settings.', ILogger::ERROR);
				return new DataResponse([
					'success' => false,
					'message' => 'Can not modify an existing signature profile without a valid profile ID!',
				]);
			}

			$existing_profile_id = -1;
			$existing_unique_profile_key = '';

			foreach ($app_config_keys as $app_key => $app_value) {
				if (str_starts_with($app_value, 'cloudsuite_gears_profile_id_')) {
					\OCP\Util::writeLog('signlive', 'Found profile_id candidate app value: '.$app_value, ILogger::DEBUG);

					$profile_id_parts = explode(';;', $app_value);
					\OCP\Util::writeLog('signlive', 'profile_id_parts count = '.count($profile_id_parts), ILogger::DEBUG);
					if(count($profile_id_parts) === 2) {
						\OCP\Util::writeLog('signlive', 'profile_id_parts[0] (numerical ID) = '.$profile_id_parts[0], ILogger::DEBUG);
						\OCP\Util::writeLog('signlive', 'profile_id_parts[1] (string ID)    = '.$profile_id_parts[1], ILogger::DEBUG);
						
						$current_profile_unique_key = $profile_id_parts[1];
						$current_profile_id_string = str_replace('cloudsuite_gears_profile_id_', '', $profile_id_parts[0]);
						$current_profile_id = intval($current_profile_id_string);

						if ($current_profile_unique_key === $provided_unique_profile_key &&
							$current_profile_id === $provided_profile_id) {
							$existing_unique_profile_key = $profile_id_parts[1];
							$existing_profile_id_string = str_replace('cloudsuite_gears_profile_id_', '', $profile_id_parts[0]);
							$existing_profile_id = intval($existing_profile_id_string);
							\OCP\Util::writeLog('signlive', 'Found matching existing_profile_id: '.$existing_profile_id, ILogger::DEBUG);
							break;
						}
					}
				}
			}
			
			if ($existing_profile_id < 0) {
				\OCP\Util::writeLog('signlive', 'Can not delete signature profile ID \''.$profile_id.'\': No existing signature profile with the specified ID found!', ILogger::ERROR);
				return new DataResponse([
					'success' => false,
					'message' => 'Can not delete signature profile ID \''.$profile_id.'\': No existing signature profile with the specified ID found!'
				]);
			} else {
				$profile_id_to_delete = $existing_profile_id.';;'.$existing_unique_profile_key;
				\OCP\Util::writeLog('signlive', 'Proceeding to delete signature profile ID: '.$profile_id_to_delete, ILogger::DEBUG);
	
				$this->config->deleteAppValue( 
					'signlive',
					'cloudsuite_gears_profile_id_'.$profile_id_to_delete);
				$this->config->deleteAppValue( 
					'signlive',
					'cloudsuite_gears_profile_name_'.$profile_id_to_delete);
				$this->config->deleteAppValue( 
					'signlive',
					'cloudsuite_gears_profile_type_'.$profile_id_to_delete);
				$this->config->deleteAppValue( 
					'signlive',
					'cloudsuite_gears_default_profile_'.$profile_id_to_delete);
				$this->config->deleteAppValue( 
					'signlive',
					'cloudsuite_gears_profile_json_'.$profile_id_to_delete);
				$this->config->deleteAppValue( 
					'signlive',
					'cloudsuite_gears_profile_nextcloud_groups_'.$profile_id_to_delete);
				$this->config->deleteAppValue( 
						'signlive',
						'cloudsuite_gears_profile_menu_priority_'.$profile_id_to_delete);
				
				return new DataResponse([
					'success' => true,
					'message' => 'Successfully deleted signature profile ID \''.$profile_id_to_delete.'\'.'
				]);
			}
		} else {
			\OCP\Util::writeLog('signlive', 'Can not delete signature profile without a valid profile ID!', ILogger::ERROR);
			return new DataResponse([
				'success' => false,
				'message' => 'Can not delete signature profile without a valid profile ID!',
			]);
		}
	}

	/**
	 * Validate JSON section (look for invalid variable substitutions)
	 * @return DataResponse
	 */
	public function validateJSONSection() {
		$request_parameters = $this->request->getParams();
		$profile_json_section = $request_parameters['cloudsuite_gears_profile_json'];
		$profile_name = $request_parameters['cloudsuite_gears_profile_name'];

		// Check if the variable substitutions in the JSON section look OK
		$json_validation_result = $this->variableSubstitution->validateJSONSubstitutionVariables($profile_json_section);
		\OCP\Util::writeLog('signlive', 'JSON section validation result: \''.print_r($json_validation_result, TRUE).'\'', ILogger::DEBUG);

		if ($json_validation_result['success'] == false) {
			\OCP\Util::writeLog('signlive', 'New JSON section contains invalid substitution placeholders: \''.$json_validation_result['status_message'].'\'', ILogger::DEBUG);
			return new DataResponse([
				'success' => false,
				'message' => 'New JSON section for profile \''.$profile_name.'\' contains invalid syntax: \''.$json_validation_result['status_message'].'\'!',
			]);
		} else {
			\OCP\Util::writeLog('signlive', 'New JSON for profile \''.$profile_name.'\' validated successfully.', ILogger::DEBUG);
			return new DataResponse([
				'success' => true,
				'message' => 'Modified JSON section for profile \''.$profile_name.'\' is valid: \''.$json_validation_result['status_message'].'\'.',
			]);
		}
	}

	/**
	 * Remove the settings flag indicating that the initial signature profile creation has run already
	 * Must be called in preparation of a reset to signature profile factory defaults 
	 * @return DataResponse
	 */
	public function resetProfileCreationFlag() {
		\OCP\Util::writeLog('signlive', 'resetProfileCreationFlag() called in signlive/lib/Controller/IntarsysSettingsController.php.', ILogger::DEBUG);

		$initial_profiles_created_value = $this->config->getAppValue('signlive', 'cloudsuite_gears_initial_profiles_created', '');
		if ($initial_profiles_created_value !== '') {
			// Initial profiles have been created before, reset the flag that indicates this explicitly
			\OCP\Util::writeLog('signlive', 'Setting \'cloudsuite_gears_initial_profiles_created\' to false/empty, initial signature profile definitions can now be recreated.', ILogger::DEBUG);
			$this->config->setAppValue('signlive', 'cloudsuite_gears_initial_profiles_created', '');
			return new DataResponse([
				'success' => true,
				'message' => 'Initial signature profile definition creation flag reset to false, factory default profiles can now be recreated.'
			]);
		}

		return new DataResponse([
			'success' => false,
			'message' => 'Initial signature profile definition creation flag is currently set to false, there is no need to reset it.'
		]);
	}

	/**
	 * Reset all signature profiles to factory defaults
	 * @return DataResponse
	 */
	public function profilesFactoryReset() {
		\OCP\Util::writeLog('signlive', 'profilesFactoryReset() called in signlive/lib/Controller/IntarsysSettingsController.php.', ILogger::DEBUG);

		$initial_profiles_created_value = $this->config->getAppValue('signlive', 'cloudsuite_gears_initial_profiles_created', '');
		\OCP\Util::writeLog('signlive', 'cloudsuite_gears_initial_profiles_created = '.$initial_profiles_created_value, ILogger::DEBUG);
		if ($initial_profiles_created_value !== '') {
			\OCP\Util::writeLog('signlive', 'cloudsuite_gears_initial_profiles_created set to non-empty value, will not recreate factory default signature profile entries.', ILogger::DEBUG);
			return new DataResponse([
				'success' => false,
				'message' => 'Initial signature profile definition creation flag is currently set to true! Will not proceed with recreation of factory default signature profiles.'
			]);
		} else {
			// First delete all existing signature profile definitions
			$restricted_profile_data = $this->getProfileList(false);
			$public_profile_data = $this->getProfileList(true);

			\OCP\Util::writeLog('signlive', 'restricted_profile_data = .'.print_r($restricted_profile_data, TRUE), ILogger::DEBUG);
			\OCP\Util::writeLog('signlive', 'public_profile_data = .'.print_r($public_profile_data, TRUE), ILogger::DEBUG);

			\OCP\Util::writeLog('signlive', 'Deleting existing restricted signature profiles.', ILogger::DEBUG);
			foreach ($restricted_profile_data['data']['restricted_signature_profiles'] as $profile_data) {
				$profile_id = $profile_data['profile_id'];
				$profile_unique_id = str_replace('cloudsuite_gears_profile_id_', '', $profile_id);
				\OCP\Util::writeLog('signlive', 'Deleting restricted signature profile \''.$profile_unique_id.'\'', ILogger::DEBUG);

				$this->config->deleteAppValue( 
					'signlive',
					'cloudsuite_gears_profile_id_'.$profile_unique_id);
				$this->config->deleteAppValue( 
					'signlive',
					'cloudsuite_gears_profile_name_'.$profile_unique_id);
				$this->config->deleteAppValue( 
					'signlive',
					'cloudsuite_gears_profile_type_'.$profile_unique_id);
				$this->config->deleteAppValue( 
					'signlive',
					'cloudsuite_gears_default_profile_'.$profile_unique_id);
				$this->config->deleteAppValue( 
					'signlive',
					'cloudsuite_gears_profile_json_'.$profile_unique_id);
				$this->config->deleteAppValue( 
					'signlive',
					'cloudsuite_gears_profile_nextcloud_groups_'.$profile_unique_id);
				$this->config->deleteAppValue(
					'signlive',
					'cloudsuite_gears_profile_menu_priority_'.$profile_unique_id);
			}

			\OCP\Util::writeLog('signlive', 'Deleting existing public signature profiles.', ILogger::DEBUG);
			foreach ($public_profile_data['data']['public_signature_profiles'] as $profile_data) {
				$profile_id = $profile_data['profile_id'];
				$profile_unique_id = str_replace('cloudsuite_gears_profile_id_', '', $profile_id);
				\OCP\Util::writeLog('signlive', 'Deleting public signature profile \''.$profile_unique_id.'\'', ILogger::DEBUG);

				$this->config->deleteAppValue(
					'signlive',
					'cloudsuite_gears_profile_id_'.$profile_unique_id);
				$this->config->deleteAppValue(
					'signlive',
					'cloudsuite_gears_profile_name_'.$profile_unique_id);
				$this->config->deleteAppValue(
					'signlive',
					'cloudsuite_gears_profile_type_'.$profile_unique_id);
				$this->config->deleteAppValue( 
					'signlive',
					'cloudsuite_gears_default_profile_'.$profile_unique_id);
				$this->config->deleteAppValue( 
					'signlive',
					'cloudsuite_gears_profile_json_'.$profile_unique_id);
				$this->config->deleteAppValue( 
					'signlive',
					'cloudsuite_gears_profile_nextcloud_groups_'.$profile_unique_id);
				$this->config->deleteAppValue( 
					'signlive',
					'cloudsuite_gears_profile_menu_priority_'.$profile_unique_id);
			}
			// Then recreate the factory defaults
			\OCP\Util::writeLog('signlive', 'Recreating factory default signature profiles.', ILogger::DEBUG);
			$signature_profile_setup = new SignatureProfilesSetup($this->config);
			$initial_profiles_created = $signature_profile_setup->setupDefaultSignatureProfiles();
			if ($initial_profiles_created === true) {
				// Finally set cloudsuite_gears_initial_profiles_created to a non-empty value to prevent repeated factory default resets
				\OCP\Util::writeLog('signlive', 'All customized signature profile definitions have been removed. Factory default signature profile definitions have been recreated.', ILogger::INFO);
				
				$this->config->setAppValue('signlive', 'cloudsuite_gears_initial_profiles_created', 'true');
				return new DataResponse([
					'success' => true,
					'message' => 'All customized signature profile definitions have been removed. Factory default signature profile definitions have been recreated.'
				]);
			} else {
				\OCP\Util::writeLog('signlive', 'Did not recreate factory default signature profiles! Failed to delete customized signature profile settings and/or to reset the SignLive! app signature profiles to factory defaults.', ILogger::ERROR);
				return new DataResponse([
					'success' => false,
					'message' => 'Did not recreate factory default signature profiles! Failed to delete customized signature profile settings and/or to reset the SignLive! app signature profiles to factory defaults.'
				]);
			}
		}
	}

	/**
	 * Check status of initial signature profile setup
	 * Create factory default signature profiles if initial setup has not been run before (e. g. by the Migration RepairStep on initial enabling of the SignLive! app)
	 * @return DataResponse
	 */
	public function checkInitialSignatureProfileSetup() {
		\OCP\Util::writeLog('signlive', 'checkInitialSignatureProfileSetup() called in signlive/lib/Controller/IntarsysSettingsController.php.', ILogger::DEBUG);

		$initial_profiles_created_value = $this->config->getAppValue('signlive', 'cloudsuite_gears_initial_profiles_created', '');
		\OCP\Util::writeLog('signlive', 'cloudsuite_gears_initial_profiles_created = '.$initial_profiles_created_value, ILogger::DEBUG);
		if ($initial_profiles_created_value !== '') {
			\OCP\Util::writeLog('signlive', 'cloudsuite_gears_initial_profiles_created set to non-empty value, will not recreate factory default signature profile entries.', ILogger::DEBUG);
			return new DataResponse([
				'success' => true,
				'message' => 'Initial signature profile definition creation flag is currently set to true. Assuming initial signature profile setup has already run.'
			]);
		} else {
			\OCP\Util::writeLog('signlive', 'cloudsuite_gears_initial_profiles_created not set or set to empty value, will create factory default signature profile entries now.', ILogger::INFO);
		
			$signature_profile_setup = new SignatureProfilesSetup($this->config);
			$initial_profiles_created = $signature_profile_setup->setupDefaultSignatureProfiles();
			if ($initial_profiles_created === true) {
				// Set cloudsuite_gears_initial_profiles_created to a non-empty value to prevent repeated factory default resets
				\OCP\Util::writeLog('signlive', 'Factory default signature profile definitions have been created.', ILogger::INFO);
				
				$this->config->setAppValue('signlive', 'cloudsuite_gears_initial_profiles_created', 'true');
				return new DataResponse([
					'success' => true,
					'message' => 'Factory default signature profile definitions have been created.'
				]);
			} else {
				\OCP\Util::writeLog('signlive', 'Did not create factory default signature profiles: Failed to initialize the SignLive! app with factory default signature profiles.', ILogger::ERROR);
				return new DataResponse([
					'success' => false,
					'message' => 'Did not create factory default signature profiles: Failed to initialize the SignLive! app with factory default signature profiles.'
				]);
			}
		}
	}
}
