<?php

/**
* This file is licensed under the Affero General Public License version 3 or
* later. See the COPYING file.
*
* @author Fabian Aichele <fabian.aichele@bpex.de>
* @copyright intarsys AG 2021
*/

namespace OCA\IntarsysCloudSuiteGearsIntegration\Notification;

use OCP\IURLGenerator;
use OCP\L10N\IFactory;
use OCP\Notification\INotification;
use OCP\Notification\INotifier;
use OCP\Util;
use OCP\ILogger;
use OCP\IConfig;

class CloudsuiteGearsNotifier implements INotifier {

	/** @var IFactory */
	protected $l10nFactory;

	/** @var IURLGenerator */
	protected $url;

	/** @var IConfig */
	protected $config;

	/**
	 * Notifier constructor.
	 *
	 * @param IFactory $l10nFactory
	 * @param IURLGenerator $url
	 */
	public function __construct(IFactory $l10nFactory, 
								IURLGenerator $url,
								IConfig $config) {
		$this->l10nFactory = $l10nFactory;
		$this->url = $url;
		$this->config = $config;
	}

	/**
	 * Identifier of the notifier, only use [a-z0-9_]
	 *
	 * @return string
	 * @since 17.0.0
	 */
	public function getID(): string {
		return 'signlive';
	}

	/**
	 * Human readable name describing the notifier
	 *
	 * @return string
	 * @since 17.0.0
	 */
	public function getName(): string {
		return $this->l10nFactory->get('signlive')->t('Sign Live!');
	}

	/**
     * Get a specific signature request job record
     * @return array
     */
    protected function getJobDefinitionEntry($signature_job_uuid) {
        \OCP\Util::writeLog('signlive', 'getJobDefinitionEntry() called in signlive/lib/Notification/CloudsuiteGearsNotifier.php. job_uuid = '.$signature_job_uuid, ILogger::DEBUG);
    
        $job_definition_entry_data = [];

        if (empty($signature_job_uuid) || $signature_job_uuid === '') {
            \OCP\Util::writeLog('signlive', 'getJobDefinitionEntry() called with an empty job_uuid!', ILogger::WARN);
            return $job_definition_entry_data;
        }

		\OCP\Util::writeLog('signlive', '==============================================================', ILogger::DEBUG);
		$job_record_key = 'csgi_job_def_'.$signature_job_uuid;
        \OCP\Util::writeLog('signlive', 'getJobDefinitionEntry() LOOKING FOR job_uuid key = '.$job_record_key, ILogger::DEBUG);
        
		$job_record_value = $this->config->getAppValue('signlive', $job_record_key, '');
		\OCP\Util::writeLog('signlive', 'getJobDefinitionEntry() GOT job_uuid data = '.$job_record_value, ILogger::DEBUG);
		\OCP\Util::writeLog('signlive', '==============================================================', ILogger::DEBUG);

        if (empty($job_record_value) || $job_record_value === '') {
            \OCP\Util::writeLog('signlive', 'getJobDefinitionEntry() found no signature request job entry for job_uuid = '.$signature_job_uuid, ILogger::DEBUG);
            return $job_definition_entry_data;
        }

        $job_record_components = explode(';;', $job_record_value);
        foreach ($job_record_components as $record_component) {
            $component_name_and_value = explode('::', $record_component);
            if (count($component_name_and_value) == 2) {
                $component_name = $component_name_and_value[0];
                $component_value = $component_name_and_value[1];
                $job_definition_entry_data[$component_name] = $component_value;
				// \OCP\Util::writeLog('signlive', 'Job definition entry: '.$component_name.' = '.$component_value, ILogger::DEBUG);
            }
        }
        
        \OCP\Util::writeLog('signlive', 'getJobDefinitionEntry() return value: '.print_r($job_definition_entry_data, TRUE), ILogger::DEBUG);

        return $job_definition_entry_data;
    }

	/**
	 * @param INotification $notification
	 * @param string $languageCode The code of the language that should be used to prepare the notification
	 * @return INotification
	 * @throws \InvalidArgumentException When the notification was not prepared by a notifier
	 */
	public function prepare(INotification $notification, string $languageCode): INotification {
		if ($notification->getApp() !== 'signlive') {
			// Not my app => throw
			throw new \InvalidArgumentException();
		}

		// Read the language from the notification
		$l = $this->l10nFactory->get('signlive', $languageCode);

		$signature_job_entry_data = $this->getJobDefinitionEntry($notification->getObjectId());

		$document_name = '<Unknown>';
		if (array_key_exists('file_name', $signature_job_entry_data)) {
			$document_name = $signature_job_entry_data['file_name'];
		}

		// $subject_parameters = $notification->getSubjectParameters();
		// $message_parameters = $notification->getMessageParameters();
		// $document_name = $subject_parameters['file_name'];

		\OCP\Util::writeLog('signlive', '=====================================================================', ILogger::DEBUG);
		// \OCP\Util::writeLog('signlive', 'Subject parameters: '.print_r($subject_parameters, TRUE), ILogger::DEBUG);
		// \OCP\Util::writeLog('signlive', 'Message parameters: '.print_r($message_parameters, TRUE), ILogger::DEBUG);
		\OCP\Util::writeLog('signlive', 'Object type: \''.$notification->getObjectType().'\'', ILogger::DEBUG);
		\OCP\Util::writeLog('signlive', 'Object ID  : \''.$notification->getObjectId().'\'', ILogger::DEBUG);
		\OCP\Util::writeLog('signlive', 'Subject    : \''.$notification->getSubject().'\'', ILogger::DEBUG);
		\OCP\Util::writeLog('signlive', 'User       : \''.$notification->getUser().'\'', ILogger::DEBUG);
		\OCP\Util::writeLog('signlive', 'Message    : \''.$notification->getMessage().'\'', ILogger::DEBUG);
		\OCP\Util::writeLog('signlive', 'Document   : \''.$document_name.'\'', ILogger::DEBUG);
		
		if ($notification->getSubject() === 'viewer_based_signature_success') {
			$notification->setParsedSubject($l->t('SignLive! successfully processed a viewer-based signature request.'));
		} else if ($notification->getSubject() === 'automatic_signature_success') {
			$notification->setParsedSubject($l->t('SignLive! successfully processed an automatic signature request.'));
		} else if ($notification->getSubject() === 'viewer_based_signature_failure') {
			$notification->setParsedSubject($l->t('SignLive! failed to process an automatic signature request!'));
		} else if ($notification->getSubject() === 'automatic_signature_failure') {
			$notification->setParsedSubject($l->t('SignLive! failed process an automatic signature request!'));
		}

		// If a status message is supplied as part of the signature job attributes, use it.
		// Otherwise just parse the given message text.  
		$message = $notification->getMessage();
		if (array_key_exists('status_message', $signature_job_entry_data)) {
			\OCP\Util::writeLog('signlive', 'Using \'status_message\' text as given in signature job record.', ILogger::DEBUG);
			$parsed_message = $signature_job_entry_data['status_message']; 
		} else {
			\OCP\Util::writeLog('signlive', 'Using message text as supplied in notification instance.', ILogger::DEBUG);
			$parsed_message = $l->t($message, $document_name);
		}
		$notification->setParsedMessage($parsed_message);

		\OCP\Util::writeLog('signlive', 'Parsed message: \''.$parsed_message.'\'', ILogger::DEBUG);
		\OCP\Util::writeLog('signlive', '=====================================================================', ILogger::DEBUG);

		// TODO: If notifications for successful jobs are relevant: 
		// Provide a direct download link to the signed document of the respective jobs in a notification?
		/*$link_to_signed_file = $this->url->linkTo('files', $document_name);
		
		if ($notification->getSubject() === 'viewer_based_signature_success' ||
			$notification->getSubject() === 'automatic_signature_success') {
			$notification->setLink($link_to_signed_file);
			\OCP\Util::writeLog('signlive', 'Setting link to signed docunebt: \''.$link_to_signed_file.'\'', ILogger::DEBUG);
		}*/

		$notification->setIcon($this->url->imagePath('signlive', 'app.svg'));

		foreach ($notification->getActions() as $action) {
			if ($action->getLabel() === 'confirm') {
				$action->setParsedLabel($l->t('Confirm'));
				$notification->addParsedAction($action);
			}
			// TODO: If notifications for successful jobs are relevant: 
			// Provide a direct download link to the signed document of the respective jobs in a notification?
			/*if ($action->getLabel() === 'download_signed_file') {
				$action->setParsedLabel($l->t('View signed file'));
				
				\OCP\Util::writeLog('signlive', 'Link to signed document: \''.$action->getLink().'\'', ILogger::DEBUG);
				
				$notification->addParsedAction($action);
			}*/
		}

		return $notification;
	}
}
