<?php

/**
* This file is licensed under the Affero General Public License version 3 or
* later. See the COPYING file.
*
* @author Fabian Aichele <fabian.aichele@bpex.de>
* @copyright intarsys AG 2021
*/

namespace OCA\IntarsysCloudSuiteGearsIntegration\Settings;

use OCP\AppFramework\Http\TemplateResponse;
use OCP\IL10N;
use OCP\Settings\ISettings;
use OCP\Template;
use OCP\Util;
use OCP\ILogger;
use OCP\IConfig;
use OCP\IGroup;
use OCP\IGroupManager;

use OCA\IntarsysCloudSuiteGearsIntegration\Util\SignLiveDefaultConfigValues;

class Admin implements ISettings {

	/** @var IConfig */
	private $config;

	/** @var IL10N */
	private $l10n;

	/** @var IGroupManager */
	private $groupManager;

	/**
	 * @param IConfig $config
	 * @param IL10N $l10n
	 * @param IGroupManager $groupManager
	 */
	public function __construct(IConfig $config,
								IL10N $l10n,
								IGroupManager $groupManager) {
		$this->l10n = $l10n;
		$this->config = $config;
		$this->groupManager = $groupManager;
	}

	/**
	 * Returns the number of defined signature profiles in the Nextcloud DB
	 * @return int
	 */
	private function getNumSignatureProfiles() {
		$num_signature_profiles = 0;
		$app_config_keys = $this->config->getAppKeys('signlive');

		foreach ($app_config_keys as $app_key => $app_value) {
			if (str_starts_with($app_key, "cloudsuite_gears_signature_profile_name_")) {
				$num_signature_profiles += 1;
			}
		}
		\OCP\Util::writeLog('signlive', 'num_signature_profiles = '.$num_signature_profiles, ILogger::DEBUG);

		return $num_signature_profiles;
	}

	/**
	 * @return TemplateResponse
	 */
	public function getForm() {
		$num_user_signature_profiles = $this->getNumSignatureProfiles();

		$app_config_keys = $this->config->getAppKeys('signlive');

		$show_verbose_failure_notifications_found = 0;
		$show_success_notifications_found = 0;
		foreach ($app_config_keys as $app_key_id => $app_key) {
			if ($app_key == "cloudsuite_gears_show_success_notifications") {
				$show_success_notifications_found = 1;
			}
			if ($app_key == "cloudsuite_gears_show_verbose_failure_notifications") {
				$show_verbose_failure_notifications_found = 1;
			}
		}

		if ($show_success_notifications_found == 0) {
			$this->config->setAppValue('signlive', 'cloudsuite_gears_show_success_notifications', 'false');
		}

		if ($show_verbose_failure_notifications_found == 0) {
			$this->config->setAppValue('signlive', 'cloudsuite_gears_show_verbose_failure_notifications', 'false');
		}

		$parameters = [
			'cloudsuite_gears_http_protocol_type' => $this->config->getAppValue('signlive', 'cloudsuite_gears_http_protocol_type', 'https'),
			'cloudsuite_gears_host' => $this->config->getAppValue('signlive', 'cloudsuite_gears_host', 'cloudsuite.intarsys.de'),
			'cloudsuite_gears_port' => $this->config->getAppValue('signlive', 'cloudsuite_gears_port', '443'),
			'cloudsuite_gears_gears_core_path' => $this->config->getAppValue('signlive', 'cloudsuite_gears_gears_core_path', '/cloudsuite-gears-nextcloud/core'),
			'cloudsuite_gears_show_default_signature_entries_in_context_menu' => $this->config->getAppValue('signlive', 'cloudsuite_gears_show_default_signature_entries_in_context_menu', 1),
			'cloudsuite_gears_show_user_profiles_in_context_menu' => $this->config->getAppValue('signlive', 'cloudsuite_gears_show_user_profiles_in_context_menu', 0),
			'cloudsuite_gears_global_attribute_company_name' => $this->config->getAppValue('signlive', 'cloudsuite_gears_global_attribute_company_name', 'Sign City AG'),
			'cloudsuite_gears_signature_profile_default_menu_priority' => $this->config->getAppValue('signlive', 'cloudsuite_gears_signature_profile_default_menu_priority', 75),
			'cloudsuite_gears_notification_duration' => $this->config->getAppValue('signlive', 'cloudsuite_gears_notification_duration', 7),
			'cloudsuite_gears_notification_check_interval' => $this->config->getAppValue('signlive', 'cloudsuite_gears_notification_check_interval', 1),
			'cloudsuite_gears_show_success_notifications' => $this->config->getAppValue('signlive', 'cloudsuite_gears_show_success_notifications', 0),
			'cloudsuite_gears_show_verbose_failure_notifications' => $this->config->getAppValue('signlive', 'cloudsuite_gears_show_verbose_failure_notifications', 0),
			'cloudsuite_gears_display_menu_priorities' => $this->config->getAppValue('signlive', 'cloudsuite_gears_display_menu_priorities', 0),
			'cloudsuite_gears_use_unified_job_control' => $this->config->getAppValue('signlive', 'cloudsuite_gears_use_unified_job_control', 1)
		];

		for ($i = 0; $i < $num_user_signature_profiles; $i++) {
			$user_profile_name_str = 'cloudsuite_gears_user_profile_name_'.$i;
			$user_json_section_1_str = 'cloudsuite_gears_user_profile_json_section_1_'.$i;
			$user_profile_type_str = 'cloudsuite_gears_user_profile_type_'.$i;
			
			$usr_profile_name = $this->config->getAppValue('signlive', $user_profile_name_str, '');
			$usr_json_section_1 = $this->config->getAppValue('signlive', $user_json_section_1_str, '');
			$usr_profile_type = $this->config->getAppValue('signlive', $user_profile_type_str, '');

			\OCP\Util::writeLog('signlive', 'User profile '.$i.': profile name ='.$usr_profile_name, ILogger::DEBUG);
			\OCP\Util::writeLog('signlive', 'User profile '.$i.': profile type ='.$usr_profile_type, ILogger::DEBUG);
			\OCP\Util::writeLog('signlive', 'User profile '.$i.': JSON section 1 ='.$usr_json_section_1, ILogger::DEBUG);

			if ($usr_profile_name != '' && $usr_json_section_1 != '' && $usr_profile_type != '') {
				$parameters[$user_profile_name_str] = $usr_profile_name;
				$parameters[$user_json_section_1_str] = $usr_json_section_1;
				$parameters[$user_profile_type_str] = $usr_profile_type;
			}
		}

		$nextcloud_groups = $this->groupManager->search('');
		$nextcloud_group_names = [];
		$num_nextcloud_groups = 0;
		foreach ($nextcloud_groups as $nextcloud_group) {
			$nextcloud_group_names[$nextcloud_group->getGID()] =  $nextcloud_group->getDisplayName();
			$num_nextcloud_groups += 1;
		}

		$parameters['num_nextcloud_groups'] = $num_nextcloud_groups;
		$parameters['nextcloud_group_names'] = $nextcloud_group_names;

		\OCP\Util::writeLog('signlive', 'Nextcloud groups list: '.print_r($nextcloud_group_names, TRUE), ILogger::DEBUG);

		\OCP\Util::writeLog('signlive', 'getForm() returns parameters: '.print_r($parameters, TRUE), ILogger::DEBUG);
		return new TemplateResponse('signlive', 'admin_settings', $parameters);
	}

	/**
	 * @return string the section ID, e.g. 'sharing'
	 */
	public function getSection() {
		return 'signlive';
	}

	/**
	 * @return int whether the form should be rather on the top or bottom of
	 * the admin section. The forms are arranged in ascending order of the
	 * priority values. It is required to return a value between 0 and 100.
	 *
	 * E.g.: 70
	 */
	public function getPriority() {
		return 5;
	}
}
