<?php

/**
* This file is licensed under the Affero General Public License version 3 or
* later. See the COPYING file.
*
* @author Fabian Aichele <fabian.aichele@bpex.de>
* @copyright intarsys AG 2021
*/

namespace OCA\IntarsysCloudSuiteGearsIntegration\Util;

use OCP\Util;
use OCP\ILogger;
use OCP\IUserSession;
use OCP\IUserManager;
use OCP\Accounts\IAccount;
use OCP\Accounts\IAccountManager;
use OCP\IConfig;
use OCP\AppFramework\Utility\ITimeFactory;
use OCP\Files\IMimeTypeDetector;
use OCP\Files\IRootFolder;

use OCA\IntarsysCloudSuiteGearsIntegration\Util\VariableSubstitution;

class JSONQueryCreator {
    /** @var \OCP\IConfig */
    protected $config;

    /** @var \OCP\Files\IMimeTypeDetector */
    protected $mimeTypeDetector;

    /** @var \OCP\Files\IRootFolder */
    protected $storage;

    /** @var ITimeFactory */
	protected $timeFactory;

    /** @var IUserSession */
    protected $userSession;

    protected $userManager;
    protected $accountManager;

    protected $variableSubstitution;

    public function __construct(IConfig $config,
                                IMimeTypeDetector $mimeTypeDetector,
                                IRootFolder $storage,
                                IUserSession $userSession,
                                IUserManager $userManager,
                                IAccountManager $accountManager,
                                ITimeFactory $timeFactory) {       
        $this->config = $config;
        $this->mimeTypeDetector = $mimeTypeDetector;
        $this->storage = $storage;
        $this->userSession = $userSession;
        $this->userManager = $userManager;
        $this->accountManager = $accountManager;
        $this->timeFactory = $timeFactory;
        
        $this->variableSubstitution = new VariableSubstitution($this->config, $this->userSession, $this->userManager, $this->accountManager);
    }

    public function getInitialViewerJSON($file_name_to_sign, $profile_data, $json_without_base64_document = false) {
        \OCP\Util::writeLog('signlive', 'getInitialViewerJSON('.$file_name_to_sign.')', ILogger::DEBUG);
        return $this->getJSONQuery($file_name_to_sign, $profile_data, $json_without_base64_document);
    }

    public function getViewerResultJSON($viewer_conversation_id) {
        \OCP\Util::writeLog('signlive', 'getViewerResultJSON('.$viewer_conversation_id.')', ILogger::DEBUG);
        $acknowledge_request_string = "{
            \"conversation\": \"".$viewer_conversation_id."\"
        }";

        return $acknowledge_request_string;
    }

    protected function insertDocumentSection($signature_request_json, $profile_data, $json_without_base64_document = false) {
        $signature_type_found = array_key_exists('signature_type', $profile_data);
        $encoded_document_found = array_key_exists('base64_encoded_document', $profile_data);
        $file_name_found = array_key_exists('file_name', $profile_data);
        
        $logging_format_active = 0;
        if ($json_without_base64_document === true) {
            $logging_format_active = 1;
        }

        \OCP\Util::writeLog('signlive', 'insertDocumentSection() called. $file_name_found = '.$file_name_found.', $encoded_document_found = '.$encoded_document_found.', $signature_type_found = '.$signature_type_found.', $json_without_base64_document = '.$logging_format_active, ILogger::DEBUG);
        if ($signature_type_found && 
            $encoded_document_found &&
            $file_name_found) {
            $document_placeholder_text = '/\$\{sl\.system\.documents_section\}/';
            \OCP\Util::writeLog('signlive', 'signature_type = '.$profile_data['signature_type'], ILogger::DEBUG);
            $document_section_identifier = '';
            if ($profile_data['signature_type'] == 'viewer_based_signature') {
                $document_section_identifier = 'document';
            } else if ($profile_data['signature_type'] == 'automatic_signature') { 
                $document_section_identifier = 'documents';
            }
            \OCP\Util::writeLog('signlive', 'document_section_identifier = '.$document_section_identifier, ILogger::DEBUG);

            $document_section = '';
            if ($document_section_identifier != '') {
                if ($profile_data['signature_type'] == 'viewer_based_signature') {
                    if ($json_without_base64_document === false) {
                        $document_section = "\"".$document_section_identifier."\": {
                            \"type\": \"d\",
                            \"name\": \"".$profile_data['file_name']."\",
                            \"content\": \"".$profile_data['base64_encoded_document']."\"
                        }";
                    } else {
                        $document_section = "\"".$document_section_identifier."\": {
                            \"type\": \"d\",
                            \"name\": \"".$profile_data['file_name']."\",
                            \"content\": \"<base64 encoded document omitted>\"
                        }";
                    }
                } else if ($profile_data['signature_type'] == 'automatic_signature') {
                    $file_path_with_replaced_separators = str_replace('/', ':::', $profile_data['file_name']);
                    if ($json_without_base64_document === false) {
                        $document_section = "\"".$document_section_identifier."\": [ {
                            \"type\": \"d\",
                            \"name\": \"".$file_path_with_replaced_separators."\",
                            \"content\": \"".$profile_data['base64_encoded_document']."\"
                        } ]";
                    } else {
                        $document_section = "\"".$document_section_identifier."\": [ {
                            \"type\": \"d\",
                            \"name\": \"".$file_path_with_replaced_separators."\",
                            \"content\": \"<base64 encoded document omitted>\"
                        } ]";
                    }
                }

                // Do not log the potentially very long JSON section with the encoded base64 content here.
                // This is only relevant when a signature request causes an error
                // These errors are handled and logged in IntarsysJobSignatureController
                // \OCP\Util::writeLog('signlive', 'json_request_with_document = '.$json_request_with_document, ILogger::DEBUG);
                // \OCP\Util::writeLog('signlive', 'document_section = '.$document_section, ILogger::DEBUG);

                $json_request_with_document = preg_replace($document_placeholder_text, $document_section, $signature_request_json);

                return $json_request_with_document;
            }
        } else {
            \OCP\Util::writeLog('signlive', 'insertDocumentSection() failed to place base64-encoded document into JSON signature query!', ILogger::ERROR);
        }
        return '';
    }

    public function getJSONQuery($file_name_to_sign, $profile_data, $json_without_base64_document = false) {
        \OCP\Util::writeLog('signlive', 'getJSONQuery('.$file_name_to_sign.')', ILogger::DEBUG);
        $processed_json_request = $this->variableSubstitution->substituteVariableValues($profile_data, $json_without_base64_document);
        \OCP\Util::writeLog('signlive', 'JSON query with replaced variable placeholders received: '.$processed_json_request, ILogger::DEBUG);
        $complete_json_request = $this->insertDocumentSection($processed_json_request, $profile_data, $json_without_base64_document);
        if ($complete_json_request == '') {
            \OCP\Util::writeLog('signlive', 'Inserting base64-encoded document into JSON query failed! Can not continue.', ILogger::ERROR);
            return '';
        }
        return $complete_json_request;
    }
}