<?php

/**
* This file is licensed under the Affero General Public License version 3 or
* later. See the COPYING file.
*
* @author Fabian Aichele <fabian.aichele@bpex.de>
* @copyright intarsys AG 2021
*/

namespace OCA\IntarsysCloudSuiteGearsIntegration\Util;

use OCP\Util;
use OCP\ILogger;
use OCP\IUserSession;
use OCP\IUserManager;
use OCP\Accounts\IAccount;
use OCP\Accounts\IAccountManager;

use OCP\IConfig;
use OCA\IntarsysCloudSuiteGearsIntegration\Util\SignLiveDefaultConfigValues;

class VariableSubstitution {
    /** @var \OCP\IUserSession */
    protected $userSession;

    /** @var \OCP\IConfig */
    protected $config;

    protected $userManager;
    protected $accountManager;

    protected $nc_variable_placeholders;

    protected $sl_variable_placeholders;

    protected $user;
    protected $user_account;

    public function __construct(IConfig $config,
                                IUserSession $userSession,
                                IUserManager $userManager,
                                IAccountManager $accountManager) {
        $this->config = $config;
        $this->userSession = $userSession;
        $this->userManager = $userManager;
        $this->accountManager = $accountManager;

        try {
            $this->uid = $this->userSession && $this->userSession->isLoggedIn() ? $this->userSession->getUser()->getUID() : false; 
            // $this->userSession->getUser()->getUID();
            $this->user = $this->userManager->get($this->uid);
            $this->userAccount = null;
            if ($this->user !== null) {
                $this->userAccount = $this->accountManager->getAccount($this->user);
            }
            // \OCP\Util::writeLog('signlive', 'User UID determined in signlive/lib/Util/VariableSubstitution.php: '.$this->uid, ILogger::DEBUG);
        } catch (Exception $e) {
            \OCP\Util::writeLog('signlive', 'Failed to determine user UID in signlive/lib/Util/VariableSubstitution.php: '.$e->getMessage(), ILogger::ERROR);
            $this->uid = false;
            $this->user = null;
            $this->userAccount = null;
        }

        if ($this->uid !== false) {
            // These arrays could be expanded programmatically instead of just holding pre-defined values
            $this->nc_variable_placeholders = [
                'nc.displayname' => $this->user->getDisplayName(),
                'nc.email' => $this->user->getEMailAddress(),
                'nc.phonenumber' => $this->userAccount->getProperty('phone')->getValue(),
                'nc.address' => $this->userAccount->getProperty('address')->getValue(),
                'nc.website' => $this->userAccount->getProperty('website')->getValue()
            ];

            $this->sl_variable_placeholders = [
                'sl.givenname' => $this->config->getUserValue($this->uid, 'signlive', 'cloudsuite_gears_user_attribute_given_names', ''),
                'sl.surname' => $this->config->getUserValue($this->uid, 'signlive', 'cloudsuite_gears_user_attribute_surname', ''),
                'sl.title' => $this->config->getUserValue($this->uid, 'signlive', 'cloudsuite_gears_user_attribute_title', ''),
                'sl.msisdn' => $this->config->getUserValue($this->uid, 'signlive', 'cloudsuite_gears_user_attribute_msisdn', ''),
                'sl.companyname' => $this->config->getAppValue('signlive', 'cloudsuite_gears_global_attribute_company_name', SignLiveDefaultConfigValues::$DEFAULT_GLOBAL_ATTRIBUTE_COMPANY_NAME),
                'sl.companydivision' => $this->config->getUserValue($this->uid, 'signlive', 'cloudsuite_gears_user_attribute_company_division', ''),
                'sl.signmeusername' => $this->config->getUserValue($this->uid, 'signlive', 'cloudsuite_gears_user_attribute_signme_user_name', ''),
                'sl.encodedsignatureimage' => $this->config->getUserValue($this->uid, 'signlive', 'cloudsuite_gears_user_attribute_encoded_signature_image', ''),
                'sl.country' => $this->config->getUserValue($this->uid, 'signlive', 'cloudsuite_gears_user_attribute_country', '')
            ];
        } else {
            $this->nc_variable_placeholders = [
                'nc.displayname' => '',
                'nc.email' => '',
                'nc.phonenumber' => '',
                'nc.address' => '',
                'nc.website' => ''
            ];

            $this->sl_variable_placeholders = [
                'sl.givenname' => '',
                'sl.surname' => '',
                'sl.title' => '',
                'sl.msisdn' => '',
                'sl.companyname' => '',
                'sl.companydivision' => '',
                'sl.signmeusername' => '',
                'sl.encodedsignatureimage' => '',
                'sl.country' => ''
            ];
        }

        $this->nc_variable_placeholder_names = [
            'nc.displayname',
            'nc.email',
            'nc.phonenumber',
            'nc.address',
            'nc.website'
        ];

        $this->sl_variable_placeholder_names = [
            'sl.givenname',
            'sl.surname',
            'sl.title',
            'sl.msisdn',
            'sl.companyname',
            'sl.companydivision',
            'sl.signmeusername',
            'sl.encodedsignatureimage',
            'sl.country',
            'sl.system.redirect_url',
            'sl.system.document_section'
        ];

        $this->sl_base64_variables = [
            'sl.encodedsignatureimage'
        ];
    }

    public function substituteVariableValues($profile_data, $json_without_base64_document = false) {
        \OCP\Util::writeLog('signlive', 'substituteVariableValues() called in signlive/lib/Util/VariableSubstitution.php', ILogger::DEBUG);
        \OCP\Util::writeLog('signlive', 'JSON section to process: '.$profile_data['json_section'], ILogger::DEBUG);

        $processed_json_template = '';

        // If variable substitiution values are not available for some reason: Return empty string
        if ($this->uid === false) {
            \OCP\Util::writeLog('signlive', 'substituteVariableValues() called without a valid Nextcloud user object! Variable substitution is non-functional, returning empty JSON section.', ILogger::WARN);
            return $processed_json_template;
        }

        if (array_key_exists('json_section', $profile_data)) {
            $processed_json_template = $profile_data['json_section'];
            
            // Substitute variable values from the Nextcloud user profile
            foreach ($this->nc_variable_placeholders as $placeholder_text => $substitution_value) {
                \OCP\Util::writeLog('signlive', 'Replacing nc placeholder \''.$placeholder_text.'\' with \''.$substitution_value.'\'', ILogger::DEBUG);
                $escaped_placeholder_text = preg_replace('/\./', '\\.', $placeholder_text);
                $search_pattern = '/\$\{'.$escaped_placeholder_text.'\}/';
                $processed_json_template = preg_replace($search_pattern, $substitution_value, $processed_json_template);
            }

            // Substitute variable values specific to the the Gears app (user-specific + installation-wide)
            foreach ($this->sl_variable_placeholders as $placeholder_text => $substitution_value) {
                \OCP\Util::writeLog('signlive', 'Replacing sl placeholder \''.$placeholder_text.'\' with \''.$substitution_value.'\'', ILogger::DEBUG);
                $escaped_placeholder_text = preg_replace('/\./', '\\.', $placeholder_text);
                $search_pattern = '/\$\{'.$escaped_placeholder_text.'\}/';
                
                $replacement_text = $substitution_value;
                if ($json_without_base64_document) {
                    if (in_array($placeholder_text, $this->sl_base64_variables)) {
                        \OCP\Util::writeLog('signlive', 'Omitting sl placeholder \''.$placeholder_text.'\', it is marked as base64-encoded content and is not to be part of log messages.', ILogger::DEBUG);
                        $replacement_text = '<base64 encoded content omitted>';
                    }
                }
                
                $processed_json_template = preg_replace($search_pattern, $replacement_text, $processed_json_template);
            }

            // Substitute reserved values, except for the actual base64-encoded document to sign
            // redirectUrl
            if (array_key_exists('redirect_url', $profile_data)) {
                $redirect_url_placeholder_text = '/\$\{sl\.system\.redirect_url\}/';
                \OCP\Util::writeLog('signlive', 'Replacing sl redirect_url placeholder with: \''.$profile_data['redirect_url'].'\'', ILogger::DEBUG);
                $processed_json_template = preg_replace($redirect_url_placeholder_text, $profile_data['redirect_url'], $processed_json_template);
            }

            // Remove all occurrences of variables without an associated value here
            foreach ($this->nc_variable_placeholders as $placeholder_text => $substitution_value) {
                \OCP\Util::writeLog('signlive', 'Removing nc placeholder \''.$placeholder_text.', no or empty substitution value provided.', ILogger::DEBUG);
                $escaped_placeholder_text = preg_replace('/\./', '\\.', $placeholder_text);
                $search_pattern = '/\$\{'.$escaped_placeholder_text.'\}/';
                $processed_json_template = preg_replace($search_pattern, '', $processed_json_template);
            }

            foreach ($this->sl_variable_placeholders as $placeholder_text => $substitution_value) {
                \OCP\Util::writeLog('signlive', 'Removing sl placeholder \''.$placeholder_text.', no or empty substitution value provided.', ILogger::DEBUG);
                $escaped_placeholder_text = preg_replace('/\./', '\\.', $placeholder_text);
                $search_pattern = '/\$\{'.$escaped_placeholder_text.'\}/';
                $processed_json_template = preg_replace($search_pattern, '', $processed_json_template);
            }
        }
        \OCP\Util::writeLog('signlive', 'JSON snippet with replaced placeholders: '.$processed_json_template, ILogger::DEBUG);
        return $processed_json_template;
    }

    public function validateJSONSubstitutionVariables($json_section_text) {
        $validation_result = [
            'success' => true,
            'status_message' => 'Validated successfully.' 
        ];

        $pattern = '/\$\{([^\$\{]+)\}/'; //a-zA-Z0-9\s [^\{\$]
        preg_match_all($pattern, $json_section_text, $matches);

        \OCP\Util::writeLog('signlive', '=========================================================', ILogger::DEBUG);
        \OCP\Util::writeLog('signlive', 'JSON section: '.print_r($json_section_text, TRUE), ILogger::DEBUG);
        \OCP\Util::writeLog('signlive', 'Validation regex search result: '.print_r($matches, TRUE), ILogger::DEBUG);
        \OCP\Util::writeLog('signlive', 'matches[0] size: '.count($matches[0]), ILogger::DEBUG);
        \OCP\Util::writeLog('signlive', 'matches[1] size: '.count($matches[1]), ILogger::DEBUG);
        
        if (count($matches) > 0 && count($matches[0]) > 0) {
            for ($i = 0; $i < count($matches[0]); $i++) {
                $placeholder_candidate = $matches[0][$i];
                $placeholder_candidate = trim($placeholder_candidate);
                $placeholder_candidate = str_replace(["$", "{", "}", "\"", "\n", "\t"], ["", "", "", "", "", ""], $matches[0][$i]);
                $placeholder_candidate = preg_replace('/\s+$/', '', $placeholder_candidate);
                \OCP\Util::writeLog('signlive', 'Checking variable substitution candidate: \''.$placeholder_candidate.'\'', ILogger::DEBUG);
                
                $is_valid_nc_placeholder = 0;
                $is_valid_sl_placeholder = 0;

                if ($placeholder_candidate === 'sl.system.redirect_url' ||
                    $placeholder_candidate === 'sl.system.documents_section') {
                    $is_valid_sl_placeholder = 1;
                    $is_valid_nc_placeholder = 1;

                    \OCP\Util::writeLog('signlive', 'Reserved placeholder \''.$placeholder_candidate.'\': \''.$is_valid_sl_placeholder.'\' == \''.$is_valid_sl_placeholder.'\'', ILogger::DEBUG);
                } else {
                    foreach ($this->sl_variable_placeholder_names as $valid_sl_placeholder) {
                        \OCP\Util::writeLog('signlive', 'Comparing: \''.$valid_sl_placeholder.'\' == \''.$placeholder_candidate.'\'', ILogger::DEBUG);
                        if ($valid_sl_placeholder == $placeholder_candidate) {
                            $is_valid_sl_placeholder = 1;
                            break;
                        }
                    }
    
                    foreach ($this->nc_variable_placeholder_names as $valid_nc_placeholder) {
                        \OCP\Util::writeLog('signlive', 'Comparing: \''.$valid_nc_placeholder.'\' == \''.$placeholder_candidate.'\'', ILogger::DEBUG);
                        if ($valid_nc_placeholder == $placeholder_candidate) {
                            $is_valid_nc_placeholder = 1;
                            break;
                        }
                    }
                }

                \OCP\Util::writeLog('signlive', 'Is a valid NC placeholder: \''.$is_valid_nc_placeholder.'\'; is a valid SL placeholder: \''.$is_valid_sl_placeholder.'\'', ILogger::DEBUG);
                if ($is_valid_sl_placeholder == 0 && $is_valid_nc_placeholder == 0) {
                    \OCP\Util::writeLog('signlive', 'Rejected placeholder: \''.$placeholder_candidate.'\'', ILogger::DEBUG);
                    if ($placeholder_candidate !== 'sl.system.redirect_url' ||
                        $placeholder_candidate !== 'sl.system.documents_section') {
                        $validation_result['success'] = false;
                        $validation_result['status_message'] = 'Variable placeholder \''.$placeholder_candidate.'\' is not a valid SignLive variable name! Please remove or replace it from the JSON section of your signature profile definition.';
                    }
                    break;
                }
            }
        }
        \OCP\Util::writeLog('signlive', '=========================================================', ILogger::DEBUG);

        return $validation_result;
    }
}
